/* File InOut.c; TRANSIENT V4.3; 30 October 2023 */
/* Copyright (C) 1995-2018,2021,2023 Miroslav Kolar */

/*	Input and output procedures		*/

#if defined(MSDOS) || defined(__BORLANDC__) || defined(_MSC_VER) || defined(__MINGW32__) || defined(__MINGW64__)
# define _PC_
#endif

#if defined(_PC_) || (!defined(__CYGWIN__) && defined(__CYGWIN32__))
# define _PCorCYGold_
#endif
/* In Cygwin B20 and older, only __CYGWIN32__ was defined;
   in newer versions, both __CYGWIN32__ and __CYGWIN__ are defined,
   and these are more UNIX-like: Ctrl-D, ...
*/

#if defined(vms) || defined(VMS)
# define _VMS_
/*# define _Exit exit*/
#endif
#ifdef __DJGPP
# define _Exit exit
#endif

#if defined(_PC_) /* Add here other cases when the times() function is missing */
# define _ONE_CLOCK_
#endif

#include "incl_pri.h"
#include <errno.h>

#include <ctype.h>
#ifndef _ONE_CLOCK_
#include <sys/times.h>
#include <unistd.h>
// See line 969 for comment on unistd.h
#else
#include <time.h>
#ifndef CLK_TCK
# define CLK_TCK	CLOCKS_PER_SEC
#endif
#endif

#include "DateLen.h"

/* Max length + 1 of the input parameter names and values and the file names: */
#ifdef _VMS_
# define MAX1_STR_LEN	120
#else
# define MAX1_STR_LEN	80
#endif

/* Defined in DateTime.c: */
extern char *DateAndTime(void);

/* Defined in transien.c: */
extern void DiagEntry_at_t(void);

/* Defined in a problem definition module (PDM) - the module describing
   each individual model (set of reaction-diffusion equations).
   Or indirectly in transien.h since version V2.5:  */
extern int INPT_model_param(int words, char *str1, char *str2, int *w);
extern int WRONG_model_param(void);
extern void pr_some_PDMpar(char *,int);
extern void pr_model(int,char *);
extern void pr_profiles(int);

/* Defined in nl3bandv.c or in nl3band1.c: */
extern void report_last_NR_error(FILE *out, char comment_char);
extern void endband(void);


/* Input parameters (control) and version number
   --------------------------

1. Imported from transien.c: */

#define IN_PAR(name,type,iV,oFrmt,Expl) extern type name;
#include "CTRLpar.h"
#undef IN_PAR

extern char *versn, *title, *time_units;
extern double t, tStep, ttEnd, tSign;
extern unsigned int do_Oth_Prof;
extern int Nexcl;
extern FILE *diag_out, *par_out, *out, *timeDep_out;
extern void out_profiles(int);
extern void out_Other(int);
extern double (* cDifMxf)(double);
extern long strict1failCount;

/* 2. Exported from here to transien.c and the PDM: */
char *init_from_file = NULL, *cDiffMfExcl = NULL;
// To PDM only:
FILE *stat_out=NULL;
// To transien.c only:
short int init_from_file_dyn=0, cDiffMfExcl_dyn=0;
// These two variables will be set to 1 when the storage for the
// respective strings is created dynamically.

/* Variables global to this file: */
static int o_n_len, def_par = 1;
static char str1[MAX1_STR_LEN], str2[MAX1_STR_LEN], *progName, *rootName, *defRoot = NULL;
static double clk_tck;
#ifdef _ONE_CLOCK_
static clock_t cl0, clTr0, clLast, clTo;
#else
static clock_t us0, sy0, usTr0, syTr0, usLast, syLast, usTo, syTo;
static struct tms cpuT;
#endif
static long stCoLast=0, rbCoLast=0, itCoLast=0, tDepHeadEndPos;
static char jobStartTime[DATELEN], defCmntC;
static double cDifMxA, cDifMxB;

/* Needed also in signals.c, dynarr.c and nl3_Decl.h: */
// MSG_LENGTH must be larger than about MAX1_STR_LEN+95 and than about 2*MAX1_STR_LEN+78:
#if MAX1_STR_LEN+95 > 2*MAX1_STR_LEN+80
#define MSG_LENGTH (MAX1_STR_LEN+95)
#else
#define MSG_LENGTH (2*MAX1_STR_LEN+80)
#endif
const int msgLength = MSG_LENGTH;
char msg[MSG_LENGTH] = "\0";

#define PRl	(void)snprintf(msg,(size_t)msgLength,
#define PRlc	i=strlen(msg); (void)snprintf(msg+i,(size_t)(msgLength-i),


/* These are defined below, but used before their definitions: */
void pr_par(int);
void open_statOut(void);
void exitTR(int,char*) __attribute__ ((noreturn));
void exitTRsimple(int) __attribute__ ((noreturn));
void file_header(char*,FILE*);


int isTooLong(int lng)
/*==================*/
{
 return (lng>=MAX1_STR_LEN);
}


void get_defRoot(void)
/*==================*/
{
if(!defRoot) {
#ifdef DEF_OUTPUT_ROOT
	defRoot = DEF_OUTPUT_ROOT;
#else
	unsigned int i, l;
	l = strlen(progName);
	defRoot = malloc((l+2)*sizeof(char));
	for(i=0;i<l;i++) defRoot[i]= toupper(progName[i]);
	defRoot[l] = 0;
#endif
}
}


void pr_model_help(int explain)
/*===========================*/
{
#if PDMtrace
PE"... {PDM header file} - ouput\n");
#endif
pr_model(explain,"");
}


double linInter(double in, int NpointsM2, double *inb, double *v, int *lowI)
/*==========================================================================
 linear interpolation of  v  between inb[] points:
 inb[] is independent variable;
 in is its value for which v is to be calculated;
 v[] are tabulated dependent variable values;
 NpointsM2 is the # of points minus 2;
 *lowI is the index of the lower bound of v from the previous call
 - as usually the new value of  in  will be close to the previous one	*/
{
 register int i, lI;
 if(NpointsM2==-1) return *v;
 i = lI = *lowI;
 while(lI > 0 && inb[lI] > in) lI--;
 if(i == lI) { while(lI < NpointsM2 && inb[lI+1] <= in) lI++; }
 i = lI + 1;
 *lowI = lI;
 return (v[lI] +
	(v[i] - v[lI]) * (in - inb[lI]) / (inb[i] - inb[lI]));
}


double cDifMxf0(/*@unused@*/double tt __attribute__((unused)))
/*==========================================================*/
{ return cDifMxInit; }

double cDifMxf1(double tt)
/*======================*/
{ return cDifMxA+cDifMxB*tt; }

double cDifMxf2(double tt)
/*======================*/
{ return cDifMxA*pow(tt,cDifMxB); }


double cDifMxEnd= 1.;
static double *cDifMxP, *tcDifMxP= NULL;
static int iP=0, lowP=0, tcDPconvJ=0;
 // iP corresponds to the highest cDifMxP? point entered, may be higher than cDifMxMode;
 // tcDPconvJ is the index of the highest cDifMxP? converted (-tStart, possibly *(-1));
 // the unused cDifMxP? points are never converted;
 // initially no points are converted, that's why tcDPconvJ is needed


double cDifMxf3(double tt)
/*======================*/
{ return linInter(tt,iP,tcDifMxP,cDifMxP,&lowP); }


void ptitle(void)
/*=============*/
{
PE"\n%s", title);
PE"%s", versn);
}


#define CLOSE(File) { (void)fclose(File); File= NULL; }


static char
#ifdef _VMS_
	 *sIn = "SYS$INPUT",
	 *sOut = "SYS$OUTPUT",
	 *sErr = "SYS$ERROR";
#else
	 *sIn = "`stdin`",
	 *sOut = "`stdout`",
	 *sErr = "`stderr`";
#endif


int read_inp_value(char *name, void *var, char *type, char *src_string)
/*===================================================================*/
{
 char *end;
 int Ret = 0;

 errno = 0;
 if(!strcmp(type,"int")) *(int *)var = (int)strtol(src_string, &end, 10);
 else if(!strcmp(type,"short int")) *(short int *)var = (short int)strtol(src_string, &end, 10);
 else if(!strcmp(type,"long int")) *(long int *)var = (long int)strtol(src_string, &end, 10);
 else if(!strcmp(type,"unsigned int")) {
	*(unsigned int *)var = (unsigned int)strtoul(src_string, &end, 10);
	if(*end=='u' || *end=='U') end++;
 } else if(!strcmp(type,"float")) *(float *)var = strtof(src_string, &end);
 else if(!strcmp(type,"double")) *(double *)var = strtod(src_string, &end);
 else if(!strcmp(type,"long double")) *(long double *)var = strtold(src_string, &end);
 else if(!strcmp(type,"char")) {
	*(char *)var = *src_string;
	end = src_string+1;
 } else {
	PRl"Should never got here: failure of default in check_PAR_value() in transien.h\
for type \"%s\" used for parameter \"%s\".", type, name);
	exitTRsimple(-5);
 }
 if(errno != 0) { perror("strtol"); Ret= 1; }
 if(end == src_string || *end != '\0') {
	PRl"  The input value of \"%s\" (%s), %s, has illegal trailing characters: %s\n",
		name, type, src_string, end);
	PE"%s", msg);
	if(!stat_out) open_statOut(); Ps"%s", msg);
	Ret= 1;
 }
 return Ret;
}


int read_inp_str(char *name, char *var, size_t maxLng, char *src_string)
/*====================================================================*/
{
 size_t sLeng= strlen(src_string);

 if(sLeng > maxLng) {
	PRl"  The input value of \"%s\" (string), %s, is longer than %d characters reserved for it.\n",
		name, src_string, (int)maxLng);
	PE"%s", msg);
	if(!stat_out) open_statOut(); Ps"%s", msg);
	return 1;
 }
 memcpy(var, src_string, sLeng);
 var[sLeng]='\0';
 return 0;
}


void info(void)
/*===========*/
{
ptitle();
PE"Usage:\n\n To get help:  %s -h  ...  prints this page\n", progName);
PE"               %s -d  ...  lists all input parameters and\n",progName);
PE"                                     their default values\n");
PE"               %s -p  ...  as -d plus any available info on parameters\n", progName);
PE"               %s -c  ...  lists only control parameters, their purpose\n",progName);
PE"                                     and default values\n\n");
PE" To run:  %s\n", progName);
PE"          %s 'Output'\n", progName);
PE"          %s 'Output' 'Input'\n", progName);
#ifndef _VMS_
PE"          %s < 'Input'\n", progName);
#endif
PE"     'Output' :  root of names of all output files: For a steady-state run, or\n");
PE"        if sepProf=0 is set in a transient run, there is a single file 'Output'\n");
PE"        for all c (concentration) profiles (for all output times in a transient\n");
PE"        run). Otherwise profiles for I-th output time are written into 'Output'.I.\n");
PE"        Initial profiles are optionally copied into a separate file 'Output'.0.\n");
PE"        Input parameters are written out into a file with a name created\n");
PE"        by adding extension \".par\" to the root 'Output'. Additional output\n");
PE"        (mostly time dependent quantities) generated by init_Other, update_Other\n");
PE"        and pr_Other is written into a file with extension \".tdep\". (Although\n");
PE"        PDM author can direct additional output to any of the output files.)\n");
PE"        Diagnostic output (minima and maxima of tStep, cDiffM, and number of\n");
PE"        all iterations between pr_Other times, or at most in 1000 time\n");
PE"        steps) is written into a file with extension \".diag\".\n");
PE"        CPU times and more are written into a file with extension \".stat\".\n");
PE"     'Input'  :  name of the input file; only some or none input\n");
PE"        parameters need be read in, defaults are used for the rest.\n");
PE"     If sepProf=-1, no profiles output files (except for optional 'Output'.0)\n");
PE"     and no 'Output'.diag are created.\n");
PE"     If no input and/or output file names are specified, the following default\n");
PE"     names will be used: 'Input'  :  %s\n", sIn);
get_defRoot();
PE"                         'Output' :  %s\n", defRoot);
PE"        ( [%s.0]   - the optional echo of initial profiles;\n", defRoot);
PE"          %s  or  %s.1, %s.2, %s.3, ...   - c profiles;\n", defRoot, defRoot, defRoot, defRoot);
PE"          %s.par, [%s.tdep,] [%s.diag,] %s.stat   - see above )\n", defRoot, defRoot, defRoot, defRoot);
PE"     %s : interactively from command line, or from a pipe.\n\n", sIn);
FE
}


void interactInfo(void)
/*===================*/
{
PE"\n To get help on how to run this program (e.g., specify the input and\n\
\t output files), enter:  ???\n\n\
 If you continue now, the following output files will be created:\n");
if(echoInit) PE" %s.0,", rootName);
if(sepProf>-1) {
 if(steady==1||!sepProf) PE" %s,", rootName);
 else PE" %s.1, %s.2, %s.3, ...,\n", rootName, rootName, rootName);
}
PE" %s.par, %s.tdep", rootName, rootName);
if(sepProf>-1) PE", %s.stat and %s.diag .\n\n", rootName, rootName);
else PE" and %s.stat .\n\n", rootName);
if(def_par) {
 PE" If you end input now, default input parameter values would be used.\n\n");
 PE" You can at first modify any input parameters in arbitrary order, one\n");
 PE" at a time, by entering the name of an input parameter");
} else PE" To modify an input parameter, enter its name");
PE" followed by its\n");
PE" new value, separated from the name by spaces or/and an equal sign.\n");
PE" Missing value is identical to a zero number, empty character or empty string.\n\n");
PE" To end input (start calculation), enter  end  or  ");
#ifdef _PCorCYGold_
PE" Ctrl+Z");
#elif defined(_VMS_)
PE"press  Ctrl+Z");
#else
PE"press  Ctrl+D");
#endif
PE" .\n\n At any time during this input phase, you can also enter:\n");
PE"  ?c-      to list the current values of all control parameters;\n");
PE"  ??c-     to also get extensive help on the role of all control parameters;\n");
PE"  ?m-      to list the current values of all model parameters;\n");
PE"  ??m-     to also get available help on model parameters;\n");
PE"  ?<str>   to list the current values of all input parameters whose\n");
PE"		names start with <str>;  e.g. ?tSt;\n");
PE"  ?\?<str>  to also get help for input parameters starting with <str>;\n");
PE"  ?        to reprint this interactive menu;\n");
PE"  qu[it]   to terminate execution;\n");
PE"  by[e]      	- \" -\n\n");
PE"  Ctrl+C   will terminate execution at any stage.\n");
}


const char
*init_from_file_Descr =
"   Name of file containing the initial profiles; if non-NULL, the\n\
   user supplied function read_profiles() will be invoked just after\n\
   Initialize(), and initial profiles set in Initialize() will be\n\
   overwritten by those read in in read_profiles(). Can be used to\n\
   continue an interrupted run (possibly after changing some parameters),\n\
   by reading in the last good profile output file in read_profiles().",
*cDiffMfExcl_Descr =
"   'Exclusion list' of the indices of equations (dependent variables c)\n\
   that will not be used in cresidT/M and fresidT/M in NR iterations, and\n\
   in transient runs also when calculating the \"relative\" change in c\n\
   in each time step, cDiffM. Only those c not in this list will\n\
   contribute to the above 'error indicators'.\n\
   The list is comma separated and preceded by the number of indices in\n\
   the list (Nexcl) and a colon, no spaces.  Eg: cDiffMfExcl = 3:2,3,9 .\n\
   Empty list: all c are used in the calculation of the convergence indicators.",
*cDifMxEnd_Descr =
"   Value of cDifMxf at t = tEnd (used only for cDifMxMode>0)";


char *prettyPr(double D, char *frmt)
/*================================*/
{
	char *i, *e;
	PRl frmt, D);
	i= msg;
	while(*(++i) != 'e') ;
	e= i;
	while(*(--i) == '0') ;
	if(*i == '.') i--;
	while(*e != '\0') *(++i) = *(e++);
	*(++i) = '\0';
	return msg;
}


void pr_some_CTRLpar(char *pName, int Xplain)
/*=========================================*/
{
size_t len= strlen(pName);
if(len==0) return;
#ifndef ultrix
# define IN_PAR(name,typ,iV,outFrmt_unit,expl) if(!strncmp(pName,#name,len)){\
	Pp #name" = ");\
	if(!(strcmp(#name,"tEnd")&&strcmp(#name,"tOutInit"))) Pp"%s\n", prettyPr(name,#outFrmt_unit));\
	 else Pp #outFrmt_unit"\n", name);\
	if(Xplain && strlen(#expl)) Pp"   "#expl"\n");}
#else
# define IN_PAR(name,typ,iV,outFrmt_unit,expl) if(!strncmp(pName,"#name",len)){\
	Pp "name = ");\
	if(!(strcmp("name","tEnd")&&strcmp("name","tOutInit"))) Pp"%s\n",prettyPr(name,"outFrmt_unit"));\
	 else Pp"outFrmt_unit\n", name);\
	if(Xplain && strlen("expl")) Pp"   expl\n");}
#endif
#include "CTRLpar.h"
#undef IN_PAR
if(iP && !strncmp(pName,"cDifMxP",len<8?len:7)) {
	register int j;
	if(len<8) {
		for(j=1; j<iP-1; j++) if(cDifMxP[j]!=-1.) {
			Pp"cDifMxP%d = %s,%.8g\n", j,
			prettyPr((j<=tcDPconvJ?tStart+tSign*tcDifMxP[j]:tcDifMxP[j]),"%.10e"), cDifMxP[j]);
		}
	} else {
		register int i= atoi(pName+7), mf=10;
		if(i>0 && i<iP-1) {
			if(cDifMxP[i]!=-1.) Pp"cDifMxP%d = %s,%.8g\n", i,
				prettyPr((i<=tcDPconvJ?tStart+tSign*tcDifMxP[i]:tcDifMxP[i]),"%.10e"), cDifMxP[i]);
			while((i*=10)<iP-1) {
				for(j=i; j<i+mf && j<iP-1; j++) if(cDifMxP[j]!=-1.)
					Pp"cDifMxP%d = %s,%.8g\n", j,
					prettyPr((j<=tcDPconvJ?tStart+tSign*tcDifMxP[j]:tcDifMxP[j]),"%.10e"), cDifMxP[j]);
				mf*=10;
			}
		}
	}
}
if(!strncmp(pName,"cDifMxEnd",len)) {
	Pp"cDifMxEnd = %s\n", prettyPr(cDifMxEnd,"%.8e"));
	if(Xplain) Pp"%s\n", cDifMxEnd_Descr);
}
if(!strncmp(pName,"cDiffMfExcl",len)) {
	Pp"cDiffMfExcl =");
	if(cDiffMfExcl) Pp" %s", cDiffMfExcl);
	Pp"\n");
	if(Xplain) Pp"%s\n", cDiffMfExcl_Descr);
}
if(!strncmp(pName,"init_from_file",len)) {
	Pp"init_from_file =");
	if(init_from_file) Pp" %s", init_from_file);
	Pp"\n");
	if(Xplain) Pp"%s\n", init_from_file_Descr);
}
}


#define EXC	Xplain>0&&Xplain<3
#ifndef ultrix
# define IN_PAR(name,type,iV,outFrmt_unit,expl) if(EXC) Pp"\n");\
 Pp "%s"#name" = ", cmntS);\
 if(!(strcmp(#name,"tEnd")&&strcmp(#name,"tOutInit"))) Pp"%s\n", prettyPr(name,#outFrmt_unit));\
  else Pp #outFrmt_unit"\n", name);\
 if(EXC && strlen(#expl)) Pp"   "#expl"\n");
#else
# define IN_PAR(name,type,iV,outFrmt_unit,expl) if(EXC) Pp"\n");\
 Pp "%sname = ", cmntS);\
 if(!(strcmp("name","tEnd")&&strcmp("name","tOutInit"))) Pp"%s\n", prettyPr(name,"outFrmt_unit"));\
  else Pp"outFrmt_unit\n", name);\
 if(EXC && strlen("expl")) Pp"   expl\n");
#endif


extern int *equExcl;

void pr_CTRLpar(int Xplain, char *cmntS)
/*====================================*/
{
if(Xplain<3) {
	Pp"%sControl parameters", cmntS);
	if(Xplain==2 && def_par) Pp", their default values and their purpose");
	Pp":\n%s------------------\n", cmntS);
}
#include "CTRLpar.h"
if(EXC) Pp"\n");
if(iP && (Xplain>=0 || cDifMxMode>2)) {
	register int j;
	if(Xplain<0 && iP>cDifMxMode) iP=cDifMxMode;
	for(j=1; j<iP-1; j++) if(cDifMxP[j]!=-1.) Pp"%scDifMxP%d = %s,%.8g\n", cmntS, j,
		prettyPr((j<=tcDPconvJ?tStart+tSign*tcDifMxP[j]:tcDifMxP[j]),"%.10e"), cDifMxP[j]);
	if(EXC) Pp"\n");
}
if(cDifMxMode || Xplain>0) {
	Pp"%scDifMxEnd = %s\n", cmntS, prettyPr(cDifMxEnd,"%.8e"));
	if(EXC) Pp"%s\n\n", cDifMxEnd_Descr);
}
Pp"%scDiffMfExcl =", cmntS);
if(Xplain==-2) {
 if(Nexcl) Pp" %d:???", Nexcl);
} else {
 if(Nexcl) {
	register int k;
	Pp" %d:%d", Nexcl, equExcl[0]);
	for(k=1;k<Nexcl;k++) Pp",%d", equExcl[k]);
 }
}
Pp"\n");
if(EXC) Pp"%s\n\n", cDiffMfExcl_Descr);
Pp"%sinit_from_file =", cmntS);
if(init_from_file) Pp" %s", init_from_file);
Pp"\n");
if(EXC) Pp"%s\n\n", init_from_file_Descr);
}

#undef IN_PAR


void read_cDifMxP(int words, char *is, char *str2, int isCmntC, int isEqualS, int *wrong)
/*=====================================================================================*/
{
int i;
register int j;
double tP, cP;
char *end1, *end2;

if(words && (i=atoi(is+7))>0 && ((tP= strtod(str2,&end1)), *end1==',' && ++end1)
		&& (cP= strtod(end1,&end2)) && end2!=end1 && *end2=='\0') {
	if(!iP) {
		iP=i+2;
		if(iP<cDifMxMode) iP=cDifMxMode;
		tcDifMxP=dvector(iP);
		cDifMxP=dvector(iP);
		for(j=1;j<iP-1;j++) cDifMxP[j]=-1.;
	} else if(i>=(j=iP-1)) {
		iP=i+2;
		if(iP<cDifMxMode) iP=cDifMxMode;
		tcDifMxP=rdvector(tcDifMxP,iP);
		cDifMxP=rdvector(cDifMxP,iP);
		for(;j<iP-1;j++) cDifMxP[j]=-1.;
	}
	tcDifMxP[i]=tP;
	cDifMxP[i]=cP;
	return;
}
PRl"  Illegal input line \"");
if(isCmntC) { PRlc"%c ", CmntC); }
PRlc"%s", is);
if(isEqualS) { PRlc" ="); }
if(words) { PRlc" %s", str2); }
PRlc"\" ignored.\n");
PE"%s", msg);
if(!stat_out) open_statOut(); Ps"%s", msg);
*wrong= 1;
}


static FILE *fp_in;


void bad_input(int Xpln)
/*====================*/
// Xpln is -1 or -2
{
if((par_out= fopen(strcat(strcpy(str1,rootName),".par"),"w")) == NULL) par_out= stderr;
file_header(str1,par_out);
pr_par(Xpln);
Ps"\n");
exitTR(-2,"Correct the illegal input value(s) and rerun!");
}


#define BAD_INPUT_MSG { wrong= 1; PE"%s\n", msg); Ps"%s\n", msg); }


int read_input(int argc, char **argv)
/*=================================*/
{
int StdInp;
static char RESTofL[MAX1_STR_LEN];
int wrong=0;

defCmntC=CmntC;

#if defined(_PC_) || defined(__CYGWIN__) || defined(__CYGWIN32__)
progName = strrchr(argv[0],'.');
// To hide the terminating ".EXE":
if(progName && (progName[1]=='e'||progName[1]=='E')) *progName = 0;
#endif

#ifdef _PC_
progName = strrchr(argv[0],'\\');
if(progName == NULL) progName = argv[0]; else progName++;
#elif defined(_VMS_)
if((progName = strrchr(argv[0],';'))) *progName = 0;
		/* To hide the VMS version numbers */
progName = strrchr(argv[0],']');
if(progName == NULL) progName = argv[0]; else progName++;
#else
progName = strrchr(argv[0],'/');
if(progName == NULL) progName = argv[0]; else progName++;
#endif

#ifdef _PCorCYGold_
 par_out = stdout;
#else
 par_out = stderr;
#endif

if(argc > 1 && argv[1][0]=='-') {
 switch(argv[1][1]) {
	default:  info(); break;
	case 'd': pr_par(0); break;
	case 'p': pr_par(1); break;
	case 'c': pr_par(2); break;
 }
 _Exit(0);
}

if(argc==1) {
	get_defRoot();
	argc = 2;
	argv[1] = defRoot;
}

rootName = argv[1];
StdInp= strlen(rootName)+6;
if(StdInp>MAX1_STR_LEN) {
	PRl"Output root name, %s, is too long. Either shorten it, or increase MAX1_STR_LEN to at least %d",
		rootName, StdInp);
	exitTRsimple(-5);
}

if(argc < 3) fp_in= stdin;
else {
	if((fp_in= fopen(argv[2],"r")) == NULL) {
		register int i;
		for(i=0;i<70+strlen(argv[2]);i++) fputc('!',stderr);
		fputc('\n',stderr);
		PE"!!!! Couldn't open input file '%s', switching to interactive input: !!!!\n\n", argv[2]);
		fp_in= stdin;
	}
}
StdInp= fp_in == stdin;

if(StdInp) {
	interactInfo();
	PE"\n");
} else open_statOut();

{
 register int s;
 register char *c;
 int words = 0, isCmntC = 0, isEqualS = 0;

 if(StdInp) PE"%c ",CmntC);
 c = str1;
 while((s = fgetc(fp_in)) != EOF) {
   if(c && s != '\n') {
     if((isspace(s) && words<3) || (s == CmntC && words == 0 && (isCmntC=1))
		   || (s == '=' && words == 1 && (isEqualS=1))) {
	if(words == 1) { *c = '\0'; c = str2; }
	else if(words == 2) { *c = '\0'; c = RESTofL; }
     } else {
	if(c == str1 || c == str2 || c == RESTofL) words++;
	*c++ = s; /* Here the first 2 words of an input line are
			copied into str1 and str2, and the rest into RESTofL */
	if(words==2 && c==str2+MAX1_STR_LEN) {
		str2[MAX1_STR_LEN-1] = '\0';
		while((s = fgetc(fp_in)) != EOF && s != '\n') ;
		PRl"  Input value \"%s...\" for \"%s\" is longer than %d chars.\n", str2, str1, MAX1_STR_LEN-1);
		PE"%s", msg);
		if(!stat_out) {
			memcpy(RESTofL,str2,MAX1_STR_LEN);
			open_statOut();
			memcpy(str2,RESTofL,MAX1_STR_LEN);
		}
		Ps"%s", msg);
		wrong= 1;
		//goto End;
	}
	if(words==3 && c==RESTofL+MAX1_STR_LEN) {
		*--c = '\0';
		*--c = '.';
		*--c = '.';
		c = NULL; //to ignore the end of line that does not fit into RESTofL
	}
     }
   }
   if(s == '\n') {
     if(c!=NULL) *c = '\0';
     if(words) {
	words--;
	if(!strcmp(str1,"end")) break;
	if(str1[0]=='?') {
	 if(StdInp) {
		if(!strncmp(str1,"???",3)) { info(); _Exit(0); }
		else if(!strncmp(str1,"?c-",3)) pr_par(3);
		else if(!strncmp(str1,"??c-",4)) pr_par(2);
		else if(!strncmp(str1,"?m-",3)) pr_model_help(4);
		else if(!strncmp(str1,"??m-",4)) pr_model_help(5);
		else if(!strcmp(str1,"?")) interactInfo();
		else {
			int explain= str1[1]=='?';
			pr_some_CTRLpar(&str1[1+explain],explain);
			pr_some_PDMpar(&str1[1+explain],explain);
		}
	 }
	 goto End;
	}

#ifndef ultrix
# define IN_PAR(name,type,iV,oFrmt,Expl)  else if(!strcmp(str1,#name))\
	{ if(!words) name = 0; else if(read_inp_value(#name,&name,#type,str2)) wrong=1; }
#else
# define IN_PAR(name,type,iV,oFrmt,Expl)  else if(!strcmp(str1,"name"))\
        { if(!words) name = 0; else if(read_inp_value("name",&name,"type",str2)) wrong=1; }
#endif
#include "CTRLpar.h"
#undef IN_PAR

	else if(!strcmp(str1,"cDifMxEnd")) {
	  if(!words) cDifMxEnd = 0.;
	  else if(read_inp_value("cDifMxEnd",&cDifMxEnd,"double",str2)) wrong=1;
	}
	else if(!strcmp(str1,"cDiffMfExcl")) {
	  if(cDiffMfExcl_dyn) { free_str(1,&cDiffMfExcl); cDiffMfExcl_dyn=0; }
	  if(words) {
		cDiffMfExcl= makestr((int)strlen(str2));
		cDiffMfExcl_dyn= 1;
		memcpy(cDiffMfExcl,str2,strlen(str2));
	  }
	}
	else if(!strcmp(str1,"init_from_file")) {
	  if(init_from_file_dyn) { free_str(1,&init_from_file); init_from_file_dyn=0; }
	  if(words) {
		init_from_file= makestr((int)strlen(str2));
		init_from_file_dyn= 1;
		memcpy(init_from_file,str2,strlen(str2));
	  }
	}
	else if(!strncmp(str1,"cDifMxP",7)) read_cDifMxP(words, str1, str2, isCmntC, isEqualS, &wrong);
	else if(
#if PDMtrace
		(PE"... {PDM header file} - input of \"%s\"\n", str1),0),
#endif
		INPT_model_param(words, str1, str2, &wrong)) {
	} else {
	 size_t len= strlen(str1);
	 if(len<2) len=2;
	 if(!strncmp(str1,"quit",len) || !strncmp(str1,"bye",len)) {
		if(stat_out) {
			CLOSE(stat_out)
			unlink(strcat(strcpy(str2,rootName),".stat"));
		}
		PRl"(2) Interrupted by user");
		exitTRsimple(102);
		/* Same exit status as for the SIGINT signal */
	 }
	 PE"  Unknown input parameter \"%s\" : input line \"", str1);
	 if(isCmntC) PE"%c ", CmntC);
	 PE"%s", str1);
	 if(isEqualS) PE" %c", '=');
	 if(words) PE" %s", str2);
	 if(words==2) PE" %s", RESTofL);
	 PE"\" ignored.\n");
	 FE
	}
     }
     def_par = 0; /* Some of the parameters may no longer have
		default values; used in interactInfo() and pr_CTRLpar() */
  End:
     words = 0;
     isCmntC = 0;
     isEqualS = 0;
     c = str1;
     if(StdInp) PE"%c ",CmntC);
   }
 }
}
CLOSE(fp_in)

PE"\n");
if(StdInp) {
	if(!stat_out) open_statOut();
}

/* Check some obvious bounds: */
#ifndef ultrix
# define CHECK_PAR(cond) if(!(cond)) {PE"  Required: "#cond"\n");\
	Ps"  Required: "#cond"\n"); wrong=1;}
#else
# define CHECK_PAR(cond) if(!(cond)) {PE"  Required: cond\n");\
	Ps"  Required: cond\n"); wrong=1;}
#endif

CHECK_PAR(steady>=0&&steady<=2)
CHECK_PAR(monitor>=-1&&monitor<40)

CHECK_PAR(TOL>1e-300)
CHECK_PAR(TOLC>0.)

if(!(steady&1)) {
	CHECK_PAR(tStepInit>0.)
	CHECK_PAR((tOutInit-tStart)*(tEnd-tStart)>0.)
	CHECK_PAR((tEnd-tOutInit)*(tEnd-tStart)>=0.)
	CHECK_PAR(maxScale>1.)
	CHECK_PAR(tOutMulF>0.)
	CHECK_PAR(Tau>=0.&&Tau<=1.)
	CHECK_PAR(NIG>=-0.5&&NIG<=1.)
	CHECK_PAR(nOutOTH!=0)
	CHECK_PAR(nIterM>2)
	CHECK_PAR(strict>=-1&&strict<=1)
	CHECK_PAR(sepProf>=-1&&sepProf<=1)
	if(strict>-1) {
		CHECK_PAR(cDifMxMode==0 || cDifMxMode==1 ||
			(cDifMxMode==2 && tEnd*tStart>=0.) || cDifMxMode>2)
		CHECK_PAR(cDifMxInit>0.)
		CHECK_PAR(cDifMxMode==0 || cDifMxEnd>0.)
	}
	// attemptMAX and attemptMaxSTEADY are defined in defs.h or possibly in a makefile
	//CHECK_PAR(attemptMAX>1 || attemptMAX<-1)
	CHECK_PAR(attemptMAX!=1)
		// for attemptMAX=-1 or 0, run continues after the 1st unsuccessful attempt
}
if(steady) {
	CHECK_PAR(attemptMaxSTEADY>9)
}
#undef CHECK_PAR

#if PDMtrace
PE"... {PDM header file} - CHECK_PAR\n");
#endif
if(WRONG_model_param()) wrong= 1;

if(steady!=1) {
	tSign = tEnd > tStart ? 1. : -1.;
	ttEnd = tSign * (tEnd - tStart);
}

if(!(steady&1) && strict>-1) {
	switch(cDifMxMode) {
		case 0: cDifMxf= cDifMxf0;
			break;
		case 1: cDifMxf= cDifMxf1;
			cDifMxB= (cDifMxEnd-cDifMxInit)/(ttEnd-tStepInit);
			cDifMxA= cDifMxInit - cDifMxB*tStepInit;
			break;
		case 2: cDifMxf= cDifMxf2;
			cDifMxB= log(cDifMxEnd/cDifMxInit) / log(ttEnd/tStepInit);
			cDifMxA= cDifMxInit / pow(tStepInit,cDifMxB);
			break;
		default:
			if(cDifMxMode>iP) {
				PRl"  Input for cDifMxMode=%d is not complete.", cDifMxMode);
				BAD_INPUT_MSG
			} else {
				register int j;
				cDifMxf= cDifMxf3;
				tcDifMxP[0] = tStepInit;
				cDifMxP[0] = cDifMxInit;
				for(j=1; j<cDifMxMode-1; j++) {
					if(cDifMxP[j]==-1.) {
						PRl"  Input for cDifMxP%d is missing.", j);
						BAD_INPUT_MSG
					} else {
						tcDifMxP[j] -= tStart;
						if(tEnd<tStart) tcDifMxP[j] = - tcDifMxP[j];
						tcDPconvJ = j;
					}
					if(j==1) {
						if(cDifMxP[1]>0. && tcDifMxP[0]>=tcDifMxP[1]) {
							PRl"  tStart%ctStepInit and the time for cDifMxP1 are out of order.",
								tEnd>tStart?'+':'-');
							BAD_INPUT_MSG
						}
					} else {
						if(tcDifMxP[j-1]>0. && cDifMxP[j]>0. && tcDifMxP[j-1]>=tcDifMxP[j]) {
							PRl"  Times for cDifMxP%d and cDifMxP%d are out of order.", j-1, j);
							BAD_INPUT_MSG
						}
					}
				}
				if(tcDifMxP[j-1]>0. && tcDifMxP[j-1]>=ttEnd) {
					PRl"  Time for cDifMxP%d and tEnd are out of order.", j-1);
					BAD_INPUT_MSG
				}
				//These will be executed only in pr_par after all printing is done:
				//tcDifMxP[j] = ttEnd;
				//cDifMxP[j] = cDifMxEnd;
			}
	}
}
return wrong;
}


char cmntS[3]="  ";

#define pTITLE	while(*ti != 0) {Pp"%c", *ti);if(*ti++ == 10) Pp"%c ", CmntC);}

void pr_par(int explain)
/*====================*/
/* explain = -1  ...  print to the 'Output'.par file
	      0  ...  list only the current values of all parameters
	      1  ...  give also all available help on the role of each parameter
	      2  ...  list only control parameter values and description of their role
	      3  ...  list only the current values of the control parameters
	     -2  ...  the same as -1, used only from bad_input() when cDiffMfExcl has illegal format
*/
{
char *ti = title;
register int j;

if(explain >= 0) cmntS[0]='\0'; else cmntS[0]=CmntC;

Pp"%s\n", cmntS);
if(explain<2) {
 if(explain < 0) {
  Pp"%c ", CmntC);
  pTITLE
  ti= versn;
  pTITLE
 }
 Pp"Input parameters");
 if(explain==1) Pp",");
 else if(def_par) Pp" and");
 if(def_par) Pp" their default values");
 if(explain==1) Pp" and their description");
 Pp":\n%s================", cmntS);
 if(def_par) Pp"=========================");
 if(explain==1) Pp"===================");
 Pp"\n%s\n", cmntS);
}

pr_CTRLpar(explain,cmntS);

if(explain<2) pr_model(explain,cmntS);

if(explain<0 && !(steady&1) && strict>-1) {
 char *TT;
 if(cDifMxMode) {
	if(tStart != 0.) {
		if(cDifMxMode<3) TT= tEnd>tStart ? "(t-tStart)" : "(tStart-t)";
		else TT= tEnd>tStart ? "t-tStart" : "tStart-t";
	} else if(tEnd<tStart) {
		if(cDifMxMode<3) TT= "(-t)"; else TT= "-t";
	} else TT= "t";
 }
 Pp"%s   cDifMxf(t) = ", cmntS);
 switch(cDifMxMode) {
	case 0: Pp"%g for all t", cDifMxInit);
		break;
	case 1: Pp"%g + (%g) * %s", cDifMxA, cDifMxB, TT);
		break;
	case 2: Pp"%g * %s^(%g)", cDifMxA, TT, cDifMxB);
		break;
	default:
		iP=cDifMxMode-1;
		Pp"piece-wise linear function between the\n%s   following points (%s,cDifMx):\n",
					cmntS, TT);
		if(tcDifMxP==NULL) {
			Pp"%s\t... not yet defined ...", cmntS);
			break;
		}
		Pp"%s\t(tStepInit=%s, %.8g)\n", cmntS, prettyPr(tcDifMxP[0],"%.10e"), cDifMxP[0]);
		for(j=1; j<iP; j++) {
			// if input was incomplete, we still get here before the run is aborted:
			if(cDifMxP[j]==-1.) Pp"%s\t(?,?)\n", cmntS);
			else Pp"%s\t(%s, %.8g)\n", cmntS, prettyPr((j<=tcDPconvJ?tcDifMxP[j]:tSign*(tcDifMxP[j]-tStart)),"%.10e"), cDifMxP[j]);
		}
		if(tStart != 0.) TT= tEnd>tStart ? "tEnd-tStart" : "tStart-tEnd";
		else if(tEnd<tStart) TT= "-tEnd";
		else TT= "tEnd";
		Pp"%s\t(%s=%s, %.8g)", cmntS, TT, prettyPr(ttEnd,"%.10e"), cDifMxEnd);
		tcDifMxP[j] = ttEnd;
		cDifMxP[j] = cDifMxEnd;
		iP--;	// still needed in cDifMxf3()
 }
 Pp"\n%s\n", cmntS);
}

}


#define PRf (void)fprintf(file,

void file_header(char *filename, FILE *file)
/*========================================*/
{
#if PDMtrace
PE"TRANSIENT (InOut.c): starting file_header for %s\n", filename);
#endif
if(file==par_out) Pp"%c CmntC = %c	(a control input parameter)\n",
	defCmntC,CmntC);
PRf"%c\n%c File: ", CmntC,CmntC);
if(file == stdout) PRf"%s", sOut);
else if(file == stderr) PRf"%s", sErr);
else PRf"%s", filename);
PRf"     Created: %s\n", DateAndTime());
if(file != par_out)
	PRf"%c Parameter values are in file: %s\n%c\n",
		CmntC, (par_out != stderr ? str1 : sErr), CmntC);
}

#undef PRf
// Without #undef, the following redefinition of PRf would generate a benign warning
// during compilation, unnecessarily alarming a PDM programmer
#define PRf (void)fprintf(out,

void echo_init_profile(void)
/*==========================
called after open_output() */
{
FILE *out1 = out;
// Here str2 can contain either the 'rootName' of length o_n_len for sepProf=0,
// or 'rootName'.1 for sepProf=1 or 'rootName'.tdep for sepProf=-1 :
int lenI = (int)strlen(str2) > o_n_len;
(void)sprintf(str2 + o_n_len, ".0");
if((out= fopen(str2,"w")) == NULL) out= stdout;
file_header(str2,out);
PRf"%c Record of initial profiles", CmntC);
if(init_from_file != NULL) PRf" (after reinitialization from %s)", init_from_file);
PRf"\n");
#if PDMtrace
PE"... pr_profiles for initial profiles\n");
#endif
pr_profiles(0);
CLOSE(out)
out= out1;
if(lenI) str2[o_n_len+1]= '1'; else str2[o_n_len]= 0;
// If sepProf=-1, 'rootName'.tdep is not restored in str2 in the previous line,
// but that doesn't matter, the main thing is that the beginning of the string
// contains 'rootName', which all that is needed below to unlink(...tdep) if steady=1.
}

#undef PRf


void change_file(int num)
/*=====================*/
{
if(num<0) {
 out= fopen("/dev/null","w");
 return;
}
if(num>1) CLOSE(out)
if(num) (void)sprintf(str2 + o_n_len, ".%0d", num);
else str2[o_n_len]=0;
if((out= fopen(str2,"w")) == NULL) out= stdout;
file_header(str2,out);
}


void startClock(void)
/*=================*/
{
(void)snprintf(jobStartTime,DATELEN,"%s",DateAndTime());
#ifdef _ONE_CLOCK_
/* Even if long int (clock_t) is only 4 bytes long, and there are 1000 ticks
   per second, CPU time differences up to 49.7 days should be correct */
clk_tck= (double)CLK_TCK;
cl0= clock();
#else
clk_tck= (double)(sysconf(_SC_CLK_TCK));
 /* If you don't have <unistd.h>, where _SC_CLK_TCK is defined, look
    for the definition of CLK_TCK in <time.h>. If no such definition
    could be find, in Unix try clk_tck = 100 tics per second.
 */
(void)times(&cpuT);
us0= cpuT.tms_utime;
sy0= cpuT.tms_stime;
#endif
//PE"clk_tck=%g\n", clk_tck); exitTR(777,"test");
}


void startTransClock(void)
/*======================*/
{
#ifdef _ONE_CLOCK_
clTr0= clLast= clock();
#else
 (void)times(&cpuT);
usTr0= usLast= cpuT.tms_utime;
syTr0= syLast= cpuT.tms_stime;
#endif
}


#define Pd (void)fprintf(diag_out,
#define LINE {register int i; for(i=0;i<64+o_n_len;i++) Ps"%c",'-'); Ps"\n");}


void diagHdrLines(void)
/*===================*/
{
 Pd"%c Minima and maxima between the time instants of the 1st column\n", CmntC);
 Pd"%c (starting from tStart=%g); usually between pr_Other times.\n", CmntC, tStart);
 // On the next line, cDiffMfExcl is still set when open_output() is called, and Nexcl
 // is set up correctly when steadyDiag() is called (when steady=2):
 if(cDiffMfExcl || Nexcl) {
	Pd"%c                     tStep (sec)       cDiffMexcl cDiffMused  allIter \
  # t steps\n", CmntC);
	Pd"%c   t (sec)         min        max         max        max      min  max\
  sampled\n", CmntC);
 } else {
	Pd"%c                     tStep (sec)              cDiffM          allIter \
  # t steps\n", CmntC);
	Pd"%c   t (sec)         min        max         min        max      min  max\
  sampled\n", CmntC);
 }
}


void open_statOut(void)
/*===================*/
{
if((stat_out= fopen(strcat(strcpy(str2,rootName),".stat"),"w")) == NULL)
					 stat_out= stderr;
//PE"Opening stat_out: str2=%s, (stat_out==stderr)=%d\n", str2, stat_out==stderr);
}


extern int lufacCompiled(void);
extern void nl3bandCompiled(char), transienCompiled(void);

void open_output(void)
/*==================*/
{
#if PDMtrace
PE"TRANSIENT (InOut.c): starting open_output for output root name: %s\n", rootName);
#endif

// Input parameters echo (par file):
if((par_out= fopen(strcat(strcpy(str1,rootName),".par"),"w")) == NULL)
					 par_out= stderr;
file_header(str1,par_out);
if(lufacCompiled()<0) Pp"%c Compilation:\n", CmntC); else
 Pp"%c lufacv.c compiled with: TRUE_FABS=%d\n", CmntC, lufacCompiled());
transienCompiled();
nl3bandCompiled(CmntC);

#ifdef _VMS_
/* The VMS specific function fgetname will replace str1 with the actual
   full VMS filename, including the version number, and the preceding
   dot '.' added by VMS to names without extensions: */
(void)fgetname(par_out,str1);
#endif

/* Now str1 contains the name of the .par file, to be used in file_header()
to print the name of the .par in the header of all other output files. */

// Statistics:
// Already opened in open_statOut(); remains to write out the header:
file_header(strcat(strcpy(str2,rootName),".stat"),stat_out);
// After this, rootName remains in the beginning of str2

o_n_len = (int)strlen(rootName);

// Diagnostics:
if(sepProf>=0) {
 (void)sprintf(str2 + o_n_len, ".diag");
 if((diag_out= fopen(str2,"w")) == NULL) diag_out= stderr;
 file_header(str2,diag_out);
 if(steady) Pd"%csteady = %d calculation\n%c~~~~~~~~~~~~~~~~~~~~~~\n%c\n",
		CmntC, steady, CmntC, CmntC);
 else diagHdrLines();
 (void)fflush(diag_out);
} else {
 diag_out= fopen("/dev/null","w");
}

// Time dependencies:
(void)sprintf(str2 + o_n_len, ".tdep");
if((timeDep_out= fopen(str2,"w")) == NULL) timeDep_out= stderr;
file_header(str2,timeDep_out);
tDepHeadEndPos= ftell(timeDep_out);

// Profiles (out) file:
if(sepProf>0) sepProf = 1;
change_file(sepProf);

if(steady) Ps"\nsteady = %d calculation\n\n", steady);
if(steady != 1) {
 char *whatTi =
#ifdef MSDOS
	" elapsed";
#else
	"CPU_time";
#endif
 { register int i= 6-(int)strlen(time_units),
		offset= sepProf>-1?o_n_len+10:13;
   Ps"   simulated  %*s_(sec)   time  runband       NR\n",offset+6,whatTi);
   // Maximum i=5 when  time_units  consists of a single character:
   if(i<1) {
	Ps"time_(%s) ", time_units);
	if(!i) Ps" %-*s", offset, "file");
	else {
		i= offset+1+i;
		if(i > 4) Ps"%-*s", i, "file");
		else if(i > 0) Ps"%*s", i, " ");
	}
   } else if(i<5) Ps"%*stime_(%s)  %-*s", i, " ", time_units, offset, "file");
   else Ps"    time_(%s)   %-*s", time_units, offset, "file"); // i=5
 }
#ifdef _ONE_CLOCK_
 Ps"           ");
#else
 Ps"user   syst");
#endif
 Ps"   steps   calls  iterations\n");
 LINE
 Ps"%#13g\n", tStart);
}
(void)fflush(stat_out);
}


long int CPUtime(double t, long stepCounter, long runbandCounter, long iterCounter)
/*===============================================================================*/
{
 long int NOtSteps = stepCounter-stCoLast;

 if(sepProf==-1) Ps"%#13g  n/a     ", t);
 else Ps"%#13g  %s%*s", t, str2, 5+o_n_len-(int)strlen(str2), " ");
#ifdef _ONE_CLOCK_
 clTo= clock();
 Ps"%9.2f       ", (double)(clTo-clLast)/clk_tck);
 clLast = clTo;
#else
 (void)times(&cpuT);
 usTo= cpuT.tms_utime;
 syTo= cpuT.tms_stime;
 Ps"%9.2f %6.2f", (double)(usTo-usLast)/clk_tck, (double)(syTo-syLast)/clk_tck);
 usLast = usTo;
 syLast = syTo;
#endif
 Ps" %7ld %7ld %8ld\n",
	NOtSteps, runbandCounter-rbCoLast, iterCounter-itCoLast);
 (void)fflush(stat_out);
 stCoLast= stepCounter;
 rbCoLast= runbandCounter;
 itCoLast= iterCounter;
 return NOtSteps;
}

void steadyDiags(int attempt, int iter)
/*===================================*/
{
 Pd"%cSteady state: %d attempt(s); %d iterations in the last attempt\n",
	CmntC, attempt, iter);
 Pd"%cIn last iteration:\n", CmntC);
 report_last_NR_error(diag_out, CmntC);
 if(steady > 1) {
	Pd"%c\n%cTransient calculation starting from the steady state:\n",
		CmntC, CmntC);
	diagHdrLines();
 }
}


void close_tDepOut(void)
/*====================*/
{
 if(timeDep_out) {
	if(tDepHeadEndPos >= ftell(timeDep_out)) {
		CLOSE(timeDep_out)
		(void)sprintf(str2 + o_n_len, ".tdep");
		if(unlink(str2) == 0) PE"\nFile %s deleted as nothing was written into it\n", str2);
	} else CLOSE(timeDep_out)
 }
}


void exitTR(int statu, char *descr)
/*===============================*/
{
static int count=0;

PE"\n %d: %s\n", statu, descr);
#if PDMtrace
PE"!!! exitTR invoked");
if(count) PE" for the 2nd time; exiting immediately\n\n");
else PE"; the following are called from exitTR:\n\n");
#else
if(count) PE"\t2nd invocation of exitTR()!!!\n");
#endif
count++;
if(count>1) _Exit(statu);

// do_Oth_Prof can be nonzero for a transient calculation only.
// It is checked and reset in these two functions:
	out_Other(0);
	out_profiles(0);
if(par_out) CLOSE(par_out)
close_tDepOut();
if(diag_out) {
	if(steady != 1) DiagEntry_at_t();
	Pd"\n");
	if(strict1failCount)
	 Pd"%c strict=1 failed to reduce cDiffM to the required value \
altogether %ld times\n", CmntC, strict1failCount);
	CLOSE(diag_out)
}

if(stat_out) {
	if(steady != 1 && statu != -2) {
	 LINE
	 Ps"%*s", 18+o_n_len, "Totals");
#ifdef _ONE_CLOCK_
	 Ps"%11.2f       ", (double)(clTo-clTr0)/clk_tck);
#else
	 Ps"%11.2f %6.2f", (double)(usTo-usTr0)/clk_tck, (double)(syTo-syTr0)/clk_tck);
#endif
	 Ps" %7ld %7ld %8ld\n\n", stCoLast, rbCoLast, itCoLast);
	} else Ps"\n");

	if(statu != -2) {
	 register double sec;
	 register int hm;
	 char *whatTi =
#ifdef MSDOS
		"elapsed";
#else
		"CPU";
#endif
	 Ps"Whole job %s time", whatTi);
#ifdef _ONE_CLOCK_
	 clTo= clock();
	 sec= (double)(clTo-cl0)/clk_tck;
#else
	 (void)times(&cpuT);
	 usTo= cpuT.tms_utime;
	 syTo= cpuT.tms_stime;
	 sec= (double)(usTo+syTo-us0-sy0)/clk_tck;
	 Ps" (user+system)");
#endif
	 Ps": %g sec", sec);
	 if(sec>=60.) {
		Ps" = ");
	 	if(sec>=3600.) {
			hm= (int)floor(sec/3600.);
			sec -= hm * 3600;
			Ps"%d h", hm);
		}
		if(sec>=60.) {
			hm= (int)floor(sec/60.);
			sec -= hm * 60;
			Ps" %d min", hm);
		}
		if(sec>0.) {
			Ps" %g sec", sec);
		}
	 }
	 Ps"\n\n");
	}
	Ps"Job started on %s,\n", jobStartTime);
	Ps" terminated on %s.\n", DateAndTime());
	Ps"\n%d: %s\n", statu, descr);
	if(statu && do_Oth_Prof & 8u && steady != 1)
		Ps"  Terminated when trying to advance the solution to t = %le %s\n",
			 t, time_units);
	(void)fflush(stat_out);
	CLOSE(stat_out)
}

if(out) CLOSE(out)

endband();
_Exit(statu);

}


void exitTRsimple(int statu)
/*========================*/
// The purpose of this function is to preserve error messages
// generated during parameter input, e.g., in case of batch processing
{
 PE"\n%d: %s\n", statu, msg);
 FE
 close_tDepOut();
 if(stat_out) {
	Ps"\n%d: %s\n", statu, msg);
	(void)fflush(stat_out);
	CLOSE(stat_out)
 }
 _Exit(statu);
}
