/* File defs.h; TRANSIENT V4.3; 30 July 2017 */
/* (C) 1999-2005,2011,2015,2017 Miroslav Kolar   */

/* Header file for several macro definitions to control the inclusion
   of various features into TRANSIENT during compilation.
Included in InOut.c, nl3bandv.c, nl3band1.c, transien.c and lufacv.c.

One can change the defaults set in this file by editing the
#define directives below. Note that they are sandwiched between
the #ifndef and #endif directives in order that the (g)cc command
line preprocessor -D flags (e.g., -DNO_X=1) can take
precedence over the macro definitions in this file.

   Macros to be set (or discussed here):
PDMtrace, NO_TIME_ZONE, _ONE_CLOCK_, TRUE_FABS,
preV4_3, RESETtStepAT99,
NO_X, NO_Y, NR_BOUNDS, NodeNumberChangeENABL,
DEF_OUTPUT_ROOT, attemptMAX, attemptMaxSTEADY,
PATHtoITSELF, and PRINT_EXIT_CODES_REMINDER.
*/ 



/* To include/not include most of the following features,
 * the corresponding macro name must be set to 1/0:	*/


/* - Set PDMtrace to 1 when debugging a PDM: to print the name of each
   PDM function before it is invoked, and to report passage through
   a few other points in TRANSIENT code (this used to be the
   functionality of monitor=3 before version V1.4):		*/
#ifndef PDMtrace
#define PDMtrace	0
#endif



/* - If you don't like the timezone field in the time stamps in the
   TRANSIENT generated output files, enable (uncomment) the line
#define NO_TIME_ZONE
at the beginning of the DateTime.c file.
Or add -DNO_TIME_ZONE (UNIX) or /DNO_TIME_ZONE (MS) or
/define=NO_TIME_ZONE (VMS) among the compiler flags in the respective makefiles.
*/



/* - If you don't have  times()  function returning separately user and system
   CPU times (an error message about times being undeclared/unreferenced
   appears), or you don't like the separate reporting of the two CPU times
   (see the the .stat output files), enable (uncomment) the following line:
#define _ONE_CLOCK_
   (in any case, _ONE_CLOCK_ is automatically defined in  InOut.c  for native
   Windows/DOS compilers).
*/

/* Remember that in VMS, to get the last two features (timezones and user and
   system CPU time) working, one has to use the following CC flags:
   /DECC/PREFIX_LIBRARY_ENTRIES=ALL_ENTRIES (see the MAKEDEMOS.COM file).
*/



/* - Optimize speed of matrix inversion:
 ***	This was much more important in 1995 than in the era of 1+ GHz CPU's:
	It was devised for optimizing  lufacv.c  speed:
 On most (all?, at least on DEC Ultrix and OSF, Sun Sparc, and on SGI)
 UNIX machines, multiplication, a*a, is much faster (5 to 40 times)
 than fabs(a). In lufacv.c, one calculates the maximum of a very large number
 of fabs(a). Replacing it by the maximum of a*a (followed by a single sqrt)
 would for very small  a  mean that the 'machine zero' is aproached sooner.
 A suitable compromise is the (if(a<0) a = -a;) construct which is comparable
 to (a*a) in the CPU time use (or at most 4 times slower on the DEC machines).
 On the other hand, under DOS/Windows, fabs(a) is faster than a*a
 or the above if construct.
 One can switch (in lufac.c and transien.c) between these two possibilities
 by explicitly defining TRUE_FABS macro:
 #define TRUE_FABS 1
	will calculate absolute value using the fabs() function;
 #define TRUE_FABS 0
	will use the above if construct instead.
 If TRUE_FABS is not explicitly defined, the following automatic definition
 will be used:
*/
#ifndef TRUE_FABS
#if defined(MSDOS) || defined(_Windows) || defined(__CYGWIN__) || defined(__CYGWIN32__)  || defined(__MINGW32__) || defined(__MINGW64__)
#define TRUE_FABS 1
#else
#define TRUE_FABS 0
#endif
#endif

// Finally, here is an optimized abs value function (macro) ABSV(x)
// that can be used in a PDM instead of fabs(x):
#if TRUE_FABS
#define ABSV(x) fabs(x)
#else
/*@unused@*/
static double ABSV() __attribute__ ((always_inline, const, unused));
static inline double ABSV(double x) {return (x<0 ? -x:x);}
#endif



/* - Linking with a PDM written for the older TRANSIENT versions (pre V4.3):
 * Try compiling TRANSIENT with preV4_3 macro set to 1. If the PDM is not
 * too old, you may not even need to do any changes to the PDM. Prior to
 * version V4.3, c was extrapolated to the new time step, and c moved to c_o,
 * before the call to the PDM's update_Other(). Since version V4.3, it is done
 * after the call to update_Other() because tStep is allowed to be changed in
 * update_Other(). Prior to V4.3, update_Other() had to use c_o[]; since V4.3,
 * it must use new c[], not yet moved to c_o[]. Macro preV4_3 takes care of this:
*/
#ifndef preV4_3
#define preV4_3 0
#endif

#if preV4_3
/* - Enable resetting ttStep if update_Other() returns an integer larger than 1.
 * This macro was intended for special uses, and is now deprecated. One can
 * still use it only with preV4_3 set to 1.
 * It is now recommended that instead of using RESETtStepAT99, the PDM programmer
 * resets tStep directly in  update_Other().
 * Default: DISABLED
*/
#ifndef RESETtStepAT99
#define RESETtStepAT99 0
#endif
#endif



/* For the following two macros, see Eq. (15) of the TRANSIENT manual */

/* - One should set NO_X to 1 if all elements of the JACOBIAN SUBMATRIX X(0)
 *   are always equal to 0:
*/
#ifndef NO_X
#define NO_X 0
#endif



/* - One should set NO_Y to 1 if all elements of the JACOBIAN SUBMATRIX Y(N-1)
 *   are always equal to 0:
*/
#ifndef NO_Y
#define NO_Y 0
#endif



/* - As a default, the BOUNDS for "concentrations" c are applied in nl3band?.c
 * only at the end of a time step, after the NR iteration process had converged.
 * If we can live with that c's can get out of bounds during NR iterations,
 * this usually gives much faster execution (NR convergence) than applying the bounds
 * after each NR iteration.
 * To apply bounds after each NR iteration, set NR_BOUNDS to 1:
*/
#ifndef NR_BOUNDS
#define NR_BOUNDS 0
#endif


/* - If you need to change (vary) the NUMBER OF NODES during a (transient) run,
 * set NodeNumberChangeENABL to 1 (# of nodes cannot exceed the value first
 * specified at the beginning of the run):
*/
#ifndef NodeNumberChangeENABL
#define NodeNumberChangeENABL 0
#endif



/* - Default name for the ROOT of all OUTPUT FILES when none is given on the command line:
 * 1. If DEF_OUTPUT_ROOT below or in a makefile is defined, the root name is equal to
 *	DEF_OUTPUT_ROOT (do not forgot the double quotes; in a makefile, one has to
 *	escape them, e.g., -DDEF_OUTPUT_ROOT=\"MD1\" );
 * 2. Otherwise it is equal to the capitalized name of the invoked executable file.
*/
//#define DEF_OUTPUT_ROOT "TR"



/* - Max # of attempts: Transient run:
 * When no NR convergence is achieved in a time step, the time step is reduced as a rule,
 * and the current step repeated. The following macro controls the maximum number of such
 * attempts. Two scenarios are possible:
 *  If attemptMAX is positive (the usual default), the run is terminated when the number of
 * tStep reduction attempts would exceed attemptMAX.
 *  If attemptMAX is negative, the maximum number of attempts is -attemptMAX, tStep cannot
 * be reduced below tStepInit, and the run is continued (advanced to the next time step)
 * after -attemptMAX attempts (or earlier if tStepInit is reached) in the hope that the situation
 * will correct itself. Negative attemptMAX can be useful in conjunction with resetting the
 * time step in update_Other() (see above RESETtStepAT99 for the pre V4.3 TRANSIENTs),
 * when trying to simulate a sequence of modified system (such as a growing boundary).
*/
#ifndef attemptMAX
#define attemptMAX -25
//#define attemptMAX 15
#endif



/* - Max # of attempts: Steady-state run:
 * Maximum number of attempts to find a steady-state solution. The initial solution
 * estimate is randomized before each new attempt.
*/
#ifndef attemptMaxSTEADY
#define attemptMaxSTEADY 100
#endif



/* - Path to the binary:
 * Enable the following macro, if you need in the PDM the function pathToThisExecutable()
 * that returns a pointer to a buffer containing the full path to the current executable:
*/
#ifndef PATHtoITSELF
#define PATHtoITSELF 0
#endif



/* - To print at the beginning of each run a reminder about the ranges of termination
 * (exit) codes available to the PDM authors, uncomment the following #define.
 * One of those ranges depends on the number of signals available on your
 * system/compiler. Thus you may want to uncomment this macro temporarily whenever
 * moving to a different platform.
*/ 
//#define PRINT_EXIT_CODES_REMINDER
