/* File demo/demo1/demo1.c; TRANSIENT V4.3; 2 Nov 2023 */
/* (C) 1995-2005,2023 Miroslav Kolar */

/* ====================================================================	*/
/*  Corrosion of a metal surface adjacent to a saturated porous medium	*/
/*  of thickness L1 followed by another porous medium of thickness L2	*/
/*  Non-equidistant spatial grid (geom. progression for grid steps)	*/
/*  Demonstration module No. 1 for  TRANSIENT				*/
/* ====================================================================	*/


#define PDM_H "demo/demo1/demo1.h"
	/* path of demo1.h relative to the TRANSIENT directory */
#include "../../transien.h"
	/*  >>>> READ the information in transien.h of TRANSIENT <<<<	*/

/* Uncomment to enable intermediate output (debugging):
#define DEBUG
*/
#ifdef DEBUG
#include "../../matrixv.h"
#endif

/*
 Average year length in the current Gregorian calendar
 is 365.2425 days = 31556952 sec.
 This is still 26 sec longer than the Mean Astronomical Year
 of 365 days, 5 hr, 48 min, 46 sec = 31556926 sec.
 In one million years, this difference will add up to 300.926 days.
 301 more leap days will have to be dropped over that period from the
 present calendar to remain in sync with the nature.
 Thus the best time unit for extremely long time periods is the
 Mean Astronomical Year:
*/
#define Y2S 31556926.

#define INFTY	1e25
	/* To represent infinite current or potential */
#define ICORREPS	2e-16
	/* Precision of iCorr calculation; should be > 1e-17 */

#define MEDIA	2
#define REGIONS	3       /* REGIONS = 2 * MEDIA - 1 */
#define difRGNS	4       /* difRGNS = 3 * MEDIA - 2 */

#define NEQNS	6	/* all species */
#define DIFFUS	3	/* diffusive species */

#define S0	0
#define S1	1
#define S2	3
#define S3	2
#define S4	4
#define S5	5
/* The code in this file relies on the fact that the
concentrations of all the diffusive species at each node
precede in memory those of the non-diffusive species.
The above definitions can be changed under the condition that
the diffusive species S0, S1 and S3 are assigned numbers 0, 1 and 2
in arbitrary order, and the non-diffusive species S2, S4 and S5
the remaining numbers 3, 4 and 5 again in arbitrary order.
See below the declaration of array  order[]  containing the
order of species for output to the profile files. It is the
same order in which the species occur in the reaction-diffusion
equations in the manual (or in the table in the following comment).
Changing the order of species in memory (withing the above
limitations) can be used to verify the accuracy:
ideally, the results should not change.
When using WEIGHTS, remember that e.g. wei[2] will correspond to
species S3 for the above definitions, etc.
*/

/****************************************************************/
/* C[j][S0] = c_0^s - oxygen pore-water concentration		*/
/* C[j][S1] = c_1^s - Cu(I) pore-water concentration		*/
/* C[j][S2] = c_1^p - Cu(I) precipitate bulk concentration	*/
/* C[j][S3] = c_2^s - Cu(II) pore-water concentration		*/
/* C[j][S4] = c_2^p - Cu(II) precipitate bulk concentration	*/
/* C[j][S5] = rho * s_2 - Cu(II) sorbant concentration		*/
/****************************************************************/

extern void Randomize(double Amplitude, double *c_o, double *c);

typedef double
	ROW[NEQNS],
	difROW[DIFFUS];

/* The PDM specific input parameters are defined in demo1.h: */

/* Other global variables: */
static double
	New, k5New, km5New, k9New, km9New,
	k11New, km11New, k7New, k8New,
	f, f1, fSq,
	*x,
	Econst, iCorr, iCorrInt2, iCorrOld,
	cBarS1, S0const, S1const,
	a2, d2, d20, d20New,
	eps[REGIONS], smax[REGIONS], EPS, SMAX,
	t000,   /* = tStart for the 1st in a series of continued runs */
	tMean, tStepLastBut1;	/* used in pr_Other */
static ROW
	*C, *C_o,
	*A, *B, *D; /* X, Y Jacobian sub-matrices not needed in this PDM */
static double *F;
static difROW
	DD[difRGNS];
static int
	nnodesM1,
	trueSteady,
	order[NEQNS] = {S0,S1,S2,S3,S4,S5},
	trans_to_trans = 0; /* with init_from_file only: is reset to 1
		only if initialized from a file with transient state
		and if currently steady=0 */

// This needs to be improved (turned into a function) to report when output
// is truncated, and thus MAXLETT should be increased:
#define ErrMsg (void)snprintf(text,MAXLETT,


int get_neqns_etc(void)
/*===================*/
{
title = "			TRANSIENT\n\
		    Demonstration No. 1\n\
   Uniform Corrosion and Transport of Corrosion Products";
time_units = "sec";
tEnd = 1.6e16;
tOutInit = 2629800.;
Tau = 0.1;
cDifMxMode= 1;
cDifMxInit= 0.01;
cDifMxEnd= 1e-5;
return NEQNS;
}

void generate_grid(void)
/*====================*/
{
nnodesM1 = interv1 + interv2;
nnodes = nnodesM1 + 1; /* interv1 = 1 + the # of intervals in medium 1;
			  interv2 = the # of intervals in medium 2;
			  nnodes = total number of nodes  */
x= dvector(nnodes);
{ register double fold, a1, a2, a3;
  f = 3.;
  a1 = (double)interv1;
  a2 = 1. / (double)nnodesM1;
  a3 = (L1 + L2) / L1;
  do {  fold = f;
	f = pow(a3 * (pow(fold,a1)-1.) + 1., a2);
     } while(ABSV(fold-f) > 5e-17);
}
{ register int i;
  register double x1fm1;
  x[1] = f;
  for(i=2; i<nnodes; i++)  x[i] = f * x[i-1];
  /* Now x[i] = f^i */
  x[0] = 0.;
  x1fm1 = L1 / (x[interv1]-1.);
  for(i=1; i<nnodes; i++) x[i] = x1fm1 * (x[i] - 1.);
}
f1 = f + 1.;
fSq = f * f;

Pp"%c Grid: initial x step = %g m\n",CmntC,x[1]);
Pp"%c       quotient f = %g\n",CmntC,f);
Pp"%c       rightmost x step in medium 1 = %g m\n",CmntC,x[interv1]-x[interv1-1]);
Pp"%c       rightmost x step = %g m\n",CmntC,x[nnodesM1]-x[nnodes-2]);
}


void Initialize(double *c_vec, double *c_o_vec)
/*===========================================*/
{
int j,k;

/* Following test more practical here than using CHECK_PAR in demo1.h: */
j=strlen(frmt)-1;
if(frmt[j]!='e' && frmt[j]!='g' && frmt[j]!='f' && frmt[j]!='E' && frmt[j]!='G')
		exitTR(16,"Illegal input: frmt  must end with e,f,g,E or G");
for(++j;j>=0;j--) frmt[j+1]=frmt[j];
frmt[0]=' ';

tMean= t000= tStart; /* may be reset if init_from_file */
tStepLastBut1= 0.;

C = (ROW *)c_vec;
C_o = (ROW *)c_o_vec;

			   /*  REGION:      */
eps[0] = eps1A;		   /*  0: medium 1  */
eps[1] = eps2;		   /*  1: medium 2  */
eps[2] = eps1A + f * eps2; /*  2: interface between
			media 1 and 2 (regions 0 and 1)*/

smax[0] = s2max;
smax[1] = 0.;
smax[2] = (eps2 * smax[1] * f + eps1A * smax[0]) / eps[2];

eps[2] /= f1;

 for(j= 0; j<interv1; j++) C_o[j][S0]= C[j][S0]= Cmed1;
 C_o[j][S0]= C[j][S0]=
  (eps1A * Cmed1 + f * eps2 * Cmed2) / (eps1A + f * eps2);
 for(++j; j<nnodes; j++) C_o[j][S0]= C[j][S0]= Cmed2;
 for(j= 0; j<nnodes; j++)
	C_o[j][S1]= C[j][S1]= C_o[j][S2]= C[j][S2]= C_o[j][S3]= C[j][S3]=
	C_o[j][S4]= C[j][S4]= C_o[j][S5]= C[j][S5]= 0.;
/*
if(steady) {
 for(j= 0; j<nnodes; j++)
	for(k=0; k<NEQNS; k++) C[j][k]= 0.;
 C[nnodesM1][S0]= Cmed2;
}
*/
if(steady) Randomize(0.5*(Cmed1+Cmed2), (double *)C_o, (double *)C);

}



#define CLOSI  (void)fclose(infile);
#define MAXLETT 120
#define GETLIN0  p= fgets(text,MAXLETT,infile);
#define GETLINE  GETLIN0 if(p == NULL) break;\
 if(text[strlen(text)-1] != 10) {\
 PE"\nInput file has lines longer then %d characters!\n", MAXLETT-1);\
 ErrMsg"MAXLETT in "__FILE__" must be increased"); exitTR(13,text); }

static char text[MAXLETT];

static FILE *infile;

void read_failed(int status, char *howmuch, char *filename)
/*=======================================================*/
{
 CLOSI
 ErrMsg"%s has %s data than for %d nodes", filename, howmuch, nnodes);
 exitTR(status,text);
}

void read_profiles(char *filename)
/*==============================*/
/* It is assumed that infile was created by pr_profiles */
{
  char *p, comnt;
  int j=0, k;

  if( (infile = fopen(filename,"r")) == NULL) {
	PE"Failed to open  %s\n", filename);
	PE"Initializations made in Initialize() unchanged\n");
	return;
  }

  for(;;) {
	GETLINE
	comnt= *p; /* CmntC may differ between runs */
	break;
  }
  do {
	GETLINE
	if(*p != '\n' && *p != comnt) break;
	if(*(++p) == ' ' && *(p + 1) == 'T') {
	   if(sscanf(p, " Transient results for t = %le sec = %*e years", &t)) {
		PE"Transient-run profiles for t = %.12g sec read in from file %s\n",t,filename);
		if(!steady) {
		   tMean= t000= tStart= t;
		   PE"This time will become the new tStart\n");
		   if(t>=tOutInit || t>=tEnd) {
			ErrMsg"%s is for time t >= tOutInit or tEnd",filename);
			exitTR(14,text);
		   }
		   trans_to_trans=1;
		}
	   }
	} else if(strstr(text,"Steady State") != NULL)
		PE"Steady-state run results (profiles) read in from %s\n",filename);
  } while(1);

  do {
    if(j == nnodes) read_failed(7,"more",filename);
		    // The above failure must always happen for a file created
		    // with sepProf=0 --> files with name not containing a '.' and
		    // a numerical extension need not be considered in init_Other()
    if(sscanf(text, " %*e %le %le %le %le %le %le", &C_o[j][S0], &C_o[j][S1],
		&C_o[j][S2], &C_o[j][S3], &C_o[j][S4], &C_o[j][S5])<6) {
	CLOSI
	PE"%s\n", text);
	Ps"%s\n", text);
	ErrMsg"The above unexpected line encountered in file %s", filename);
	exitTR(9,text);
    }
    j++;
    GETLINE
  } while(1);
  if(j<nnodes) read_failed(8,"less",filename);
  CLOSI

  for(j=0; j<nnodes; j++)
		for(k=0; k<NEQNS; k++) C_o[j][k] *= 1000.;

  if(onlyCuI==2) {
	/* First convert all c2 to c1 */
	for(j=0;j<interv1;j++) C_o[j][S1] += C_o[j][S2]/eps1A;
	j++; C_o[j][S1] += C_o[j][S2]/eps[2];
	for(;j<nnodes;j++) C_o[j][S1] += C_o[j][S2]/eps2;
	for(j=0;j<nnodes;j++) C_o[j][S2]= 0.;
  }

  if(!(steady & 1))
	for(j=0; j<nnodes; j++)
		for(k=0; k<NEQNS; k++) C[j][k]= C_o[j][k];
  PE"Read in: C_o[0][S0] = %.10g, etc.\n", C_o[0][S0]);
}


void auxiliary_parameters_for_evalf(void)
/*=====================================*/
{
New = 1. - Tau;

/*
DD[reg][k] store diffusion coefficients divided by
respective capacity factors (accessible porosities) in
region 'reg':	0,1 - medium 0 and 1;
		2 - left side of the interface;
		3 - right side of    - " -
Inside evalf(), rows of DD for individual regions
are in turns pointed to by DDm.
*/

DD[0][S0] = tau1*eps1E*D0/eps1A;
DD[0][S1] = tau1*eps1E*D1/eps1A;
DD[0][S3] = tau1*eps1E*D3/eps1A;
DD[1][S0] = tau2*D0;
DD[1][S1] = tau2*D1;
DD[1][S3] = tau2*D3;

DD[2][S0] = tau1*eps1E*D0/eps[2];
DD[2][S1] = tau1*eps1E*D1/eps[2];
DD[2][S3] = tau1*eps1E*D3/eps[2];
DD[3][S0] = tau2*eps2*D0/eps[2];
DD[3][S1] = tau2*eps2*D1/eps[2];
DD[3][S3] = tau2*eps2*D3/eps[2];
#ifdef DEBUG
show_dvector("DD",(double *)DD,0,11);
#endif

d20= 2./(x[1] * x[1]);
d20New= d20*New;
d2= d20*f/f1;
a2= d2 * f;

Econst = Rgas * Tabs / Farad;
S0const= -2./(eps1A*x[1]*n0*Farad);
S1const= 2./(eps1A*x[1]*n1*Farad);

if(onlyCuI==2) k5=0.;

k5New = k5 * New;
k7New = k7 * New;
k8New = k8 * New;
k9New = k9 * New;
k11New = k11 * New;
km5New = km5 * New;
km9New = km9 * New;
km11New = km11 * New;

}


void initNL3BANDV(double***LBp, double***UBp, double*Fnl3,
	double*A_vec, double*B_vec, double*D_vec)
/*======================================================*/
{
static double *UB[NEQNS], *LB[NEQNS], Upper[5], Lower[2], sZero[2];
int j;

*LBp = LB;
*UBp = UB;

Lower[0] = 1.;  /* Bound type: constant */
Lower[1] = 0.;  /* Bound value */
if(!onlyCuI) {
 sZero[0] = ZERO==0. ? 1.:4.;  /* Bound type: set zero cutoff (or constant) */
 sZero[1] = ZERO;  /* Zero cutoff value - input parameter */
 Upper[0] = 2.;  /* Bound type: piece-wise constant */
 Upper[1] = s2max; /* First bound value */
 Upper[2] = interv1; /* Last node where Upper[1] applies */
 Upper[3] = 0.;    /* Second bound value */
 Upper[4] = 0.;    /* => Upper[3] applies for all remaining nodes */

 for(j=0; j<NEQNS; j++) LB[j]= Lower; /* All lower bounds are identical */
 LB[S0] = sZero;				/* except for the one for S0 */

 /* First five species have no upper bounds: */
 for(j=0; j<NEQNS; j++) UB[j]= NULL;
 /* Bound for C[][S5] defined in Upper: */
 UB[S5] = Upper;
} else {
 Cmed2=0.;
 for(j=0; j<NEQNS; j++) UB[j]= LB[j]= Lower; /* Force everything to be zero */
 UB[S1]= NULL; /* except for the Cu I concentration(s) */
 if(onlyCuI==1) UB[S2]= NULL; /* else k5 set to 0 above */
}

/* Addresses of F, Jacobian submatrices: */
F= Fnl3;
A = (ROW *)A_vec;
B = (ROW *)B_vec;
D = (ROW *)D_vec;

}


inline double max0(double x)
/*========================*/
{ return (x > 0. ? x : 0.); }


void reaction_part0(double *c)
/*==========================*/
/* Only non-diffusive species, to be used at the last node.
   Values of EPS and SMAX for the current mode are set in evalf().
*/
{
 register double R;

 R = k9 * max0(c[S3] - c3sat);
 F[S4] = EPS * R - km9 * c[S4];
 B[S4][S3] = R == 0. ? 0. : EPS * k9New;
 B[S4][S4] = -km9New;
 if(SMAX != 0) {
    /* 0 <= c[S5] <= SMAX; see the bounds LB, UB in Initialize */
  R = SMAX - c[S5];
  B[S5][S3] = EPS * k11New * R;
  R *= k11 * c[S3];
  F[S5] = EPS * R - km11 * c[S5];
  B[S5][S5] = -k11New * c[S3] * EPS - km11New;
 } else {
  F[S5] = 0.;
  B[S5][S5] = 1.; /* to make B nonsingular */
 }
 R = k5 * max0(c[S1] - c1sat);
 F[S2] = EPS * R - km5 * c[S2];
 B[S2][S1] = R == 0. ? 0. : EPS * k5New;
 B[S2][S2] = -km5New;
}


void reaction_part(double *c)
/*=========================*/
/* Values of EPS and SMAX for the current mode are set in evalf(). */
{
 register double R;

 R = k9 * max0(c[S3] - c3sat);
 F[S4] = EPS * R - km9 * c[S4];
 B[S4][S3] = R == 0. ? 0. : EPS * k9New;
 B[S4][S4] = -km9New;
 if(SMAX != 0) {
    /* 0 <= c[S5] <= SMAX; see the bounds LB, UB in Initialize */
  R = SMAX - c[S5];
  B[S5][S3] = EPS * k11New * R;
  R *= k11 * c[S3];
  F[S5] = EPS * R - km11 * c[S5];
  B[S5][S5] = -k11New * c[S3] * EPS - km11New;
 } else {
  F[S5] = 0.;
  B[S5][S5] = 1.; /* to make B nonsingular */
 }
 R = k5 * max0(c[S1] - c1sat);
 F[S2] = EPS * R - km5 * c[S2];
 B[S2][S1] = R == 0. ? 0. : EPS * k5New;
 B[S2][S2] = -km5New;

 F[S1] = -F[S2] / EPS;
 B[S1][S1] = R == 0. ? 0. : -k5New;
 B[S1][S2] = km5New / EPS;
 if(SMAX != 0) {
  F[S3] = -(F[S4] + F[S5]) / EPS;
  B[S3][S3] = -(B[S4][S3] + B[S5][S3]) / EPS;
  B[S3][S5] = -B[S5][S5] / EPS;
 } else {
  F[S3] = -F[S4] / EPS;
  B[S3][S3] = -B[S4][S3] / EPS;
 }
 B[S3][S4] = km9New / EPS;

 R = -k7 * c[S0] * c[S1];
 F[S0] = R;
 R = k8 * c[S3] + 4. * R;
 F[S1] += R;
 F[S3] -= R;
 R = -k7New * c[S1];
 B[S0][S0] = R;
 R *= 4.;
 B[S1][S0] = R;
 B[S3][S0] = -R;
 R = -k7New * c[S0];
 B[S0][S1] = R;
 R *= 4.;
 B[S1][S1] += R;
 B[S3][S1] = -R;
 B[S1][S3] = k8New;
 B[S3][S3] -= k8New;

}


void init_tStep_for_evalf(void) { }
/*===============================*/


void evalf(int j)
/*===============
This procedure can easily be generalized to arbitrary
number of media; just number all the inter-media interfaces
as follows:
interface between regions i and i+1 will be
region No. i+MEDIA (for indices of EPS, SMAX);
"effective" diffusion coefficients for the left side of the
interface (DD-) will be in DD[i+MEDIA], and those for the right
side (DD+) in DD[i+REGIONS].

X and Y not needed here at all (due to mass-balance discretization; thus -DNO_X -DNO_Y).
All elements of A, B, D (and thus X and Y) are set to zero
in NL3BANDV before evalf is called. Thus only their nonzero
elements need to be set below.
*/
{
static double ja2, jd2, ja2New, jd2New, *cBarM, *cBar0, *cBarP, *DDm;
static ROW VM, V0, VP;
static int medium;
register int k;

if(j<nnodesM1) {

if(j==0){

 /*
 Assume that the first call to evalf() is with j=0 and
 on each subsequent call j is increased by 1 (THIS IS NOT
 CHECKED HERE!).
 So first do some initializations needed later for j > 0:
 */
 for(k=0; k<NEQNS; k++) {
	V0[k] = Tau * C_o[0][k] + New * C[0][k];
	VP[k] = Tau * C_o[1][k] + New * C[1][k];
 }
 if(!trueSteady) tMean= t - Tau/tStepInv; /* = time instant within tStep at
	which everything is being evaluated in the chosen
	differencing scheme of implicitness Tau */
 medium = 0;
 EPS = eps[0];
 SMAX = smax[0];
 DDm = (double *)DD[0];
 cBarM = (double *)VM;
 cBar0 = (double *)V0;
 cBarP = (double *)VP;
 ja2 = a2;
 jd2 = d2;

 /* Now do the x=0 boundary conditions: */
 reaction_part(cBar0);
 { /* Calculate iCorr: same as in iCorr_f() below,
      plus the derivatives w.r.t. the two concentrations:
   */
   register double A, r, Bb, Del, den, DiCorr0, DiCorr1;
	//PE"S0=%g, S1=%g\n",cBar0[S0],cBar0[S1]);
	// This algorithm relies on the fact that c1>0; thus NR_BOUNDS=1 is essential
   if(cBar0[S0]==0.) {
	iCorr= DiCorr1= 0.;
	DiCorr0= cBar0[S1]==0 ? INFTY : sqrt(A1/(A2*cBar0[S1]));
   } else {
	if(iCorr==0.) iCorr= pow(A2*cBar0[S0],0.333333333333333333);
	do {
		r= iCorr;
		A= A2*cBar0[S1];
		Bb= A1*cBar0[S0];
		den= 3.*r+A+A;
		Del= r*r*(r+A) - Bb*cBar0[S0];
		Del /= (r *= den);
		iCorr -= Del;
		//PE"%g %g\n", iCorr, Del);
	//} while (ABSV(Del) > ICORREPS*iCorr);
	} while (ABSV(Del) > ICORREPS);
	DiCorr0= (Bb+Bb)/r;
	DiCorr1= -A2*iCorr/den;
   }
   DiCorr0 *= New;
   DiCorr1 *= New;
   cBarS1= cBar0[S1];
   F[S0] += d20 * DDm[S0] * (cBarP[S0] - cBar0[S0]) + S0const*iCorr;
   F[S1] += d20 * DDm[S1] * (cBarP[S1] - cBar0[S1]) + S1const*iCorr;
   F[S3] += d20 * DDm[S3] * (cBarP[S3] - cBar0[S3]);
   B[S0][S0] += S0const*DiCorr0 - (D[S0][S0]= DDm[S0] * d20New);
   B[S0][S1] += S0const*DiCorr1;
   B[S1][S1] += S1const*DiCorr1 - (D[S1][S1]= DDm[S1] * d20New);
   B[S1][S0] += S1const*DiCorr0;
   B[S3][S3] -= (D[S3][S3]= DDm[S3] * d20New);
 }

} else {

 /* First update ja2,..., cBar's: */
 ja2 /= fSq; jd2 /= fSq;
 ja2New = ja2 * New; jd2New = jd2 * New;
 { double *p;
   p = cBarM;
   cBarM = cBar0;
   cBar0 = cBarP;
   cBarP = p;
   for(k=0; k<NEQNS; k++)
	cBarP[k] = Tau * C_o[j+1][k] + New * C[j+1][k];
 }

 if(j == interv1) { /* medium 1/medium 2 interface */
  /* For more than two media, change the preceding line to:
       if(j == interv1 || j == node23 || j == node34 || ...)
     Here node23, node34, ..., are the other interface nodes.
     The rest of this procedure would remain essentially the same
     for an arbitrary number of media!
  */
  EPS = eps[medium+MEDIA];
  SMAX = smax[medium+MEDIA];
  reaction_part(cBar0);
#ifdef DEBUG
	  show_dmatrix("c",6,c,interv1-1,interv1+1,0,5);
	  show_dvector("cBarM",cBarM,0,5);
	  show_dvector("cBar0",cBar0,0,5);
	  show_dvector("F - reaction part only",F,0,5);
	  show_dmatrix("B - reaction part only",6,B,0,5,0,5);
	  show_dvector("DDm a",DDm,0,2);
#endif
  { /* Diffusion part: */
    register double dM, dP;
    DDm = (double *)DD[medium+MEDIA];
    for(k=0; k<DIFFUS; k++) {
	A[k][k]= (dM= ja2 * DDm[k]) * New;
	F[k] -= dM * (cBar0[k] - cBarM[k]);
    }
    DDm = (double *)DD[medium+REGIONS];
    for(k=0; k<DIFFUS; k++) {
	  D[k][k]= (dP= jd2 * DDm[k]) * New;
	  B[k][k] -= A[k][k] + D[k][k];
	  F[k] += dP * (cBarP[k] - cBar0[k]);
    }
  }
  DDm = (double *)DD[++medium];
  EPS = eps[medium];
  SMAX = smax[medium];

 } else { /* Internal regions of both (all) media: */
  reaction_part(cBar0);
  /* Diffusion part: */
  for(k=0; k<DIFFUS; k++) {
	  F[k] += DDm[k] * (ja2*(cBarM[k]-cBar0[k]) + jd2*(cBarP[k]-cBar0[k]));
	  A[k][k] = DDm[k] * ja2New;
	  D[k][k] = DDm[k] * jd2New;
	  B[k][k] -= A[k][k] + D[k][k];
  }
 } /* end of if(j==interv1) else ... */

 /* Time derivatives: */
 for(k=0; k<NEQNS; k++) {
	F[k] += (C_o[j][k] - C[j][k]) * tStepInv;
	B[k][k] -= tStepInv;
 }

} /* end of if(j==0) (else j<nnodesM1) */

} else { /* j=nnodesM1:  x=L1+L2 boundary conditions */

	reaction_part0(cBarP); /* cBar's have not been rotated
		for this node ==> now cBarP plays the role of cBar0 */
	/* Const conc b.c. for diffusive species: */
	F[S0]= C[j][S0]-Cmed2;
	B[S0][S0]= 1.;
	F[S1]= C[j][S1];
	B[S1][S1]= 1.;
	F[S3]= C[j][S3];
	B[S3][S3]= 1.;
	/* Time derivatives for non-diffusive species: */
	for(k=DIFFUS; k<NEQNS; k++) {
		F[k] += (C_o[j][k] - C[j][k]) * tStepInv;
		B[k][k] -= tStepInv;
 	}

}

}


void pr_profiles(int MaxIter)
/*=========================*/
{
int j, i;

if(MaxIter) {
 if(trueSteady) {
  PR"%c\t\tSteady State results\n%c\n", CmntC, CmntC);
  PR"%c No. of iterations done = %d\n", CmntC, MaxIter);
 } else {
  PR"%c Transient results for t = %.10e sec = %e years\n", CmntC, t, t/Y2S);
  PR"%c Last tStep = %e sec = %.10e years\n", CmntC, tStepLast, tStepLast/Y2S);
  PR"%c Max. # of NR iterations done since the previous profile printout = %d\n",CmntC, MaxIter);
 }
 report_max_NR_error(out, CmntC);
} else {
 // MaxIter = 0 indicates that this is the echo of the initial profiles
 PR"%c\n", CmntC);
}

PR"%c\t\t All concentrations in mol/dm^3\n",CmntC);
PR"%c  x (m)      [O_2]      [Cu(I)_s]  [Cu(I)_p]  [Cu(II)_s] [Cu(II)_p] Cu(II)_surf\n", CmntC);
for(j= 0;j<nnodes;j++) {
	PR" %.10f", x[j]);
	for(i=0; i<NEQNS; i++) PR frmt, 1e-3*C_o[j][order[i]]);
	PR"\n");
}
(void)fflush(out);
}


void iCorr_f(double cS0, double cS1)
/*==================================
Calculates corrosion current iCorr for given values of
concentrations of species S0 and S1 at x=0. It is the
solution of a cubic equation which always has a single
non-negative root corresponding to iCorr. It is easier
to calculate this root by Newton method.
*/
{
 cBarS1= cS1;
 if(cS0==0.) {
	iCorr=0.;
	return;
 }
 iCorr= pow(A2*cS0,0.33333333333333333333);
 if(cS1>0.) {
	register double A, r, Del;
	do {
		r= iCorr;
		A= A2*cS1;
		Del= (r*r*(r+A) - A1*cS0*cS0) / (r*(3.*r+A+A));
		iCorr -= Del;
	//} while (ABSV(Del) > ICORREPS*iCorr);
	} while (ABSV(Del) > ICORREPS);
 }
}


#define GETVAL(X) {while(*p != ' ') p++; while(*p == ' ') p++; X =  atof(p);}

#define NOTFOUND {ErrMsg"Time %.12g not found in %s",tStart,init_from_file);exitTR(15,text);}

void init_Other(void)
/*=================*/
{
static int count = 0; /* For steady=2, count enables to take different
	 action below for steady state calculation proper and for the
	 subsequent transient calculation */

trueSteady= steady && !count++;
if(trueSteady) return;

iCorr_f(C_o[0][S0],C_o[0][S1]);
iCorrInt2 = 0.;

if(trans_to_trans) {
/* Profiles were reinitialized from a file with transient state,
   and currently steady=0.
   Initialize also iCorr and iCorrInt2 from the respective line of the old .tdep file:
*/

register int i, l = strlen(init_from_file);
register char *p;
char comnt=CmntC, *fileName;
double oth_t, aux, DT, DT0;

/* Copy the init_from_file string into a newly created buffer, and replace
 * with 'tdep' the original numerical extension (must be always present
 * - see above near  read_failed(7,...)): */
i= l-1;
while(i && init_from_file[i] >= '0' && init_from_file[i] <= '9') i--;
if(init_from_file[i] == '.') l= i;
fileName= makestr(l+5);	// allocates additional byte and fills it with 0
memcpy(fileName,init_from_file,l);
memcpy(fileName+l,".tdep",6);

if((infile = fopen(fileName,"r")) == NULL) {
	if(MAXLETT < 21+l+6) {
		// Could get here only if fileName is very loooooooong :-)
		exitTR(98,"MAXLETT needs to be increased!");
		// because the following ErrMsg needs to place 21+l+6 bytes into text[]
	}
	ErrMsg"Failed to open file: %s", fileName);
	free_str(1,&fileName);
	exitTR(99,text);
}
free_str(1,&fileName);

GETLIN0
comnt= *p;
for(;;) {
	GETLINE
	if(*p != comnt) break;
	if(*(p += 2) == 'I')
	 if(sscanf(p,"Int_icorr is Int(icorr dt) from %le", &t000)) break;
}
GETLIN0
GETLIN0
GETLIN0
GETLIN0
GETVAL(oth_t)
DT0=tStart-oth_t*Y2S;
if(DT0<0) NOTFOUND
	/* ? all times in the old .tdep file LARGER than the profiles time ? */
for(;;) {
 GETLIN0
 if(p==NULL) NOTFOUND	/* ?  - " -  SMALLER  - " -  ? */
 GETVAL(oth_t)
 oth_t *= Y2S;
 DT= tStart-oth_t;
 if(DT<=0.) {
	if(DT0>-DT) {
		GETVAL(iCorr)
		GETVAL(aux)
		GETVAL(iCorrInt2)
	} else oth_t += DT-DT0;
	iCorr *= 1e-6;
	iCorrInt2 *= 2.;
	break;
 }
 GETVAL(iCorr)
 GETVAL(aux)
 GETVAL(iCorrInt2)
 DT0=DT;
}
PE"Time %.12g sec from %s used to initialize iCorr and iCorrInt2\n", oth_t, init_from_file);

} /* end of: if(trans_to_trans) */

iCorrOld = iCorr;
Pm"%c Int_icorr is Int(icorr dt) from %g %s to present\n%c\n",
	CmntC,t000,time_units,CmntC);
Pm"%ctime (years)  icorr (microA/m^2)  Ecorr (V)  Int_icorr (C/m^2)\n%c",
	CmntC,CmntC);
pr_Other(1);
Fm
}


int update_Other(void)
/*==================*/
{
 /*
 iCorrInt2 = 2 * integral of iCorr from t=0 to current t
 by trapezoidal rule;
 iCorr corresponds to tMean. Integration of iCorr must also be
 carried up to tMean. Following is trapezoidal rule for the time
 interval from the previous to the present tMean (its length is
 given by the expression within the last ()):
 */
 if(!trueSteady) {
	iCorrInt2 += (iCorr + iCorrOld) * (tStepLast*New + tStepLastBut1*Tau);
	iCorrOld = iCorr;
	tStepLastBut1= tStepLast;
 }
 if(t < 500.*tStepInit) return 1;
 else return 0;
}


void pr_Other(int only_Other)
/*===========================
 In transient.c,
 only_Other =   1  if only pr_Other is called at this time t
		0  if also pr_profiles is called
*/
{
double ECorr;

ECorr= iCorr + A2 * Gamma * cBarS1;
ECorr= ECorr>0. ? A3 + Econst * log(A4 * ECorr) : -INFTY;

if(trueSteady)	{
	PR"\n%cSteady state: iCorr = %.12g microA/m^2, ECorr = ", CmntC, iCorr*1e6);
	if(ECorr == -INFTY) PR"- infinity\n\n");
	else PR" %.12g V\n\n", ECorr);
} else {
	Pm frmt, tMean/Y2S);
	Pm frmt, iCorr*1e6);
	Pm frmt, ECorr);
	Pm frmt, 0.5 * iCorrInt2);
	Pm"\n");
}

if(!(only_Other || trueSteady)) {
	/* Add output for the exact time at which profiles are outputted: */
	iCorr_f(C_o[0][S0],C_o[0][S1]);
	ECorr= iCorr + A2 * Gamma * cBarS1;
	ECorr= ECorr>0. ? A3 + Econst * log(A4 * ECorr) : -INFTY;
	iCorrInt2 += (iCorr + iCorrOld) * tStepLast*Tau;
	iCorrOld = iCorr;
	tStepLastBut1= 0.;
	Pm" %.10g", t/Y2S);
	Pm frmt, iCorr*1e6);
	Pm frmt, ECorr);
	Pm frmt, 0.5 * iCorrInt2);
	Pm"\n");
}

Fm
}
