/* File demo/demo2/demo2common.c; TRANSIENT V4.3;  31 Oct. 2023 */
/* (C) 1995,2001,2005,2023 Miroslav Kolar */

/* ============================================================ */
/* Diffusion of gas (e.g., H_2) generated inside a cylindrical	*/
/* container at a constant rate out of this container through	*/
/* a (narrow) opening (fill pipe).				*/
/* Demonstration modules No. 2a and 2b for TRANSIENT		*/
/* ============================================================ */

/* Common functions for both demo2a and demo2b			*/

#define _ADDITIONAL_SOURCE_

#define PDM_H "demo/demo2/demo2.h"
	/* path of demo2.h relative to the TRANSIENT directory */
#include "../../transien.h"
	/* >>>> READ the information in transien.h of TRANSIENT <<<< */

#if defined(MSDOS) || defined(_MSC_VER)
#define M_PI 3.14159265358979
#endif

#if VERSN == 1
#define VersioN "a"
extern double delta, h;
extern int interface;
#else
#define VersioN "b"
extern double h1, h2;
#endif

extern int bcr;

int get_neqns_etc(void)
/*===================*/
{
title = "			TRANSIENT\n\
		    Demonstration No. 2"VersioN"\n\
		Gas generation and diffusion";
time_units = "sec";
cDifMxInit = 0.001;

/* To achieve steady state for default input
parameters; needed for succesful Laplace T. comparison: */
tEnd = 10519200.; /* 4 months */
// E.g., for l1=2, l2=200, d1=0.6 cm, one needs: tEnd = 1577880000.; /* 50 years */
tOutMulF = 2.;

/* Convert string BCr to a numerical value: */
bcr= strcmp(BCr,"3point");

return 1;  /* = # of equations */
}

extern double
	A1, A2,
	*C, *C_O, D_T,
	*k, *ki,
	c0, cl1, cl1l2, c0p, cl1p, cl1l2p;

void Initialize(double *c_lin, double *c_o_lin)
/*===========================================*/
{
int j;

C = c_lin;
C_O = c_o_lin;

for(j=0; j<nnodes; j++) C_O[j] = C[j] = 0.;

c0= cl1= cl1l2= c0p= cl1p= cl1l2p= 0.;
}

/********************************************************/
/* C[j] = c_lin[j]  - hydrogen concentration		*/
/********************************************************/


void init_tStep_for_evalf(void) { }
/*===============================*/


void read_profiles(char *filename)
/*==============================*/
{
PE"\nNo action in read_profiles() implemented\n");
PE"Initializations made in Initialize() unchanged\n\n");
}


void init_Other(void)
/*=================*/
{
int j;

if(steady) return;

/* k[j] will be used to accumulate contributions to the
   numerical Laplace transform, k(x,s), of c(x,t) for a single
   value of the Laplace variable s (x = j * h);
   to be compared with the known analytical results for k(x,s)
   for c(x,0)=0. Does make sense for steady=0 only.
   Assuming tStart=0, tEnd>0; assured in demo2.h .
*/
k= dvector(nnodes);
ki= dvector(nnodes);
for(j=0; j<nnodes; j++) ki[j]= k[j]= 0.;
}


int update_Other(void)
/*==================*/
{
if(steady) return 0;

{
 int j;
 register double pre, aux = exp(-s * t);

 for(j=0; j<nnodes; j++) {
	pre= ki[j];
	ki[j] = aux * C_O[j];
	k[j] += (pre + ki[j]) * tStepLast;
 }
 /* Trapezoidal integration; factor 0.5 missing in k[j] */
 return 0;
}
}


void pr_Other(int only_Other)
/*=========================*/
/* In transient.c,
   only_Other = 1  if only pr_Other is called at this time t
		0  if also pr_profiles is called */
/* The only purpose of this procedure is to print two columns
   comparing the numerical and analytical  Laplace Transform for a
   single value of  s  at the end of the integration at tEnd */
{
if(steady == 0 && !only_Other && t >= tEnd) {
 double LT1, LT2, LT3, LT4, LT5;
 int
#if VERSN == 1
	inlet_last = interface,
#endif
	j;
 register double aux, aux1;
 
 aux1= fabs(c0-c0p);
 aux= fabs(cl1-cl1p);
 if(aux>aux1) aux1= aux;
 aux= fabs(cl1l2-cl1l2p);
 if(aux>aux1) aux1= aux;
 aux= 3.*aux1/(fabs(c0)+fabs(cl1)+fabs(cl1l2));
 if(aux > STEADYtol) 
  Pm"%c WARNING: Steady state possibly not yet reached (rel diff %g):\n%c\
	numerical Laplace transform may have no meaning!\n%c\n",
	CmntC,aux,CmntC,CmntC);

 aux = exp(-s * t) / s;

#if VERSN == 1
 if(delta < 0.) inlet_last--;
#endif

 /* Calculate the coefficients in the analytical Laplace Transform
   (LT5 and LT1 are used as auxilliary variables first):  */
 LT2 = G / (s * s);
 LT4 = sqrt(s/D_T);
 LT1 = A2 * tanh(LT4 * l2);
 LT5 = A1 + LT1 * tanh(LT4 * l1);
 if(LT4 * (l2>l1 ? l2 : l1) + log(LT5) > 691) {
  PE"s  is too large, cannot calculate analytical Laplace transform\n");
  return;
 }
 LT1 *= LT2 / (cosh(LT4 * l1) * LT5);
 LT3 = LT2 * A1 / (cosh(LT4 * l2) * LT5);
#if VERSN == 1
 LT5 = LT4 * (l1 + l2);
 LT4 *= h;
#endif

     /* First add to k(x,s) the contribution from tEnd to
	infinity, assuming that steady state has already
	been achieved at tEnd, i.e., that c(x,t)=c(x,tEnd)
	for all t > tEnd: */
 for(j=0; j<nnodes; j++) k[j] = 0.5*k[j] + aux * C_O[j];

 Pm"%c Laplace transform comparison for s = %g\n", CmntC, s);
 Pm"%c   x (cm)      Numerical    Analytical\n", CmntC);
 
#if VERSN == 1
 for(j= 0;j<=inlet_last;j++) Pm" %f  %e  %e\n", j * h, k[j],
				 LT1 * sinh(LT4 * j));
 for( ;j<nnodes;j++) Pm" %f  %e  %e\n", j * h, k[j],
#else
 LT5 = LT4 * h1;
 for(j= 0;j<=interv1;j++) Pm" %f  %e  %e\n", j * h1, k[j],
				 LT1 * sinh(LT5 * j));
 LT5 = LT4 * (l2 + interv1 * h2);
 LT4 *= h2;
 for( ;j<nnodes;j++) Pm" %f  %e  %e\n", l1 + (j - interv1) * h2, k[j],
#endif
				LT2 - LT3 * cosh(LT5 - LT4 * j));
 Fm
}
}
