/* File lufacv.c; TRANSIENT V4.1; 20 Feb 2011 */
/* (C) 1995,2001,2005,2011 Miroslav Kolar */

/* lufacv.c  - solve n x n set of linear eqns

* 1995: VECTORIZATION
* lufac() now returns -1 if there is no unique solution;
  linear representation of l and u introduced.
* 2001: TRUE_FABS

* Direct factorization with maximal column pivoting
* based on Algorithm 6.5 from "Numerical Analysis, 3rd Ed." by
  R. L. Burden and J. Douglas Faires, Prindle, Weber & Schmidt,
  Boston, Massachussetts, U.S.A. (1985).
*/

#include <stdio.h>
#include <math.h>
#include "dynarr.h"
#include "defs.h"

/* M. Kolar:
1. #define MACHINEZERO 1e-40 eliminated;
 first decreased to 1e-300, then found that a finite value is not needed at all;
 here an implicit MACHINEZERO=0 is used satisfactorily.

2. On most (all?, at least on DEC Ultrix and OSF, Sun Sparc, and on SGI)
 UNIX machines, multiplication, a*a, is much faster (5 to 40 times)
 than fabs(a). But calculating below the maximum of a set of a*a instead
 of fabs(a) means for very small a that the 'machine zero' is aproached sooner.
 A suitable compromise is the (if(a<0) a = -a;) construct
 which is comparable to (a*a) in the CPU time use (or at most 4 times
 slower on the DEC machines).
   On the other hand, under DOS/Windows, fabs(a) is faster than a*a or
 the above if construct.
   See definition of TRUE_FABS in defs.h, and the 3 instances of its use below.
*/

#define PE (void)fprintf(stderr,
#define FE (void)fflush(stderr);

#define Ti(j)		(j * (2 * neqns - j - 1) >> 1)
	/* Nonzero elements of a lower triangular matrix l are arranged
	   column-wise in a vector, such that l[i,j] is stored as the
	   (i+Ti(j))-th element of this vector.
	   Similarly, an upper triangular matrix u is stored row-wise,
	   such that u[i,j] is the (Ti(i)+j)-th element of a vector.
	*/


int lufacCompiled(void) { return TRUE_FABS; }
//===========================================


void switchrows(double *m, int ncol, int rowa, int rowb)
/*====================================================*/
/* Linearized matrix with ncol columns */
{
int i;
register double temp;

    rowa *= ncol;
    rowb *= ncol;
    for (i=0;i<ncol;i++) {
	temp = m[rowa+i];
	m[rowa+i] = m[rowb+i];
	m[rowb+i] = temp;
    }
} /* end of switchrows */

void switchrowsl(double *l, int neqns, int rowa, int rowb)
/*======================================================*/
{  /* Only columns less than rowb; for lower triang matrix only */
int i, ii;
register double temp;

    temp = l[rowa]; l[rowa] = l[rowb]; l[rowb] = temp;
    for (i=1;i<rowb;i++) {
	ii = i * (2 * neqns - i - 1) >> 1;
	temp = l[rowa+ii]; l[rowa+ii] = l[rowb+ii]; l[rowb+ii] = temp;
    }
} /* end of switchrows */


/* l[i][i] = 1 */
int lufac(int neqns, double *a, double *l, double *u, int *nrows)
/*=============================================================*/
{
int i, j, k, kk, p, nn = neqns*neqns;
register double sum;

    for (i=0;i<neqns;i++) nrows[i] = i;
    {	register double large, c;
#if TRUE_FABS
	large = fabs(a[0]);
#else 
	c = a[0]; large = c < 0. ? -c : c;
#endif
	p = 0;
	for(i=neqns; i<nn; i += neqns) {
#if TRUE_FABS
		c = fabs(a[i]);
#else 
		c = a[i]; if(c < 0.)  c = -c;
#endif
		if(c > large) { large = c; p = i/neqns; }
	}
	/* if(large < MACHINEZERO) */
	if(large == 0.) {
		PE"lufac(pivot):  no unique solution exists"); FE
		return(-1);
	}
    }
    if (p != 0) {
              nrows[p] = 0;
              nrows[0] = p;
         switchrows(a,neqns,p,0);
    }

    l[0] = 1.0; u[0] = a[0];

    for (j=1;j<neqns;j++) {
         u[j] = a[j];
         l[j] = a[j*neqns]/u[0];
    }

    for (i=1;i<neqns-1;i++) {
	int ii = Ti(i);

	/*  p = pivot(i,neqns,a,l,u);  substituted by the body of
					the former pivot() procedure:  */
	/* return index p to row such that
	   a[p][i] - sum(l[p][k]*u[k][i]) k=0,i-1 is a maximum for i<=p<=n */
	{ register double large=0.;
	  for (j=i;j<neqns;j++) {
	    sum = a[j*neqns+i];
	    for (k=0;k<i;k++) { kk = Ti(k); sum -= l[j+kk] * u[kk+i]; }
#if TRUE_FABS
	    sum = fabs(sum);
#else
	    if(sum < 0.) sum = -sum;
#endif
	    if (sum > large) {
	         large = sum;
	         p = j;
	    }
	  }
	  /* if(large < MACHINEZERO)  */
	  if(large == 0.) {
		PE"lufac: col #0 is all zeroes"); FE
		return(-1);
	  }
	}
	/* End of pivot */

        if (p != i) {
              k = nrows[p];
              nrows[p] = nrows[i];
              nrows[i] = k;
              switchrowsl(l,neqns,p,i);
              switchrows(a,neqns,p,i);
        }

        l[i+ii] = 1.0;
        sum = a[i*neqns+i];
        for (k=0;k<i;k++) { kk = i + Ti(k); sum -= l[kk] * u[kk]; }
        u[ii+i] = sum;

        for (j=i+1;j<neqns;j++) {
              sum = a[i*neqns+j];
              for (k=0;k<i;k++) { kk = Ti(k); sum -= l[i+kk] * u[kk+j]; }
              u[ii+j] = sum;
              sum = a[j*neqns+i];
              for (k=0;k<i;k++) { kk = Ti(k); sum -= l[j+kk] * u[kk+i]; }
              l[j+ii] = sum/u[ii+i];
        }
    }

    sum = a[nn-1];
    for (k=0;k<neqns-1;k++) { kk = neqns-1 + Ti(k); sum -= l[kk] * u[kk]; }
    /* if (sum < MACHINEZERO)  */
    if (sum == 0.) {
		PE"lufac:  no unique solution exists"); FE
		return(-1);
    }
    kk = (neqns-1) * (neqns+2) >> 1;
    l[kk] = 1.0;
    u[kk] = sum;
    return(1);
} /* end of lufac */


void lusolve(int neqns, double *l, double *u, int *nrows, int ib, double *x)
/*========================================================================*/
{
int i,j;
register double sum;

 /* solve lower-triangular system Lz = b, where b[ib]=1 and b[i]=0 for i!=ib */
 /* the same memory is used for z as for x */
 x[0] = nrows[0] == ib ? 1. : 0.;
 for (i=1;i<neqns;i++) {
    sum = nrows[i] == ib ? 1. : 0.;
    for (j=0;j<i;j++) sum -= l[i+Ti(j)] * x[j];
    x[i] = sum;
 } /* Up to now x represented z */

 /* solve upper-triangular system Ux = z */
 x[neqns-1] /= u[(neqns-1)*(neqns+2) >> 1]; /* = z[neqns-1]/u[neqns-1][neqns-1] */
 for (i=neqns-2;i>=0;i--) { int ii = Ti(i);
    sum = x[i]; /* = z[i] */
    for (j=i+1;j<neqns;j++) sum -= u[ii+j] * x[j];
    x[i] = sum/u[ii+i];
 }
} /* end of lusolve */
