/* File nl3band1.c; TRANSIENT V4.3; 22 Jul 2018
 * (C) 1997-2012,2015,2018 Miroslav Kolar
 *
 * This is a version of nl3bandv.c for a single species (neqns = 1).
 *							 ~~~~~~~~~
 *==================================================================
 * Bound checking simplified.

 * Functions defined in this file:
	void startband(...)
	double runband(...)
	void endband(void)
	void cr_band_mem(int NJ)
	void ApplyBounds(double*c)
	void Randomize(double Amplitude, double *c_o, double *c)
	void changeNoNodes(int NJ)
	void SetCheckBounds(void)
	void changeBound(int e, char which, double *BDA)
	void report_max_NR_error(FILE *out, char comment_char)
	void report_last_NR_error(FILE *out, char comment_char)
	void nl3bandCompiled(char CMNtC)  ...  in nl3_Decl.h
	int lufacCompiled(void) { return -1; }

 * -------------------------------------------------------------------------
 * The arguments and return values of these functions are exactly the same
 * as in nl3bandv.c - see the beginning of that file.

 * The only variations from nl3bandv.c are:

 *  Here the only valid value for neqns is 1.
 *  LowBound and UpBound have only one element, 0th.
 *  This file has no need for matinvNMC().
    Return value of -8 in runband() still corresponds to a
    "matrix" inversion failure, but it is simply caused by B = 0.
*/


#define _NL3_PARENT_ '1'
// More declarations (common with nl3bandv.c):
#include "nl3_Decl.h"


static double A, B, D, F,
#if !NO_X
	x,
#endif
	*LB0, *UB0;	// To be used if changeBound() called

static int
	bound_type,
#if !NO_Y
	NNODE3,
#endif
	NNODE2;


/* Compared to the original Spotnitz' nl3band.c, here F and xi
   (and the original dC) have the opposite sign, the original dC is now
   put into xi, and X shares the memory with A and Y with D (they are never
   used simultaneously).  */


void cr_band_mem(int NJ)
/*====================*/
{
xi= dvector(NJ==1 ? 2 : NJ+NJ-1);
E= xi + NJ;
#if NodeNumberChangeENABL
NNODEmax=
#endif
NNODE= NJ;
NNODE1= NNODE - 1;
NNODE2= NNODE - 2;
#if !NO_Y
NNODE3= NNODE - 3;
#endif
FresidPREC= TOL*NJ*2e-16;
CresidPREC= TOLC*FresidPREC;
if(monitor2) PE"TOL=%g, FresidPREC=%g, CresidPREC=%g = TOLC*FresidPREC\n\n",
	TOL, FresidPREC, CresidPREC);
}


#if NodeNumberChangeENABL
void changeNoNodes(int NJ)
/*======================*/
{
if(NJ>NNODEmax) EXITTRm7
if(NJ<1) EXITTRm7A
NNODE= NJ;
NNODE1= NNODE - 1;
NNODE2= NNODE - 2;
#if !NO_Y
NNODE3= NNODE - 3;
#endif
FresidPREC= TOL*NJ*2e-16;
CresidPREC= TOLC*FresidPREC;
if(monitor2) PE"TOL=%g, FresidPREC=%g, CresidPREC=%g = TOLC*FresidPREC\n\n",
	TOL, FresidPREC, CresidPREC);
}
#endif


void endband(void)
/*==============*/
{
free_dvector(&xi);
#if NodeNumberChangeENABL
NNODEmax=
#endif
NNODE= 0;
}


/* The following 4 macros are used in the include files below: */
#define EQNO 0
#define FORA for(i = 0; i < NNODE; i++)
#define FORB for(i = 0; i < NNODE; i++)
#define NoDe i

#define INCONSIST { exitTR(-1,"Lower and upper bounds inconsistent"); }


void SetCheckBounds(void)
/*=====================*/
{
register int i, j;
register double upBnd, lowBnd, *lp, *up;

j= (int)((UB && UB[0]) ? UB[0][0] : 0);
if(j<0 || j>3) exitTR(-1,"Upper bound type has to be 0, 1, 2 or 3");
bound_type= (int)((LB && LB[0]) ? LB[0][0] : 0);
if(bound_type<0 || bound_type>4)
	exitTR(-1,"Lower bound type has to be 0, 1, 2, 3 or 4");
if(bound_type<4) bound_type += 4 * j;
else {
	if(LB[0][1]<0.) exitTR(-1,"Zero cutoff must be non-negative");
	bound_type = 16 + j;
}

/* Check bound consistency, if both bounds are defined, LB <= UB must hold: */
switch(bound_type) {

#include "nl3_strt.h"

}
}


void startband(int neqns, int NJ, double*cTR)
/*=========================================*/
{
RAND_WARNING

if(neqns != 1) exitTR(-1,"nl3band1.c is only for neqns=1!!!!!!");
if(NJ < 1) exitTR(-1,"Number of grid nodes must be positive!");

if(NNODE==0)
  cr_band_mem(NJ);
else if(NNODE<NJ) {
  endband();
  cr_band_mem(NJ);
}
cMax = FMax = 0.;
converged = -99.; /* means: no call to runband since last startband */
c= cTR;
initNL3BANDV(&LB,&UB,&F,&A,&B,&D);
SetCheckBounds();
}


void changeBound(int e, char which, double *BDA)
/*============================================*/
// e unused in nl3band1.c; can be only 0
{
switch(which) {
	case 'l':
		if(!LB) LB= &LB0;
		LB[0]= BDA;
		break;
	case 'u':
		if(!UB) UB= &UB0;
		UB[0]= BDA;
		break;
}
SetCheckBounds();
}


void ApplyBounds(double*c)
/*======================*/
{
register int i;
register double *lp, *up, upBnd, lowBnd;

/* Enforce the lower and upper bound: */
switch(bound_type) {

#include "nl3_ApBo.h"


#define INV_FAIL if(B==0) { PE"B=0: inversion failed at node %d\n", node);\
 if(monitor2) PE" %d: (>=%g) >=%g %d 0\n",NRiter,fresidT,fresidM,nodefresidM);\
 return(converged = -8.); }

double runband(double t)
/*====================*/
{
int
	node;
#include "nl3_Decl1.h"

converged = -1.;

for(NRiter = 1; ; NRiter++)
{

fresidT = fresidM = 0.;

for(node= 0;node<NNODE;node++)
{

B= A= D= 0.;
/* Here A is also used to store X, and D to store Y;
   X and A, and Y and D, are never used at the same time */

evalf(node);

if(monitor2) {
 { register double aux;
	aux = ABSV(F);
	fresidT += aux;
	if(fresidM < aux) { fresidM = aux;
			    nodefresidM = node; }
 }
} else fresidT += ABSV(F);

if(node==0) {
INV_FAIL
B = 1. / B;
E[0] = B * D;
#if !NO_X
x = B * A;
#endif
xi[0] = B * F;
}

else if(node<NNODE1) {
F -= A * xi[node-1];
B -= A * E[node-1];
#if !NO_X
if(node==1) D -= A * x;
#endif
INV_FAIL
A = 1. / B;
E[node] = A * D;
xi[node] = A * F;
}

else {
#if !NO_Y
A -= D * E[NNODE3];
#endif
B -= A * E[NNODE2];
#if !NO_Y
F -= D * xi[NNODE3];
#endif
F -= A * xi[NNODE2];
INV_FAIL
A = 1. / B;
xi[node] = A * F;
} /* end if(node==0) */

} /* end of  for(node= 0;node<NNODE;node++) */

for(node= NNODE2;node>0;node--)
 xi[node] -= E[node] * xi[node+1];

xi[0] -= E[0] * xi[1];
#if !NO_X
xi[0] -= x * xi[2];
#endif

{
register int i;
register double aux;
const int eqfresidM= 0, eqcresidM= 0;

#if NR_BOUNDS
register double *lp, *up, upBnd, lowBnd;

// Update c[], and check lower and upper bounds (remember: dC = -xi):
switch(bound_type) {

#include "nl3_runb.h"

}

#else
// Update c only, no bound checks:
FORA c[i] -= xi[i];
#endif

cresidT = 0.;
cresidM = 0.;
if(monitor2) {
 for(node=0;node<NNODE;node++) {
	aux = ABSV(xi[node]);
	cresidT += aux;
	if(cresidM < aux) {
		cresidM = aux;
		nodecresidM = node;
	}
 }
} else {
 for(node=0;node<NNODE;node++) {
	aux = ABSV(xi[node]);
	cresidT += aux;
	if(cresidM<aux) cresidM = aux;
 }
}

#include "nl3_conv.h"

} // End of  register double aux;

} /* end of  for(NRiter = 1; ; NRiter++)  */

if(converged>0.) {
#if !NR_BOUNDS
 ApplyBounds(c);
#endif
 if(monitor2) {
	if(cresidM > cMax) { cMax = cresidM;
			nodecMax = nodecresidM; tcMax = t; }
	if(fresidM > FMax) { FMax = fresidM;
			nodeFMax = nodefresidM; tFMax = t; }
 } else {
	if(cresidM > cMax) { cMax = cresidM; tcMax = t; }
 }
 if(cresidT > cTmax) { cTmax = cresidT; tcTmax = t; }
 if(fresidT > FTmax) { FTmax = fresidT; tFTmax = t; }
}

return(converged);
}


#include <time.h>

void Randomize(double Amp, double *c_o, double *c)
/*==============================================*/
{
register int i;
register double *lp, *up, upBnd, lowBnd, RaU, RaL;

srand((unsigned)time(0));
/* Randomize while enforcing the lower and upper bounds: */
switch(bound_type) {

#include "nl3_Rand.h"



#define PR (void)fprintf(out,

void report_max_NR_error(FILE *out, char comment_char)
/*==================================================*/
{
char *str = transCalc ? "Max average" : "Average";
if(monitor2) {
 PR"%c Max converged NR change in c, %e, was at node # %d", comment_char, cMax, nodecMax);
 if(transCalc) PR" at t = %e", tcMax);
 PR"\n%c Max residual |F|, %e, was at node # %d", comment_char, FMax, nodeFMax);
 if(transCalc) PR" at t = %e", tFMax);
 PR"\n");
 cMax = FMax = 0.;
} else {
 PR"%c Max converged NR change in a c[node][0] was %e", comment_char, cMax);
 if(transCalc) PR" at t = %e", tcMax);
 PR"\n");
 cMax = 0.;
}
PR"%c %s converged NR change in c per node was %e", comment_char, str, cTmax/NNODE);
if(transCalc) PR" at t = %e", tcTmax);
PR"\n%c %s residual |F| per node was %e", comment_char, str, FTmax/NNODE);
if(transCalc) PR" at t = %e", tFTmax);
PR"\n%c\n", comment_char);
(void)fflush(out);
cTmax = FTmax = 0.;
}

void report_last_NR_error(FILE *out, char comm)
/*===========================================*/
{
if(converged == -99.) PR"%c No call to runband() made yet\n", comm);
else if(converged == -8.)
 PR"%c inversion failed on last call to runband()\n", comm);
else {
 if(monitor2) {
  PR"%cLast fresidM = %g at node # %d\n", comm, fresidM, nodefresidM);
  //PR"%c rel cresidM = %g (|delta_c|) at node # %d\n",comm,cresidM,nodecresidM);
  PR"%c cresidM = %g (|delta_c|) at node # %d\n",comm,cresidM,nodecresidM);
 }
 PR"%c fresidT = (sum of |F|) = %g; cresidT = (sum of |delta_c|) = %g\n",
		comm, fresidT, cresidT);
}
(void)fflush(out);
}


// To indicate non-use of lufac()
int lufacCompiled(void) { return -1; }
//====================================
