/* File nl3bandv.c; TRANSIENT V4.3; 25 Nov 2018
 * (C) 1995-2012,2015,2018 Miroslav Kolar
 *
 * In 1995, this file was forked from the steady-state-only nl3band.c by Robert Spotnitz
 *
 * It uses completely redesigned modules matinvv.c, matinvv.h,
 *      lufacv.c, lufacv.h, matrixv.c, and
 *      dynarr.c and dynarr.h.
 * All matrices are linearized (vectorized ...v.c)
 * Bound checking simplified.
 * Bounds can change during a run.
 * This version can handle arbitrary number of nodes, starting
 * from one. Number of nodes can change during the run provided
 * it remains below the maximum initially specified in startband.
 * Position of nodes cannot change. Could be used for moving boundary.
*/

//#define DEBUG

/* Functions defined in this file:
	void startband(...)
	double runband(...)
	void endband(void)
	void cr_band_mem(int neqns,int NJ)
	void ApplyBounds(double*c)
	void Randomize(double Amplitude, double *c_o, double *c)
	void changeNoNodes(int)
	void changeBound(int e, char which, double *BDA)
	void report_max_NR_error(FILE *out, char comment_char)
	void report_last_NR_error(FILE *out, char comment_char)
	void fresidsA2(void)
	void fresidsA(void)
	void fresidsE2(void)
	void fresidsE(void)
	void cresidsA2(void)
	void cresidsA(void)
	void cresidsE2(void)
	void cresidsE(void)
	void nl3bandCompiled(char CMNtC)  ...  in nl3_Decl.h


 - void startband(int neqns, int NJ, double*cTR)
   ----------------------------------------------------------------------
 *  neqns  -  number of equations (concentrations, species)
 *  NJ	  -  # of nodes (can start from 1)
 *  cTR  -  address of the "concentration" array c from TRANSIENT


 - LB and UB contents now passed from a PDM directly through initNL3BANDV()
   called from startband(): 				     --------------
   -----------------------
 *  LB[k]  -  pointer to a Bound Definition Array (BDA) containing information
		about the lower bound for the k-th dependent variable
 *  UB[k]  -  the same for the upper bound
      The 0-th element of a BDA denotes the type of the bound:
		0 - no bound set
		1 - a constant bound (the same for all nodes)
		2 - piece-wise constant bound
		3 - a distinct bound for each node
		4 - special case for LB only:
				set zero cutoff, the same for all nodes
      The remaining elements of such an array are as follows:
      Type 0: no other elements necessary.
	   1: one more element equal to the constant bound.
	   2: an arbitrary number, nC, of couples of additional elements, each
		couple consisting of a bound, c_m, followed by the index, j_m,
		of the last node for which this bound applies. The last couple
		is recognized either by j_nC >= NJ-1 or j_nC <= j_(nC-1). In
		either case, the last bound, c_nC, holds for nodes j_(nC-1)+1
		trough NJ-1. The first bound, c_1, holds for nodes 0 through
		j_1, etc. j_1 >= 0 (j_1 < 0 would mean that c_1 would be
		applied to all nodes).
	   3: NJ additional elements, each equal to the bound for a single node
		starting from 0 to NJ-1.
	   4: one more element equal to the zero cutoff: if less than this
		cutoff, the value of the k-th dependent variable will be set
		exactly to 0.
      (Any of the pointers LB, LB[k], UB, UB[k] being NULL also means
	'no bound set' for all k, or a single k.)


 - double runband(double t)
   -------------------------------------------------------------------------
   Input: c     -  initial guess for concentrations (see cTR above)
	  t     -  current time, not needed inside the procedure, used only
		   to record the times for which the largest |F| residual and
		   largest converged change in c occurred,
		   to be able to report them in report_max_NR_error

   Output: c      -   final approximation for concentrations (see cTR above)
	   NRiter -   number of Newton-Raphson iterations

   RETURN VALUE:  1  if convergence was achieved w.r.t. cresidT (or fresidT)
		 >1  return value = (final cresidT)/residPREC - degree of (non)convergence
		 -8  if matinvNMC() failed (lufac() could not find a unique solution)
		 -9  if cresidT or fresidT is not finite

   TRANSIENT's exit codes generated directly in nl3bandv.c
		 -1  if various inconsistencies in concentration bounds;
			or if the number of grid points is <= 0
		 -7  if the requested new # of nodes exceeds the current maximum

   Iteration process is terminated whenever one of the following happens:
   - if convergence is achieved;
   - if apparent divergence/oscillation: cresidT or fresidT start to grow
   - if matrix inversion (matinvNMC()) fails.


 - evalf() to evaluate F_k and dF_k/dc_l - now directly declared as an
   -------					extern function


 - void endband(void)
   ------------------


 - void ApplyBounds(double*c)
   --------------------------
      Apply the above defined lower and upper bounds to c.


 - void Randomize(double Amplitude, double *c_o, double *c)
   --------------------------------------------------------
      Randomly modify c_o by at most +/-Amplitude but so that the
   modified value remains within bounds if defined, and return the result in c.
*/


#include "matinvv.h"

#define _NL3_PARENT_ 'v'
// More declarations:
#include "nl3_Decl.h"

/* NNODE can also be 1 or 2. For these cases, one has to make sure
that the X and Y matrices are empty (have all elements equal to zero).
*/

extern void mmult(int neqns, double *p, double *l, double *r);
extern void mvmult(int neqns, double *p, double *m, double *v);
extern void mmultsub(int neqns, double *ps, double *l, double *r);
extern void mvmultsub(int neqns, double *ps, double *m, double *v);
#ifdef DEBUG
extern void show_dmatrix(char *name, int order, double *m,
			int l1, int u1, int l2, int u2, char *frm);
extern void show_dvector(char *name, double *v, int l, int u);
#endif

extern int
	Nexcl, *equUsed;

static double
#if !NO_X
	*x,
#endif
	*A, *B, *D, *X, *Y, *F;
static int
	eqfresidM, eqcresidM,
	eqcMax, eqFMax, *bound_type,
	NEQNS= 0, ntN, nSq;


/* Compared to the original Spotnitz' nl3band.c, here F and xi
   (and the original dC) have the opposite sign, the original dC is now
   put into xi, and X shares the memory with A and Y with D (they are never
   used simultaneously).
   Here all matrices (c, xi, A, B, D  and E) are linearized (stored as vectors).
*/


void cr_band_mem(int neqns,int NJ)
/*==============================*/
{
nSq = neqns * neqns;
ntN= NJ * neqns;
F= dvector(ntN + neqns + nSq *
#if NO_X
3
#else
4
#endif
);
	/* F points to a continuous block of memory for F,A,B,D,x,xi */
A= F + neqns;
B= A + nSq;
D= B + nSq;
X= A;
Y= D;
#if NO_X
xi= D + nSq;
#else
x= D + nSq;
xi= x + nSq;
#endif
cr_matinv_mem(neqns);
NEQNS= neqns;
#if NodeNumberChangeENABL
NNODEmax=
#endif
NNODE= NJ;
NNODE1= NNODE - 1;
E= dvector(NJ==1 ? nSq : nSq * NNODE1);
bound_type = ivector(neqns);
FresidPREC= TOL*NJ*2e-16;
CresidPREC= TOLC*FresidPREC*(neqns-Nexcl);
if(monitor2) PE"TOL=%g, FresidPREC=%g, CresidPREC=%g = TOLC*FresidPREC*%s\n\n",
	TOL, FresidPREC, CresidPREC, Nexcl?"(neqns-Nexcl)":"neqns");

#ifdef DEBUG
   PE"<-- "__FILE__": Initial memmory addresses of various NR arrays:\n\
F: %p, A: %p, B: %p, D: %p, xi: %p, E: %p\n -->\n",

#if (defined(MSDOS) || defined(_Windows)) && !defined(NO_EXT_KEYS)
	(double far *)F, (double far *)A, (double far *)B,
	(double far *)D, (double far *)xi, (double far *)E);
#else
	F, A, B, D, xi, E);
#endif
   fflush(stderr);
#endif
}


#if NodeNumberChangeENABL
void changeNoNodes(int NJ)
/*======================*/
{
if(NJ>NNODEmax) EXITTRm7
if(NJ<1) EXITTRm7A
ntN= NJ*NEQNS;
NNODE= NJ;
NNODE1= NJ-1;
FresidPREC= TOL*NJ*2e-16;
CresidPREC= TOLC*FresidPREC*(neqns-Nexcl);
if(monitor2) PE"TOL=%g, FresidPREC=%g, CresidPREC=%g = TOLC*FresidPREC*%s\n\n",
	TOL, FresidPREC, CresidPREC, Nexcl?"(neqns-Nexcl)":"neqns");
}
#endif


void endband(void)
/*==============*/
{
free_dvector(&F);
free_dvector(&E);
free_ivector(&bound_type);
NEQNS= 0;
#if NodeNumberChangeENABL
NNODEmax=
#endif
NNODE= 0;
del_matinv_mem();
}


static void (* fresids)(void), (* cresids)(void);
static int Node; // basically a "local" variable to runband(), but also needed in fresidsA2/E2
		 // which were moved outside of runband() so as to speed-up the execution;
		 // switching between fresidsA2 and fresidsE2 is done in startband()

void fresidsA2(void)
/*================*/
{ register int e;
 register double aux;
 for(e=0;e<NEQNS;e++) {
	aux = ABSV(F[e]);
	if(fresidM < aux) {
		fresidM = aux;
		eqfresidM = e;
		nodefresidM = Node;
	}
	fresidT += aux;
 }
}

void fresidsA(void)
/*===============*/
{ register int e;
 for(e=0;e<NEQNS;e++) fresidT += ABSV(F[e]);
}


void fresidsE2(void)
/*================*/
{ register int e, i;
  register double aux;
 for(e=0;e<NEQNS-Nexcl;e++) {
	i=equUsed[e];
	aux = ABSV(F[i]);
	if(fresidM < aux) {
		fresidM = aux;
		eqfresidM = i;
		nodefresidM = Node;
	}
	fresidT += aux;
 }
}

void fresidsE(void)
/*===============*/
{ register int e;
 for(e=0;e<NEQNS-Nexcl;e++) fresidT += ABSV(F[equUsed[e]]);
}


void cresidsA2(void)
/*================*/
{ register int e, i;
  register double aux;
 cresidT= 0.;
 cresidM = 0.;
 for(e=0;e<NEQNS;e++) {
	for(i=e;i<ntN;i+=NEQNS) {
		aux = ABSV(xi[i]);
		cresidT += aux;
		if(cresidM<aux) {
			cresidM = aux;
			eqcresidM = e;
			nodecresidM = i;
		}
	}
 }
 nodecresidM /= NEQNS;
}

void cresidsA(void)
/*===============*/
{ register int i;
  register double aux;
 cresidT= 0.;
 cresidM = 0.;
 for(i=0;i<ntN;i++) {
	aux = ABSV(xi[i]);
	cresidT += aux;
	if(cresidM<aux) cresidM = aux;
 }
}


void cresidsE2(void)
/*================*/
{ register int e, i;
  register double aux;
 cresidT= 0.;
 cresidM = 0.;
 for(e=0;e<NEQNS-Nexcl;e++) {
	for(i=equUsed[e];i<ntN;i+=NEQNS) {
		aux = ABSV(xi[i]);
		cresidT += aux;
		if(cresidM<aux) {
			cresidM = aux;
			eqcresidM = equUsed[e];
			nodecresidM = i;
		}
	}
 }
 nodecresidM /= NEQNS;
}

void cresidsE(void)
/*===============*/
{ register int e, i;
  register double aux;
 cresidT= 0.;
 cresidM = 0.;
 for(e=0;e<NEQNS-Nexcl;e++) {
	for(i=equUsed[e];i<ntN;i+=NEQNS) {
		aux = ABSV(xi[i]);
		cresidT += aux;
		if(cresidM<aux) cresidM = aux;
	}
 }
}


/* The following 4 macros are used in the include files below: */
#define EQNO e
#define FORA for(i = e; i < ntN; i += NEQNS)
#define FORB for(i = e, node = 0; i < ntN; i += NEQNS, node++)
#define NoDe node

#define INCONSIST { PRl"Lower and upper bounds inconsistent for eq. # %d",e);\
		    exitTR(-1,msg); }


void startband(int neqns, int NJ, double*cTR)
/*=========================================*/
{
RAND_WARNING

if(NJ < 1) exitTR(-1,"Number of grid nodes must be positive!");

if((NEQNS==0)||(NNODE==0))
	cr_band_mem(neqns,NJ);
else if((NEQNS<neqns)||(NNODE<NJ)) {
	endband();
	cr_band_mem(neqns,NJ);
}
if(monitor2) {
	fresids= Nexcl ? fresidsE2 : fresidsA2;
	cresids= Nexcl ? cresidsE2 : cresidsA2;
} else {
	fresids= Nexcl ? fresidsE : fresidsA;
	cresids= Nexcl ? cresidsE : cresidsA;
}

cMax = FMax = cTmax = FTmax = 0.;
converged = -99.; /* ~ no call to runband since last startband */
c= cTR;
initNL3BANDV(&LB,&UB,F,A,B,D);

{
register int i, j, e;
register double upBnd, lowBnd, *lp, *up;

 for(e=0; e<NEQNS; e++) {
	j= (int)((UB && UB[e]) ? UB[e][0] : 0);
	if(j<0 || j>3) {
	  PRl"Upper bound type for eq. # %d has to be 0, 1, 2 or 3", e);
	  exitTR(-1,msg);
	}
	bound_type[e]= (int)((LB && LB[e]) ? LB[e][0] : 0);
	if(bound_type[e]<0 || bound_type[e]>4) {
	  PRl"Lower bound type for eq. # %d has to be 0, 1, 2, 3 or 4", e);
	  exitTR(-1,msg);
	}
	if(bound_type[e]<4) bound_type[e] += 4 * j;
	else {
	  if(LB[e][1]<0.) {
		PRl"Zero cutoff for eq. # %d must be non-negative", e);
		exitTR(-1,msg);
	  }
	  bound_type[e] = 16 + j;
	}
 }

 // Check bound consistency, if both bounds are defined, LB <= UB must hold:
 for(e=0; e<NEQNS; e++)  switch(bound_type[e]) {

#include "nl3_strt.h"

}
}

}


void changeBound(int e, char which, double *BDA)
/*============================================*/
{
register int i, j;
register double upBnd, lowBnd, *lp, *up;

 switch(which) {
	case 'l':
		if(!LB) {
			LB= pdvector(NEQNS);
			for(i=0;i<NEQNS;i++) LB[i]= NULL;
		}
		LB[e]= BDA;
		break;
	case 'u':
		if(!UB) {
			UB= pdvector(NEQNS);
			for(i=0;i<NEQNS;i++) UB[i]= NULL;
		}
		UB[e]= BDA;
		break;
 }

 j= (int)((UB && UB[e]) ? UB[e][0] : 0);
 if(j<0 || j>3) {
	PRl"Upper bound type for eq. # %d has to be 0, 1, 2 or 3", e);
	exitTR(-1,msg);
 }
 bound_type[e]= (int)((LB && LB[e]) ? LB[e][0] : 0);
 if(bound_type[e]<0 || bound_type[e]>4) {
	PRl"Lower bound type for eq. # %d has to be 0, 1, 2, 3 or 4", e);
	exitTR(-1,msg);
 }
 if(bound_type[e]<4) bound_type[e] += 4 * j;
 else {
	if(LB[e][1]<0.) {
		PRl"Zero cutoff for eq. # %d must be non-negative", e);
		exitTR(-1,msg);
	}
	bound_type[e] = 16 + j;
 }

/* Check bound consistency, if both bounds are defined, LB <= UB must hold: */
switch(bound_type[e]) {

#include "nl3_strt.h"

}
}


void ApplyBounds(double*c)
/*======================*/
{
register int e, i, node;
register double *lp, *up, upBnd, lowBnd;

/* Enforce the lower and upper bounds: */
for(e = 0; e < NEQNS; e++)  switch(bound_type[e]) {

#include "nl3_ApBo.h"


#include <time.h>

void Randomize(double Amp, double *c_o, double *c)
/*==============================================*/
{
register int e, i, node;
register double *lp, *up, upBnd, lowBnd, RaU, RaL;

srand((unsigned)time(0));
/* Randomize while enforcing the lower and upper bounds: */
for(e = 0; e < NEQNS; e++) {
 switch(bound_type[e]) {

#include "nl3_Rand.h"

}


#undef FORB
#undef NoDe
#define FORB for(i = e, Node = 0; i < ntN; i += NEQNS, Node++)
#define NoDe Node

#define INCOMPL_FRESIDM if(monitor2) PE" %d: (>=%g) >=%g %d %d\n",\
	NRiter,fresidT,fresidM,nodefresidM,eqfresidM);

double runband(double t)
/*====================*/
{
double
	*Ep, *xip;
#include "nl3_Decl1.h"

converged = -1.;

for(NRiter = 1; ; NRiter++)
{

fresidT= fresidM= 0.;

for(Node= 0;Node<NNODE;Node++)
{

{
register int i;
for(i= 0;i<nSq;i++) B[i]= A[i]= D[i]= 0.;
}
/* Here X is identical with A, and Y with D - see cr_band_mem();
   X and A, and Y and D, are never used at the same time */

evalf(Node);

/* omitting F[i]*= -1.; will just consistently reverse the sign
	of F, xi (and dC) in comparison with the original Spotnitz' nl3band: */

fresids();

if(Node==0) {
	if(matinvNMC(NEQNS,B,B) < 0) {
		PE"	at node 0 in "__FILE__"\n");
		INCOMPL_FRESIDM
		return(converged = -8.); }
	mmult(NEQNS,E,B,D);
#if !NO_X
	mmult(NEQNS,x,B,X);
#endif
	mvmult(NEQNS,xi,B,F);
	Ep = E;
	xip = xi;
#ifdef DEBUG
	show_dmatrix("B^-1",NEQNS,B,0,NEQNS-1,0,NEQNS-1," %.3g");
	show_dmatrix("E(0)",NEQNS,E,0,NEQNS-1,0,NEQNS-1," %.3g");
	show_dvector("xi(0)",xi,0,NEQNS-1);
#endif
} else if(Node<NNODE1) {
	mvmultsub(NEQNS,F,A,xip);
	mmultsub(NEQNS,B,A,Ep);
#if !NO_X
	if(Node==1) mmultsub(NEQNS,D,A,x);
#endif
	if(matinvNMC(NEQNS,B,B) < 0) {
		PE"	at node %d in "__FILE__"\n",Node);
#ifdef DEBUG
		show_dvector("F",F,0,NEQNS-1);
#endif
		INCOMPL_FRESIDM
		return(converged = -8.);
	}
	Ep += nSq;	/*  Now Ep ~ E(Node)	*/
	xip += NEQNS;	/*  Now xip ~ xi(Node)	*/
	mmult(NEQNS,Ep,B,D);
	mvmult(NEQNS,xip,B,F);
#ifdef DEBUG
	PE"node = %d\n", Node);
	show_dmatrix("E(node)",NEQNS,Ep,0,NEQNS-1,0,NEQNS-1," %.3g");
	show_dvector("xi(node)",xip,0,NEQNS-1);
#endif
} else { /* Here Node = NNODE1 = NJ-1; xip ~ xi(NJ-2), Ep ~ E(NJ-2) */
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
#if !NO_Y
	mmultsub(NEQNS,A,Y,Ep-nSq);
#endif
	mmultsub(NEQNS,B,A,Ep);
#if !NO_Y
	mvmultsub(NEQNS,F,Y,xip-NEQNS);
#endif
	mvmultsub(NEQNS,F,A,xip);
	if(matinvNMC(NEQNS,B,B) < 0) {
		PE"	at the last node %d in "__FILE__"\n",Node);
		if(monitor2) PE" %d: (%g) %g %d %d\n",
			NRiter,fresidT,fresidM,nodefresidM,eqfresidM);
		return(converged = -8.); }
	mvmult(NEQNS,xip+NEQNS,B,F);
#ifdef DEBUG
	show_dvector("xi(NJ-1)",xip+NEQNS,0,NEQNS-1);
#endif
} /* end if(Node==0) */

} /* end of  for(Node= 0;Node<NNODE;Node++) */

#ifdef DEBUG
for(Node=NNODE-2; xip > xi; xip -= NEQNS, Ep -= nSq, Node--) {
	mvmultsub(NEQNS,xip,Ep,xip+NEQNS);
	PE"back: node = %d\n", Node);
	show_dvector("xi(node)",xip,0,NEQNS-1);
}
#endif
for( ; xip > xi; xip -= NEQNS, Ep -= nSq) mvmultsub(NEQNS,xip,Ep,xip+NEQNS);

xip += NEQNS;
mvmultsub(NEQNS,xi,E,xip);
#if !NO_X
mvmultsub(NEQNS,xi,x,xip+NEQNS);
#endif
#ifdef DEBUG
show_dvector("xi(0)",xi,0,NEQNS-1);
#endif
//PE"---  c[76]=%g, xi[76]=%g, c[56]=%g, xi[56]=%g",c[76],xi[76],c[56],xi[56]);
#pragma GCC diagnostic pop

{
register int
	e, i;
register double
	aux;

#if NR_BOUNDS
register double
	upBnd, lowBnd, *lp, *up;

// Update c, and check lower and upper bounds (remember: dC = -xi):
for(e = 0; e < NEQNS; e++)  switch(bound_type[e]) {
#include "nl3_runb.h"
}

#else
// Update c only, no bound checks:
for(i=0;i<ntN;i++) c[i] -= xi[i];
#endif

// Calculate CRESID's:
cresids();

#include "nl3_conv.h"

} // End of  <Lbrace> register int e, i; register double aux ...;

} /* end of  for(NRiter = 1; ; NRiter++)  */
// The following output line is finished in transien.c:
//PE"%g %g %g %g %g", converged, fresidT1, fresidT, cresidT1, cresidT);

if(converged>0.) {
#if !NR_BOUNDS
 ApplyBounds(c);
#endif
 if(monitor2) {
	if(cresidM > cMax) { cMax = cresidM; eqcMax = eqcresidM;
			nodecMax = nodecresidM; tcMax = t; }
	if(fresidM > FMax) { FMax = fresidM; eqFMax = eqfresidM;
			nodeFMax = nodefresidM; tFMax = t; }
 } else {
	if(cresidM > cMax) { cMax = cresidM; tcMax = t; }
 }
 if(cresidT > cTmax) { cTmax = cresidT; tcTmax = t; }
 if(fresidT > FTmax) { FTmax = fresidT; tFTmax = t; }
}

return(converged);
}


#define PR (void)fprintf(out,

void report_max_NR_error(FILE *out, char comment_char)
/*==================================================*/
{
char *str = transCalc ? "Max a" : "A";
if(monitor2) {
	PR"%c Max converged NR change in c, %e, was in eq. # %d at node # %d",
			comment_char, cMax, eqcMax, nodecMax);
	if(transCalc) PR" at t = %e", tcMax);
	PR"\n%c Max residual |F|, %e, was in eq. # %d at node # %d",
			comment_char, FMax, eqFMax, nodeFMax);
	if(transCalc) PR" at t = %e", tFMax);
	PR"\n");
	cMax = FMax = 0.;
} else {
	PR"%c Max converged NR change in a c[node][eq] was %e", comment_char, cMax);
	if(transCalc) PR" at t = %e", tcMax);
	PR"\n");
	cMax = 0.;
}
PR"%c %sverage converged NR change in c per node and equation was %e",
			comment_char, str, cTmax/ntN);
if(transCalc) PR" at t = %e", tcTmax);
PR"\n%c %sverage residual |F| per node and equation was %e",
			comment_char, str, FTmax/ntN);
if(transCalc) PR" at t = %e", tFTmax);
PR"\n%c\n", comment_char);
(void)fflush(out);
cTmax = FTmax = 0.;
}

void report_last_NR_error(FILE *out, char comm)
/*===========================================*/
{
if(converged == -99.) PR"%c No call to runband() made yet\n", comm);
else if(converged == -8.)
 PR"%c matinvNMC() [lufac()] failed on last call to runband()\n", comm);
else {
 if(monitor2) {
	PR"%c Last fresidM = %g in eq. # %d at node # %d;\n",
		comm, fresidM, eqfresidM, nodefresidM);
	PR"%c cresidM = %g (|delta_c|) in eq. # %d at node # %d;\n",
		comm, cresidM, eqcresidM, nodecresidM);
 }
 PR"%c fresidT = (sum of |F|) = %g; cresidT = (sum of |delta_c|) = %g\n",
		comm, fresidT, cresidT);
}
(void)fflush(out);
}
