/* File signals.c; TRANSIENT V4.3; 3 Nov 2021 */
/* (C) 1999,2009,2011,2015,2021 Miroslav Kolar */

/*		TRANSIENT signal handling module

  The purpose of this file is to catch as many as possible of the
  signals causing job termination, or signals generated by error
  conditions, and to output any results not yet written out that are
  available, and to complete the 'Output'.stat job statistics file.

  It seems that each compiler defines a different number of signals,
  and maps them in a different way to numbers (see signal.h, or
  or sys/signal.h for your compiler).	 This module is an attempt
  to achieve the above stated purpose in a platform independent way.
  It may need more testing, so feel free to modify this file if
  inappropriate signals are caught, etc. In the least, you should
  keep catching SIGINT, which always seems to be assigned number 2,
  which is sent when the user interrupts the execution, e.g., by
  pressing Ctrl+C in UNIX, and SIGTERM (kill -15) sent ferom shell.
  And you would usually not want to catch SIGTSTP,
  which is sent when you press Ctrl+Z in UNIX to temporarily
  pause the execution of a program, and later on to resume it by '%%'
  or 'kill -CONT'. That's why SIGTSTP is among the signals for which
  no handler is set in signlSetup() below.  The equivalent 'kill -STOP'
  (SIGSTOP) is never caught as well as SIGKILL, and so these two need
  not be explicitly excluded in signlSetup().
  On the other hand, signals like SIGWINCH (window size change), etc.,
  should be ignored, and so must be skipped in signlSetup().
  If you have problems, see note at the end of this file.

  signlSetup() is called at the start of execution in main() (in
  transien.c), to set up the signal catching functions for individual
  signals. For most of them this function is catchSig() defined below,
  which generates a message describing the cause of job termination,
  and then calls exitTR() of InOut.c with exit status 100+signal_#,
  which does all the relevant output.
*/ 

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

/* NSIG is the number of signals plus 1: */
#ifndef NSIG
#define NSIG 35
#endif

/* Defined in InOut.c: */
extern void exitTR(int status, char *descr) __attribute__ ((noreturn));
extern void exitTRsimple(int) __attribute__ ((noreturn));
extern char msg[];
extern const int msgLength;


/* Defined below: */
void catchSig(int);
#ifdef SIGALRM
void catchAlrm(int);
#endif


int signlSetup(void)
/*=================*/
{
register int i;

for(i=0;i<NSIG;i++) {
#ifdef SIGTSTP
 if(i == SIGTSTP) continue;
#endif
#ifdef SIGCONT
 if(i == SIGCONT) continue;
#endif
#ifdef SIGWINCH
 if(i == SIGWINCH) continue;
#endif
#ifdef _SIGWINCH
 if(i == _SIGWINCH) continue;
#endif
#ifdef SIGWINDOW
 if(i == SIGWINDOW) continue;
#endif
#ifdef SIGINFO
 if(i == SIGINFO) continue;
#endif
#ifdef SIGALRM
 if(i == SIGALRM) {
	signal(i,catchAlrm);
	continue;
 }
#endif
 signal(i,catchSig);
}
return NSIG;
}


#ifdef SIGALRM
void catchAlrm(/*@unused@*/int signl)
/*=================================*/
/* This function just makes sure that the execution continues
   unaffected when clock alarm is sent (and sounds the bell).	*/
{ fprintf(stderr,"%c%c\n",(char)7,(char)7); return; }
#endif


#ifdef SIGSEGV
static int SegmCount=0;
#endif
 
#define PRl (void)snprintf(msg+le,(size_t)(msgLength-le),

void catchSig(int signl)
/*====================*/
{
int le;
le= sprintf(msg,"(%d) ",signl);
if(signl == SIGINT) PRl"Interrupted by user (Ctrl+C)");
else if(signl == SIGILL) PRl"Illegal instruction");
else if(signl == SIGFPE) PRl"Floating point exception");
else if(signl == SIGTERM) PRl"Killed from shell (SIGTERM)");
else if(signl == SIGABRT) {
	PRl"Abnormal termination");
	exitTRsimple(100+signl);
}
#ifdef SIGBREAK
else if(signl == SIGBREAK) PRl"Ctrl+Break interrupt");
#endif
#ifdef SIGHUP
else if(signl == SIGHUP) PRl"Hangup detected");
#endif
#ifdef SIGQUIT
else if(signl == SIGQUIT) PRl"QUIT");
#endif
#ifdef SIGTRAP
else if(signl == SIGTRAP) PRl"Trace trap");
#endif
#ifdef SIGIOT
else if(signl == SIGIOT) PRl"I/O trap");
#endif
#ifdef SIGEMT
else if(signl == SIGEMT) PRl"Emulator trap");
#endif
#ifdef SIGBUS
else if(signl == SIGBUS) PRl"Bus error");
#endif
#ifdef SIGSEGV
else if(signl == SIGSEGV) {
	PRl"Segmentation violation/illegal memory access/stack overflow");
	SegmCount++;
	if(SegmCount>1) {
		if(SegmCount==2) exitTRsimple(100+signl);
		else {
			(void)fprintf(stderr,"\n%d: %s\n", 100+signl, msg);
			_Exit(100+signl);
		}
	}
}
#endif
#ifdef SIGSYS
else if(signl == SIGSYS) PRl"Bad argument to system call");
#endif
#ifdef SIGPIPE
else if(signl == SIGPIPE) PRl"No one to read the pipe");
#endif
#ifdef SIGURG
else if(signl == SIGURG) PRl"Urgent I/O condition");
#endif
else PRl"Terminated upon receiving signal %d", signl);
exitTR(100+signl,msg);
}

/* If you do not have signals at all, or have difficulties to make this module
work, or want to temporarily eliminate signal handling,
remove #include <signal.h>, and replace the above function definitions
by the following do-nothing functions:

void signlSetup(void) {}

void catchSig(int signl) {}
*/
