/* File transien.c; TRANSIENT V4.3; 30 Oct 2023			*/
/* Copyright (C) 1995-2012,2015-19,2021,2023 Miroslav Kolar	*/

/* See  defs.h  for how to enable/disable various features.		*/

/* Transient solution of a system of nonlinear partial differential 	*/
/*		     equations of parabolic type 			*/
/* ================================================================	*/
/* Adaptive time step algorithm linked to the maximum change during 	*/
/* one time step in all concentrations at all grid points,		*/
/* and/or to the number of Newton-Raphson iterations in each step.	*/

#include "incl_pri.h"
#include <unistd.h>

#define Pd (void)fprintf(diag_out,
#define Fd (void)fflush(diag_out);

// Defined in InOut.c:
extern void ptitle(void);
extern void startClock(void);
extern void startTransClock(void);
extern int read_input(int argc, char **argv);
extern void bad_input(int Xplain);
extern void pr_par(int);
extern void change_file(int sepProf);
extern void open_output(void);
extern void echo_init_profile(void);
extern void exitTR(int stat, char *descr) __attribute__ ((noreturn));
extern long int CPUtime(double t, long sCount, long rCount, long iCount);
extern void steadyDiags(int attempt, int iter);

// Defined in signals.c:
extern int signlSetup(void);

// Defined in nl3bandv.c or in nl3band1.c:
extern void startband(int neqns, int Nnodes, double*c);
extern double runband(double t);
extern void changeNoNodes(int nnodes);
extern void ApplyBounds(double*c);
extern void Randomize(double Amplitude, double *c_o, double *c);
extern void report_last_NR_error(FILE *out, char comment_char);

// Defined in nl3_Decl.h:
extern void reportDEBUG(void);

// Defined in dynarr.c:
extern double *dvector(int n);

// Defined in a problem definition module (PDM) - a module corresponding
//   to each individual model (set of reaction-diffusion equations):
extern int get_neqns_etc(void);
extern void generate_grid(void);
extern void read_profiles(char *filename);
extern void init_tStep_for_evalf(void);
extern void auxiliary_parameters_for_evalf(void);
extern void pr_profiles(int MaxIter);
extern void init_Other(void);
extern int update_Other(void);
extern void pr_Other(int only_Other);
/* After V2.5, pointer to evalf() passed directly from a PDM to NL3BANDV: */
extern void Initialize(double *c, double *c_o);
// Two more imports from PDM are declared in nl3_Decl.h:
// extern void evalf(int), initNL3BANDV(....)

// Defined in transien.h, which is included in PDM:
extern void check_PAR_format(void);

/* Input parameters - Control:
   -------------------------- */
#define IN_PAR(name,type,initValue,oFrmt_u,expl) type name=initValue;
#include "CTRLpar.h"
#undef IN_PAR

extern char
	*init_from_file,
	*cDiffMfExcl;
extern short int
	init_from_file_dyn,
	cDiffMfExcl_dyn;
extern FILE *stat_out;
extern const int msgLength;
extern char	 msg[];

/* Other global variables: */
char
	*versn = "\n\
  .----------         p o w e r e d    b y        ----------.\n\
  | TRANSIENT V4.3 nonlinear parabolic diff equation solver |\n\
  '----------    http://transient.mkolar.org/    -----------'\n\n";

char *title = NULL,
     *time_units = NULL;
double
	t, tStep, tStepInv, tStepLast, ttStep, ttOld,
	cDiffM, cAVmaxOld=0., cAVmaxExOld=0., cAVmax, cAVmaxEx;
int
	neqns, nnodes, monitor2, mPause= 1, transCalc, NRiter;
int unsigned
	do_Oth_Prof= 0u; /* used in a transient run to prevent
		(i) the duplication of output in exitTR() when
		    terminating the run for whatever reason;
		(ii) keeping track of the time corresponding to the
		    solution available for output.
	Four lowest (right-most) bits of do_Oth_Prof are used for this purpose.
	Bits 1 and 3 are associated with the "Other" values (calculated/outputted
	by update_Other/pr_Other.  Bits 2 and 4 pertain to the c_o values.
	Bits 1 and 2 keep track of the availability, 3 and 4 of time.
		Bit 1: value 0 - nothing new to output
			     1 - new "Other" values ready for output
		Bit 2: value 0 - nothing new to output
			     1 - new c_o values ready for output
		Bit 3: value 0 - "Other" quantities ready for output correspond to
				 the current time t
			     1 - they still correspond to the previous time t-tStep
		Bit 4: value 0 - c_o ready for output correspond to t
			     1 - they still correspond to t-tStep	*/
FILE
	*diag_out= NULL, *par_out= NULL, *out= NULL, *timeDep_out= NULL;
double
	(* cDifMxf)(double),
	(* cDiffMf)(void);
void
	(*updtDiag)(double);
int
	*equExcl, Nexcl= 0,
	*equUsed;
long strict1failCount= 0;

double
	tSign, ttEnd;	// will be set up in InOut.c
static double
	cDiffM_sum= 0.,
	*c, *c_o;
static int
	NtN, maxIterDone;
static long
	stepCounter= 0,
	runbandCounter= 0,
	iterCounter= 0;

/* To be used with the diagnostic output file: */
static double
	cDiffMmulti, cDiffMceil, tStepFloor, tStepCeil, cDiffMEx;
	// Here cDiffMmulti is a multipurpose (two-purpose) variable:
	// if Nexcl=0, it acts like 'cDiffMfloor' (stores the minimum of cDiffM),
	// otherwise like 'cDiffMceilEx' (stores the maximum of the excluded c).
static int
	allIter, allIterFloor, allIterCeil, diagCounter, diagFlag;


void DiagEntry_at_t(void)
/*=====================*/
{ /* To be used in InOut.c; and here when strict=1 fails */
 if(diagCounter) Pd"%#15.10g  %10.3e %10.3e  %10.3e %10.3e %4d %4d %6d\n", t, tStepFloor,
  tStepCeil, cDiffMmulti, cDiffMceil, allIterFloor, allIterCeil, diagCounter);
}


void transienCompiled(void)
/*=======================*/
{
Pp"%c transien.c with: TRUE_FABS=%d, ", CmntC, TRUE_FABS);
if(steady) Pp"attemptMaxSTEADY=%d", attemptMaxSTEADY);
if(steady==2) Pp", ");
if(!(steady&1)) {
#if preV4_3
	Pp"preV4_3= %d, RESETtStepAT99=%d, ", preV4_3, RESETtStepAT99);
#endif
	Pp"attemptMAX=%d", attemptMAX);
}
Pp"\n");
}


double cDiffMf0(void)
/*=================*/
{
 register double R, cDiffM= 0.;
 register int j;

 cAVmax= 0.;
 for(j=0; j<NtN; j++) {
#if TRUE_FABS
	R = fabs(c[j] - c_o[j]);
#else
	R = c[j] - c_o[j];
	if(R<0.) R = -R;
#endif
	if(R > cDiffM) cDiffM = R;
	R = ABSV(c[j]);
	if(R > cAVmax) cAVmax = R;
 }
 cDiffM /= (cAVmax>cAVmaxOld ? cAVmax : cAVmaxOld);
 return(cDiffM);
}


double cDiffMf3(void)
/*=================*/
{
 register double R, cDiffM= 0.;
 register int i, j, jmax= 0;

 cAVmax= 0.;
 for(i=0; i<neqns; i++) {
	for(j=i; j<NtN; j += neqns) {
#if TRUE_FABS
		R = fabs(c[j] - c_o[j]);
#else
		R = c[j] - c_o[j];
		if(R<0.) R = -R;
#endif
		if(R > cDiffM) { cDiffM = R; jmax= j; }
		R = ABSV(c[j]);
		if(R > cAVmax) cAVmax = R;
	}
 }
 j= jmax/neqns;
 i= jmax - j*neqns;
 cDiffM /= (cAVmax>cAVmaxOld ? cAVmax : cAVmaxOld);
 PE"cDiffM=%g in node %d, eq %d where c=%.8g, c_o=%.8g; |c|max=%.8g, |c_o|max=%.8g\n",
	cDiffM,j,i,c[jmax],c_o[jmax],cAVmax,cAVmaxOld);
 return(cDiffM);
}


double cDiffMf0excl(void)
/*=====================*/
{
 register double R, cDiffM= 0.;
 register int i, j;

 cAVmax= 0.; cDiffMEx= 0.; cAVmaxEx= 0.;
 for(i=0; i<neqns-Nexcl; i++) {
	for(j=equUsed[i]; j<NtN; j += neqns) {
#if TRUE_FABS
		R = fabs(c[j] - c_o[j]);
#else
		R = c[j] - c_o[j];
		if(R<0.) R = -R;
#endif
		if(R > cDiffM) cDiffM = R;
		R = ABSV(c[j]);
		if(R > cAVmax) cAVmax = R;
	}
 }
 for(i=0; i<Nexcl; i++) {
	for(j=equExcl[i]; j<NtN; j += neqns) {
#if TRUE_FABS
		R = fabs(c[j] - c_o[j]);
#else
		R = c[j] - c_o[j];
		if(R<0.) R = -R;
#endif
		if(R > cDiffMEx) cDiffMEx = R;
		R = ABSV(c[j]);
		if(R > cAVmaxEx) cAVmaxEx = R;
	}
 }
 cDiffM /= (cAVmax>cAVmaxOld ? cAVmax : cAVmaxOld);
 if(cAVmaxEx>cAVmaxExOld) cDiffMEx/=cAVmaxEx;
 else if(cAVmaxExOld>0.) cDiffMEx/=cAVmaxExOld;
 return(cDiffM);
}


double cDiffMf3excl(void)
/*=====================*/
{
 register double R, cDiffM= 0.;
 register int i, j, jmax= 0, jmaxE= 0;

 cAVmax= 0.; cDiffMEx= 0.; cAVmaxEx= 0.;
 for(i=0; i<neqns-Nexcl; i++) {
	for(j=equUsed[i]; j<NtN; j += neqns) {
#if TRUE_FABS
		R = fabs(c[j] - c_o[j]);
#else
		R = c[j] - c_o[j];
		if(R<0.) R = -R;
#endif
		if(R > cDiffM) { cDiffM = R; jmax= j; }
		R = ABSV(c[j]);
		if(R > cAVmax) cAVmax = R;
	}
 }
 for(i=0; i<Nexcl; i++) {
	for(j=equExcl[i]; j<NtN; j += neqns) {
#if TRUE_FABS
		R = fabs(c[j] - c_o[j]);
#else
		R = c[j] - c_o[j];
		if(R<0.) R = -R;
#endif
		if(R > cDiffMEx) { cDiffMEx = R; jmaxE= j; }
		R = ABSV(c[j]);
		if(R > cAVmaxEx) cAVmaxEx = R;
	}
 }
 j= jmax/neqns;
 i= jmax - j*neqns;
 cDiffM /= (cAVmax>cAVmaxOld ? cAVmax : cAVmaxOld);
 PE"cDiffM(used)=%g in node %d, eq %d where c=%.8g, c_o=%.8g; |cU|max=%.8g, |c_oU|max=%.8g\n",
	cDiffM,j,i,c[jmax],c_o[jmax],cAVmax,cAVmaxOld);
 j= jmaxE/neqns;
 i= jmaxE - j*neqns;
 if(cAVmaxEx>cAVmaxExOld) cDiffMEx/=cAVmaxEx;
 else if(cAVmaxExOld>0.) cDiffMEx/=cAVmaxExOld;
 PE"cDiffMexcl=%g in node %d, eq %d where c=%.8g c_o=%.8g; |cE|max=%.8g, |c_oE|max=%.8g\n",
	cDiffMEx,j,i,c[jmaxE],c_o[jmaxE],cAVmaxEx,cAVmaxExOld);
 return(cDiffM);
}


void updtDiagEx(double cDiffM)
/*==========================*/
{
 cAVmaxOld=cAVmax;
 cAVmaxExOld=cAVmaxEx;
 if(diagCounter) {
	if(allIterFloor>allIter) allIterFloor=allIter;
	else if(allIterCeil<allIter) allIterCeil=allIter;
	if(cDiffMmulti<cDiffMEx) cDiffMmulti= cDiffMEx;
	if(cDiffMceil<cDiffM) cDiffMceil= cDiffM;
	if(tStepFloor>tStep) tStepFloor= tStep;
	else if(tStepCeil<tStep) tStepCeil= tStep;
 } else {
	allIterFloor= allIterCeil= allIter;
	cDiffMmulti= cDiffMEx;
	cDiffMceil= cDiffM;
	tStepFloor= tStepCeil= tStep;
 }
}


void updtDiagAllU(double cDiffM)
/*============================*/
{
 cAVmaxOld=cAVmax;
 if(diagCounter) {
	if(allIterFloor>allIter) allIterFloor=allIter;
	else if(allIterCeil<allIter) allIterCeil=allIter;
	if(cDiffMmulti>cDiffM) cDiffMmulti= cDiffM;
	else if(cDiffMceil<cDiffM) cDiffMceil= cDiffM;
	if(tStepFloor>tStep) tStepFloor= tStep;
	else if(tStepCeil<tStep) tStepCeil= tStep;
 } else {
	allIterFloor= allIterCeil= allIter;
	cDiffMmulti= cDiffMceil= cDiffM;
	tStepFloor= tStepCeil= tStep;
 }
}


const char* possibleSS =
 "A steady state possibly reached: negligible change since the previous ";

void out_profiles(int NewFile)
/*==========================*/
{
 int NOtSteps, SSreached;

 if(do_Oth_Prof & 2u) {
	if(do_Oth_Prof & 8u) t -= tStep; /* t for c_o */
	NOtSteps= (int)CPUtime(t,stepCounter,runbandCounter,iterCounter);
	SSreached= cDiffM_sum<NOtSteps*1e-14;
#if PDMtrace
	PE"... pr_profiles (out_profiles) \n");
#endif
	do_Oth_Prof &= 29u; /* to set the 2nd bit to zero */
	if(sepProf>=0) {
		if(!sepProf && (t > tOutInit || steady > 1)) PR"\n");
		if(SSreached) PR"%c %sprofile file!\n%c\n", CmntC, possibleSS, CmntC);
		pr_profiles(maxIterDone);
		if(monitor>0) { if(sepProf) PE" <out %d>\n",sepProf); else PE" <out>\n"); }
		if(NewFile && sepProf) change_file(++sepProf);
	}
	if(do_Oth_Prof & 8u) t += tStep;
	if(monitor>0) PE"Sum of the last %d cDiffM: %g\n", NOtSteps, cDiffM_sum);
	if(SSreached) {
		Ps"\t%stime\n", possibleSS);
		if(monitor<0) PE"\n");
		PE"%sprofile output\n", possibleSS);
	}
	cDiffM_sum=0.;
 }
}


void out_Other(int onOth)
/*=====================*/
{
 if(do_Oth_Prof & 1u) {
	if(do_Oth_Prof & 4u) t -= tStep; /* t for pr_Other values */
#if PDMtrace
	PE"... pr_Other (out_Other)\n");
#endif
	do_Oth_Prof &= 30u; /* to set the 1st bit to zero */
	/* do_Oth_Prof must be reset before pr_Other is actually called
	   to avoid infinite loop when exitTR is called from pr_Other */
	pr_Other(onOth);
	if(do_Oth_Prof & 4u) t += tStep;
	if(monitor>0) PE" <outOTH>\n");
	else PE"\n%.9g %s ", t, time_units);
 }
}


#if PATHtoITSELF
char *invokedName;

char* pathToThisExecutable(void)
/*==============================
Find the relative or absolute full path to a file (executable)
called Name somewhere in the current path, or in ((relatively to) "."
This function was added around 2008; written only for UN*X/Linux,
and not expected to work in VMS. Is somebody still using
(Open)VMS for numerical calculations today?
Should I remove all historical VMS references in the TRANSIENT code,
or make it all work in VMS again? */
{
 char *full;
 int l;
 // For Linux || FreeBSD || Solaris:
 if( (l=readlink("/proc/self/exe", msg, msgLength))>0 ||
     (l=readlink("/proc/curproc/file", msg, msgLength))>0 ||
     (l=readlink("/proc/self/path/a.out", msg, msgLength))>0 ) {
	full = makestr(l);
	if(l>=msgLength) exitTR(-3,
 "In function pathToThisExecutable() in transien.c, msgLength needs to be increased");
	return strcpy(full,msg);
 }
 // If /proc not available:
 if(invokedName[0]=='/') {
		full = makestr(strlen(invokedName));
		return strcpy(full,invokedName);
 }
 if(fopen(invokedName,"r")) { // exists in (relatively to) the current directory
		char *pwd = getenv("PWD");
		full = makestr(strlen(invokedName)+strlen(pwd)+1);
		(void)strcpy(full,pwd);
		if(strcmp(pwd,"/")) (void)strcat(full,"/");
		(void)strcat(full,invokedName);
		return full;
 } else { // is somewhere else in the path
		char	*p1, *p2,
			*path = getenv("PATH");
		p1= path;
		while((p2=strchr(p1,':'))) {
			l= (int)(p2-p1);
			full = makestr(strlen(invokedName)+l+1);
			(void)memcpy(full,p1,(size_t)l);
			full[l]='\0';
			if(strcmp(full,".")) {
				if(strcmp(full,"/")) (void)strcat(full,"/");
				(void)strcat(full,invokedName);
				if(fopen(full,"r")) return full;
			}
			free_str(1,&full);
			p1=p2+1;
		}
		// The last or single member of the path remains:
		full = makestr(strlen(invokedName)+strlen(p1)+1);
		(void)strcpy(full,p1);
		if(strcmp(full,".")) {
			if(strcmp(full,"/")) (void)strcat(full,"/");
			(void)strcat(full,invokedName);
			if(fopen(full,"r")) return full;
		}
		free_str(1,&full);
 }
 // It should never get here:
 return "Failed to find";
}
#endif


#define PRl  (void)snprintf(msg,(size_t)msgLength,
#define PRlc lg=strlen(msg); (void)snprintf(msg+lg,(size_t)(msgLength-lg),

static void wrong_cDfMfExcl(char *p, int i)
/*=======================================*/
// i - last successfully read-in equation index (counting from 1); 0 - none yet
// p pointer to the first incorrect/missing char in the input string
{
int lg;
PRl"  Illegal cDiffMfExcl=%s  :  neqns=%d; #e=%d; ", cDiffMfExcl, neqns, Nexcl);
if(i>0) { PRlc"e%d=%d; ", i, equExcl[i-1]); }
if(*p==0) {
	if(i==Nexcl) {PRlc"missing the end double quote\n");}
	else if(!i) {PRlc"missing excluded equation ind%s\n", Nexcl==1?"ex":"ices");}
	else {PRlc"missing %d more excluded equation ind%s\n",Nexcl-i,Nexcl-i==1?"ex":"ices");}
} else {
	int val= atoi(p);
	if(val || *p=='0' || *p=='-' || *p=='+') { PRlc"illegal value '%d'", val); }
	else { PRlc"illegal char '%c'", *p); }
	PRlc" at position %d\n", (int)(p-cDiffMfExcl)+1);
}
PE"\n%s", msg); Ps"%s", msg);
PRl"\tThe cDiffMfExcl list must have the form \"#e:e1,e2,...,e'#e'\",\n\
\twhere 0 < #e < neqns; 0 <= e1 < e2 < ... < neqns.\n");
PE"%s", msg); Ps"%s\n", msg);
bad_input(-2);
}

#define CDIFFMex_REP(WHATw,ANDw)	PE#WHATw" = (MAX over all j and e=%d",equExcl[0]);\
        for(k=1; k<Nexcl; k++) PE",%d",equExcl[k]);\
        PE" of |c[j][e]"#ANDw"c_o[j][e]|)");
#define CDIFFM_REP(WHATw,ESPECw,ANDw)	\
PE#WHATw" = (MAX over all j and "#ESPECw"e of |c[j][e]"#ANDw"c_o[j][e]|)");


int main(int argc, char **argv)
/*===========================*/
{
double	cDiffMprev, ConvCode,
	Scale= 1., ttOut, DttOut, tt,
	ttStepPrev, ttStepLast;
int	strictM1, strict1,
	strict01,
	attempt, k,
#if NodeNumberChangeENABL
	oldNnodes,
#endif
	exactOutOTH;
register int i;

#if PDMtrace
PE"Starting TRANSIENT\n");
#endif

#if PATHtoITSELF
invokedName= *argv;
#endif

check_PAR_format();

#if PDMtrace
PE" After TRANSIENT's check_PAR_format\n");
#endif

k=signlSetup();

#if PDMtrace || defined(PRINT_EXIT_CODES_REMINDER)
PE"Reminder: ranges of termination codes available to a PDM author:\n\
\t- to use directly in exitTR(): 1 - 100 and %d - 1000\n\
\t- when returning negative return code from update_Other(): 1001 - %s\n", 100+k,
#if defined(MSDOS) || defined(__BORLANDC__) || defined(_MSC_VER) || \
defined(__MINGW32__) || defined(__MINGW64__)
"infinity"
#else
"∞"
#endif
);
#endif
startClock();

#if PDMtrace
PE"... get_neqns_etc\n");
#endif
neqns = get_neqns_etc();
if(title == NULL) title = "                       Unnamed PDM";
if(time_units == NULL) time_units = "sec?";
ptitle();

if(steady != 1) PE" ..............................................................\n\
 | Verbosity level of this diagnostic output is controlled by |\n\
 | the value of the input parameter 'monitor'.   Default: %2d  |\n\
 | All  available  values:     -1, 0, 1, 2, 3, 10, 20, and 30 |\n\
 ^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~^~\n", monitor);

{ int wrong= 0;
 if(read_input(argc, argv)) wrong= 1;

 if(steady != 1) PE" ......................\n\
 | Using monitor = %2d |\n\
 ^~^~^~^~^~^~^~^~^~^~^~\n", monitor);

 if(cDiffMfExcl) {
	register char *p= cDiffMfExcl;
	char *end;
	int n0= 0; // isQuoted= 0;
	k=0;
	if(*p=='"') p++; // isQuoted= 1;
	if(!(Nexcl= (int)strtol(p,&end,10))) wrong_cDfMfExcl(p,0);
	if(Nexcl<1||Nexcl>=neqns) wrong_cDfMfExcl(p,0);
	if(*end==':') p=end+1; else wrong_cDfMfExcl(end,0);
	k= (int)strtol(p,&end,10);
	if(end==p || k<0 || k>=neqns) wrong_cDfMfExcl(p,0);
	equExcl = ivector(Nexcl);
	equExcl[0] = k;
	if(Nexcl>1) {
		for(i=1;i<Nexcl;i++) {
			if(*end==',') p=end+1; else wrong_cDfMfExcl(end,i);
			k= (int)strtol(p,&end,10);
			if(k>=neqns||k<=equExcl[i-1]) wrong_cDfMfExcl(p,i);
			equExcl[i] = k;
		}
	} else i= 0;
	// Must be end of string, possibly preceded by a double quote:
	/* But, let's not enforce it by uncommenting the next two lines,
	 * and allow any trailing characters in input.
	if(isQuoted) { if(*end=='"') end++; else wrong_cDfMfExcl(end,i); }
	if(*end!=0) wrong_cDfMfExcl(end,i);
	*/
	equUsed = ivector(neqns-Nexcl);
	for(k=0; k<neqns; k++) {
		for(i=0; i<Nexcl; i++) if(k==equExcl[i]) break;
		if(i==Nexcl) equUsed[n0++] = k;
	}
	free_str(cDiffMfExcl_dyn,&cDiffMfExcl);
#if PDMtrace
	PE"Nexcl=%d\n", Nexcl);
	PE"Used eq: ");
	for(i=0;i<neqns-Nexcl;i++) PE" %d", equUsed[i]);
	PE":: %d\n",equUsed[i]);
#endif
 }
 if(wrong) bad_input(-1);
}

// Initialize() may need t; so may echo_init_profile():
t = tStart;
if(steady == 1) {
	// steady-state solution only
	if(sepProf>0) sepProf = 0;
} else {
	strictM1 = strict == -1;
	strict01 = !strictM1;
	strict1 = strict == 1;
}

open_output();
pr_par(-1);
if(steady) {
       	// tt will store true Tau for transient solution:
	if(steady == 2) tt= Tau;
	Tau = 0.;
}
if(monitor>9) { mPause= 0; monitor /= 10; }
monitor2 = monitor>1 ? monitor : steady;

#if PDMtrace
PE"... generate_grid\n");
#endif
generate_grid();
#if NodeNumberChangeENABL
oldNnodes= nnodes;
#endif

NtN = nnodes * neqns;
i = steady & 1 ? NtN : NtN + NtN;
c = dvector(i);
c_o = c + (i - NtN);
 /* For steady = 1 (or 3,5,... for some future use?), we need
    only one concentration array */

#if PDMtrace
PE"... Initialize\n");
#endif
Initialize(c, c_o);
if(init_from_file) {
#if PDMtrace
	PE"... read_profiles\n");
#endif
	read_profiles(init_from_file);
}
if(echoInit) echo_init_profile();

#if PDMtrace
PE"... auxiliary_parameters_for_evalf\n");
#endif
auxiliary_parameters_for_evalf();

#if PDMtrace
PE"... initNL3BANDV (from startband)\n");
#endif
startband(neqns,nnodes,c);
#if PDMtrace
PE"... startband finished\n");
#endif
ApplyBounds(c);
if(c!=c_o) ApplyBounds(c_o);

if(steady) {
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
	double *c_o_orig, minus8failures;
#pragma GCC diagnostic pop
	transCalc=0;
	tStepInv = 0.;
	if(steady == 2) c_o_orig = c_o;
	c_o = c; /* This changes c_o only here, not in a PDM!!!! */
#if PDMtrace
	PE"... init_Other (steady)\n");
#endif
	init_Other();
	if(init_from_file && steady==1 && init_from_file_dyn) free_str(1,&init_from_file);
#if NodeNumberChangeENABL
	if(nnodes!=oldNnodes) { // can be changed in a PDM in init_Other
		NtN= nnodes*neqns;
		oldNnodes= nnodes;
		changeNoNodes(nnodes);
	}
#endif
	attempt = 1;
	minus8failures = 0; // it is used to track ConvCode=-8 failures
    Try_steady_again:
#if PDMtrace
	PE"... evalf (repeatedly from runband)\n");
#endif
	PE"NRiter: (fresidT) fresidM node# eq#  (cresidT) cresidM node# eq#\n\
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
	ConvCode= runband(0.0);
	if(ConvCode<0. || ConvCode>10.) {
	  PE" This was attempt # %d:", attempt);
	  if(ConvCode==-8.) {
		PE" Inversion failed (-8):");
		minus8failures++;
	  } else {
		minus8failures = 0;
		if(ConvCode==-9.) PE" cresidT or fresidT not finite (-9):");
		else PE" No or bad NR convergence (ConvCode=%g):", ConvCode);
		FE
	  }
	  report_last_NR_error(stderr, ' ');
	  if((ConvCode!=-8.||minus8failures<2) && attempt<attemptMaxSTEADY) {
		attempt++;
		PE"\n Randomizing solution estimate\n"); FE
		{ register double a, ampl;
		  for(ampl=0.,i=0; i<NtN; i++) {
			a=ABSV(c[i]);
			if(a>ampl) ampl=a;
		  }
		  Randomize(1e-3*ampl, c, c);
		}
		goto Try_steady_again;
	  }
	  if(ConvCode == -8.) {
		PR"\n%c %d attempts done; inversion failed\n\n",CmntC,attempt);
		exitTR(-8,"Inversion failed in runband/lufac");
	  } else if(ConvCode == -9.) {
		PR"\n%c cresidT or fresidT not finite\n\n",CmntC);
		exitTR(-9,"cresidT or fresidT not finite");
	  } else {
		PR"\n%cConvergence not achieved in %d attempts\n\n", CmntC, attempt);
		report_last_NR_error(out, CmntC);
	  	exitTR(-4,"Convergence not achieved;\n\
		  change the initial guess for the steady-state solution");
	  }
	}
	if(ConvCode>1.) PE"ConvCode=%g: Check the quality of the solution\n", ConvCode);
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
	if(steady == 2) {
		c_o = c_o_orig;
		//pr_...() use PDM's c_o:
		for(i=0; i<NtN; i++) c_o[i] = c[i];
	}
#pragma GCC diagnostic pop
#if PDMtrace
	PE"... pr_profiles (steady)\n");
#endif
	pr_profiles(NRiter);
#if PDMtrace
	PE"... update_Other (steady)\n");
#endif
	(void)update_Other();
#if PDMtrace
	PE"... pr_Other (steady)\n");
#endif
	pr_Other(0);
	steadyDiags(attempt, NRiter);
	if(steady == 1) exitTR(0,"Steady-state run completed successfully");
	else {
		PE"\n    ---   END OF STEADY-STATE CALCULATION   ---\n\n\
    Start of the transient phase for steady=2 :\n\n");
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
		Tau = tt;
#pragma GCC diagnostic pop
#if PDMtrace
		PE"... auxiliary_parameters_for_evalf\n");
#endif
		auxiliary_parameters_for_evalf();
		if(sepProf==1) {
			sepProf=2;
			change_file(sepProf);
		}
	}
} /* End of: if(steady) */

#if NodeNumberChangeENABL
if(nnodes!=oldNnodes) exitTR(-6,
	"No. of nodes cannot be changed immediately after the steady-state calculation");
#endif

/* Transient solution starts here */
monitor2 = monitor>1 ? monitor : 0;
if(nOutOTH<0) {
	nOutOTH= -nOutOTH;
	exactOutOTH= 0;
} else exactOutOTH= 1;
transCalc=1; /* Used in runband() */
if(Nexcl) {
	cDiffMf = monitor>2 ? cDiffMf3excl : cDiffMf0excl;
	updtDiag = updtDiagEx;
} else {
	cDiffMf = monitor>2 ? cDiffMf3 : cDiffMf0;
	updtDiag = updtDiagAllU;
}

/*
 ttOld is the transformed time, starting from 0, always positive.
 tt is the (new) transformed time at the end of the current time step.
 t is the (new) real time: t = sign(tEnd-tStart) * tt + tStart.
	When the current time step successfully completes, t becomes the
	current time, and corresponds to the current solution c_o in the
	output procedures.
 ttStep is always positive.
	evalf() always gets the real t and tStepInv.
*/
tOutInit = tSign * (tOutInit - tStart);
//t = tStart;  - see above
ttOld = 0.;
tt = ttStep = tStepInit;
tStep = tSign * ttStep;
DttOut = tOutInit / nOutOTH;	/* = interval between OTH output times	*/
ttOut = 0.;			/* = the next OTH output time		*/
maxIterDone = 0;
ttStepPrev = tStepInit;
do_Oth_Prof = 3u; /* This sets bits 1 and 2 to 1, to output the initial values if the run
	is aborted before the very 1st step successfully done (can be useful for
	debugging; echoInit=1 causes the input of initial c_o values for any run).
       	It makes sense to make do_Oth_Prof nonzero only after the output files are opened.  */

{ int tmpStdE;
 if(monitor>2) {
	PE"<------------- Initialization of cAVmaxOld");
	if(Nexcl) PE" and cAVmaxExOld");
	tmpStdE=dup(fileno(stderr));
	if(freopen("/dev/null","w",stderr)!=NULL) PE"\n\tTemporary stderr suppression failed");
	PE":\n Disregard the next %s:\n", Nexcl?"2 lines":"line");
 }
 (void)cDiffMf();
 cAVmaxOld= cAVmax;
 if(Nexcl) cAVmaxExOld= cAVmaxEx;
 if(monitor>2) {
	FE
	dup2(tmpStdE,fileno(stderr));
	close(tmpStdE);
	PE" ------------->\n\n");
 }
}

/* Actions for various values of the flag for diagnostic output:
   if diagFlag= 0    ... no printout; ++diagCounter;
		> 0  ... printout; initialize; diagCounter=1; --diagFlag
Initialize diagFlag:
*/
diagFlag= 2;
/* diagFlag= 2 causes a diag output to be marked by a comment char AFTER the line;
 * Used for the very first call to runband in which cDiffM may be huge due to bad
 * initial guess; and optionally for the very 1st step after tStep modified in
 * update_Other (what used to be RESETtStepAT99): */
diagCounter= 0;

#if PDMtrace
PE"... init_Other\n");
#endif
init_Other();
if(init_from_file && init_from_file_dyn) free_str(1,&init_from_file);
#if NodeNumberChangeENABL
if(nnodes!=oldNnodes) { // nnodes can be changed in a PDM in init_Other
	NtN= nnodes*neqns;
	oldNnodes= nnodes;
	changeNoNodes(nnodes);
}
#endif
if(monitor>0) {
 PE"<----- Brief explanation of the following stderr output produced by TRANSIENT\n\
\t\t(PDM may intersperse it with additional output):\n");
 reportDEBUG();
 PE"In each time step a line ");
 if(monitor2) {
  PE"of the form\n\
 iter #: (fresidT) fresidM node# eq#  (cresidT) cresidM node# eq#\n\
will be printed after each NR iteration (node#, eq# ~ where f/cresidM occur).\n");
  if(monitor>2) {
   PE"Then ");
   if(Nexcl) {
	PE"on 2 lines cDiffM and cDiffMExcl, and where they occur:");
	CDIFFM_REP(\n\040cDiffM(used),all unexcluded\040,-)
	CDIFFM_REP(/cAVmax \054\n\040\040\040 cAVmax,all unexcluded\040,| and |)
	PE" = MAX(|cU|max,|c_oU|max)");
	CDIFFMex_REP(\n\040cDiffMExcl,-)
	CDIFFMex_REP(/cAVmaxEx \054\n\040\040\040 cAVmaxEx,| and |)
	PE" = MAX(|cE|max,|c_oE|max)");
   } else {
	PE"a line with cDiffM and where it occurs:");
	CDIFFM_REP(\n\040cDiffM,all\040,-)
	CDIFFM_REP(/cAVmax \054\n\040\040\040 cAVmax,all\040,| and |)
	PE" = MAX(|c|max,|c_o|max)");
   }
   PE".\n");
  }
  PE"Then a line with current time, cDiffM, current tStep, new tStep Scale factor,\n");
 } else PE"with time, cDiffM, tStep, # of attempts, # of all iterations,\n");
 PE"and cumulative #'s of time steps, runband calls and iterations, and\n\
the \"specific\" value of cDiffM per time unit ");
 if(monitor>2 && Nexcl) PE"(cDiffM(used) divided by tStep)\n");
 else PE"(cDiffM divided by tStep)\n");
 if(mPause) {
	static const char * const sentence[] = {"------\n", " Giving you time",
		"to read", "the above information.\n"};
	int i=0;
	for(; i<4; i++) {
		sleep(1);
		PE" %s", sentence[i]);
	}
	PE"  To skip this pause, next time use monitor = %d for the same verbosity.\n\
  Will continue in 25 seconds.", monitor*10);
	sleep(25);
 }
 PE" ----->\n"); FE
}
startTransClock();
while(ttOld < ttEnd) { // Top cycle - each loop of which ends with output of profiles
  for(k=1; k<=nOutOTH; k++) {	// Secondary cycle - each loop ends at the time of
				// a regular output of "Other" values
    ttOut += DttOut;
    if(ttOut > 0.999995*ttEnd) { ttOut = ttEnd; k = nOutOTH; } // if(ttOut >= ttEnd) ...
    if(tt > ttOut) { /* This may happen only near ttEnd
		for large ttEnd when ttStep > DttOut, or when !exactOutOTH ,
		or when e.g. tOutInit or other profile output time is set equal tEnd,
		but the respective ttOut differs slightly from ttEnd due to
		the rounding error (when adding repeatedly DttOut).
		This is the only case when, after making the following adjustments,
		the new initial guess will not be obtained by proper extrapolation: */
	if(exactOutOTH) {
	  tt = ttOut;
	  if(monitor2) PE"The above Scale=%g changed to %g after printout.\n",
			Scale, Scale*(ttOut-ttOld)/ttStep);
	  ttStep = ttOut - ttOld;
	} else {
		while(tt>ttOut+DttOut && k<nOutOTH) { ttOut += DttOut; k++; }
		if(tt > ttOut) {
			tt = ttOut;
			if(monitor2) PE"The above Scale=%g changed to %g after printout.\n",
					Scale, Scale*(ttOut-ttOld)/ttStep);
			ttStep = ttOut - ttOld;
		}
	}
    }
    do { // Lowest cycle - each loop is one time step - the whole cycle covers the
	 // interval between between two regular pr_Other times
	register double aux;
	register int updtOthCode;

	//PE"     k=%d, tt=%g, ttOut=%g\n", k, tt, ttOut);
	attempt = 1;
	allIter = 0;
	// We are at the start of a new tStep, and t and tt are set to the time instant
	// for which the solution is just starting now.
	do_Oth_Prof |= 12u; /* This sets the 3rd and 4th bit to 1 to indicate that
		until this step is completed (and do_Oth_Prof reset correspondingly),
		c_0 and the "Other" values correspond to t-tStep (i.e., they contain
		the results for the previous time instant t-tStep). This instructs
		out_Other() and out_profiles what value of time to use if the
		run is terminated for whatever reason before this tStep is completed,
		and premature output is forced by exitTR() */
	cDiffMprev= -1.;
    Try_again:
	t = tSign * tt + tStart;
	tStep = tSign * ttStep;
	tStepInv = 1./tStep;
#if PDMtrace
	PE"... init_tStep_for_evalf\n");
#endif
	init_tStep_for_evalf();
#if PDMtrace
	PE"... evalf (repeatedly from runband)\n");
#endif
	ConvCode= runband(t);
	// End of output line started in nl3bandv.c (intent was: figure out some
	// correlations to be used to improve time-stepping algorithm):
	//PE" %g %g %g convg2\n", tt, ttStep, ttStep/ttStepPrev);
	runbandCounter++;
	if(ConvCode<0. || ConvCode>100.) {
	  allIter += NRiter;
	  if(NRiter > maxIterDone) maxIterDone = NRiter;
	  if(monitor>0) {
		PE"  ^- attempt #%d: t=%le, tStep=%.4g, ", attempt, t, tStep);
		if(ConvCode==-8.) PE "inversion failed (-8)\n");
		else if(ConvCode==-9.) PE"cresidT or fresidT not finite (-9)\n");
		else PE"ConvCode=%g\n", ConvCode);
		FE
	  }
	  if(ConvCode==-8.) {
	    if(attempt<3) goto ANOTHER_ATTEMPT;
	    PR"\n%cInversion failed after %d attempts for t = %le\n\n",
			CmntC, attempt, t);
	    report_last_NR_error(out, CmntC);
	    exitTR(-8,"Inversion failed in runband/lufac");
	  }
	  if(ConvCode==-9.) {
	    if(attempt<3) goto ANOTHER_ATTEMPT;
	    PR"\n%ccresidT or fresidT not finite after %d attempts for t = %le\n\n",
			CmntC, attempt, t);
	    report_last_NR_error(out, CmntC);
	    exitTR(-9,"cresidT or fresidT not finite");
	  }
	ANOTHER_ATTEMPT:
	  attempt++;
#if attemptMAX>0
	  if(attempt>attemptMAX) exitTR(-4,"Hardcoded max # of attempts reached");
#endif
	  ttStep *= (ConvCode>0. ? pow(ConvCode,-0.13) : 0.35);
#if attemptMAX>0
	  if(ttStep<tStepInit) exitTR(-10,
		"Next attempt would decrease time step below tStepInit");
#else
	  if(attempt<=-attemptMAX && (ttStep>=tStepInit || ConvCode<-1)) {
#endif
		for(i=0; i<NtN; i++) c[i] = c_o[i];
		/* Also tried:
		2:	if(ConvCode<-1) for(i=0; i<NtN; i++) c[i] = c_o[i];
			else for(i=0; i<NtN; i++) c[i] = c_o[i] + 0.1 * (c[i]-c_o[i]);
		1:	Randomize(1e-3*attempt, c_o, c);
		Maybe none of this is needed? It works OK also without any c[] resetting.
		*/
		tt = ttOld + ttStep;
		goto Try_again;
#if attemptMAX<0
	  } else {
		// ConvCode<-1 cannot get here; up to 3 attemtps are allowed in the same
		// time step even if the reduced ttStep<tStepInit - see above
		DiagEntry_at_t();
		diagFlag= 1;
		diagCounter= 0;
		attempt -= 1;
		Pd"%c%.10g: WARNING: continuing despite %d failed attempts: ConvCode=%g\n",
		    CmntC, t, attempt, ConvCode); Fd
		Fd
	  }
#endif
	} else if(ConvCode>1.&&monitor>0) PE"ConvCode=%.14g\n", ConvCode);

	allIter += NRiter;
	if(NRiter > maxIterDone) maxIterDone = NRiter;
	iterCounter += allIter;

	{ /* Calculation of new tStep and tt: */
	  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
	double cDiffM_rate;

	cDiffM = cDiffMf();
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
	if(strict01) {
#pragma GCC diagnostic pop
	  Scale = cDiffM==0.0 ? 1e99 : cDifMxf(tt)/cDiffM;
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
	  if(strict1 && Scale<1.) {
#pragma GCC diagnostic pop
	    /*	Reduce the current time step to enforce that the actual
		cDiffM never exceeds the current limit set by cDifMxf(tt): */
	    if(monitor2) PE"  ^- attempt #%d: strict: t=%le, tStep=%.4g, cDiffM=%g\n",
			attempt, t, tStep, cDiffM); FE
	    if(cDiffMprev<0.) cDiffMprev= cDiffM;
	    else {
		int small_tStep=0;
		// cDiffM may initially increase a bit, so start testing after 25 attempts
		if((attempt>25 && cDiffM>=cDiffMprev) ||
				(ttStep<tStepInit && (small_tStep=1))) {
		   // cDiffM stopped decreasing (or tStep too small)
		   if(Scale<0.97) {
			strict1failCount++;
			PE" t=%g: strict=1 failed to reduce cDiffM for the %ld%s time\n",
				t, strict1failCount, (strict1failCount==1?"st":
				(strict1failCount==2?"nd":(strict1failCount==3?"rd":"th"))));
			FE
			DiagEntry_at_t();
			diagFlag= 1;
			diagCounter= 0;
			Pd"%c%.13g: WARNING: strict=1 failed to reduce cDiffM=%g to\
 %g (attempt %d", CmntC, t, cDiffM, Scale*cDiffM, attempt);
			if(small_tStep) Pd"; due to small tStep");
			Pd")\n");
			Fd
		   } // else let's accept it (give no warning) as it's very close to the target
		   goto CONTINUE;
		} else if(cDiffM<cDiffMprev) cDiffMprev= cDiffM;
	    }
	    if(Scale>0.05) { if(Scale<=0.1) Scale=0.05; else Scale *= 0.5; }
	    ttStep *= Scale;
	    attempt++;
	    /* Interpolate to get initial guess for the reduced ttStep: */
	    Scale *= 0.5; // decreasing arbitrarily the weight of the not yet good enough c
	    for(i=0; i<NtN; i++) c[i] = c_o[i] + Scale * (c[i]-c_o[i]);
	    //for(i=0; i<NtN; i++) c[i] = c_o[i];
	    tt = ttOld + ttStep;
	    goto Try_again;
	  }
	} /* End of: if(strict01) */

	// Done with (all attempts for) the current time step:
    CONTINUE:
	cDiffM_rate = cDiffM/ttStep;
	cDiffM_sum += cDiffM;
	updtDiag(cDiffM);
	ttStepLast = ttStep;
	i= NRiter-nIterM;
	//PE"i=%d, ttStep=%g, Scale=%g, ttStepLast=%g, ttStepPrev=%g\n",
	//	i, ttStep, Scale, ttStepLast, ttStepPrev);
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
	if(strict01) {
#pragma GCC diagnostic pop
	   if(ConvCode>1.) {
		aux= (199.-ConvCode)/198.; //pow(ConvCode,-0.14); 51./(50.+ConvCode)
		if(Scale>aux) Scale= aux;
	   } //else if(attempt>1 && Scale>0.999) Scale= 0.999;
	   if(i>0) { // decrease next time step:
		/* Using aux=1/f, where  f = 2+A - (1+A) * (B+G)/(B+G*x) ;
		f(1)=1, f(infinity)=2+A; B and G control how it behaves in between.
		That is: aux = ((B+G)+G*i)/[(B+G)+G*(A+2)*i], where i=x-1 (x=i+1).
		For A=5, B=20 and G=0.05 it gives: aux=(20.05+0.05*i)/(20.05+0.35*i);
		For A=98, B=20 and G=0.05: aux=(20.05+0.05*i)/(20.05+5.*i); */
		//aux=(88.05+0.05*i)/(88.05+6.*i); ... for A=118, B=88, G=0.05
		aux=(121.+0.05*i)/(121.+6.1*i); // ... A=120, B=120.95, G=0.05
		if(Scale>aux) Scale= aux;
	   } else {
		if(ttStepPrev>ttStepLast) {
			aux= 0.5*(ttStepPrev+ttStepLast)/ttStepLast;
			if(Scale>aux) Scale= aux;
		}
		if(Scale>maxScale) Scale= maxScale;
	   }
	} else {
	   if(ConvCode>1.) Scale= (199.-ConvCode)/198.; //pow(ConvCode,-0.14); 51./....
	   else {
		if(i>0) { // decrease next time step, as above:
			Scale=(121.+0.05*i)/(121.+6.1*i);
		} else {
			if(attempt>1) Scale= 0.999;
			else {	// increase next time step:
				// NRiter=1 happens only very exceptionally (see nl3_conv.h),
				// thus NRiter=2 (i.e., i=2-nIterM) is the usual minimal
				// NRiter, and this should correspond to maxScale:
				Scale= 1. - i*(maxScale-1.)/(nIterM-2.);
				// (This exceeds maxScale only for the rare case of NRiter=1)
			}
		}
	   }
	}
	ttStep *= Scale;
	ttStepPrev = ttStepLast;
	ttOld = tt;
	tt = ttOld + ttStep; /* This is the new time for the next step */
	if(exactOutOTH) {
	   if(ttOld<ttOut && ttOld>0.999999*ttOut) ttOld= ttOut;
	   if(ttOld >= ttOut) {
		// here we are finishing tStep that hit ttOut exactly, after adjustments below
		if(ttStep > 1.000005* DttOut) {
			// to not jump over the whole next DttOut interval
			ttStep= DttOut;
			tt = ttOld + ttStep;
			Scale= ttStep / ttStepLast;
			//PE"exactOutOTH: ttStep reduced to DttOut\n");
		}
	   } else if(tt>1.00001*ttOut) {
		//PE"exactOutOTH 0: tt-ttOut=%g; tt=%g; ttOld=%g\n", tt-ttOut, tt, ttOld);
		tt = ttOut;
		ttStep = tt - ttOld;
		Scale= ttStep / ttStepLast;
		//PE"exactOutOTH 0a: tt=%g; ttStep=%g; ttOut=%g\n", tt, ttStep, ttOut);
	   } else if((aux=(ttOut-ttOld)/ttStep)<6.) {
		// to eventually hit ttOut exactly so that
		// ttStep is never increased too much
		i= ceil(aux);
		if(i==1) {
		   //PE"exactOutOTH: tt=%g; ttOld=%g; ttStepLast=%g; ttStep adjusted: from %g",
		   //	tt, ttOld, ttStepLast, ttStep);
		   tt = ttOut;
		   ttStep = tt - ttOld;
		} else {
		   /* if(i-aux>0.6 && aux*Scale<(i-1)*maxScale) i--;  <- this would make
			sure that maxScale is never exceeded, but seems to unnecessarily
			decrease tStep too much (always when Scale=maxScale, i.e., when
			it's safe to increase tStep a lot); so a compromise instead: */
		   if((i<6||Scale>=1.) && i-aux>0.85) i--;
		   //PE"exactOutOTH: i=%d; ttStep adjusted: from %g", i, ttStep);
		   ttStep = (ttOut - ttOld)/i;
		   tt = ttOld + ttStep;
		   // The same ttStep should be obtained for all i steps. But some
		   // other step reducing process (tSOne, nodeB) can interfere, and so
		   // it's better to repeat this calculation for all subsequent i.
		}
		Scale= ttStep / ttStepLast;
		//PE" to %g; newScale=%g\n", ttStep, Scale);
	   }
	}
	stepCounter++;
	// Resetting at this point ttStep<tStepInit to tStepInit was a bad idea,
	// breaking Scale, and more.
	if(monitor>0) {
		//if(Scale>maxScale) PE"!!!!! Scale=%g > maxScale !!!\n", Scale);
		PE"%.8le %g %.8le", t, cDiffM, tStep);
		if(monitor2) PE" %-18.17g", Scale);
		else PE" %2d %2d", attempt, allIter);
		PE" %ld %ld %ld  %.3e\n", stepCounter,runbandCounter,iterCounter,cDiffM_rate);
		FE
	}

#if preV4_3
	/* Extrapolate to get initial guess for c[][] for new tt: */
	{ register double NIGscale= NIG*Scale;
	  for(i=0; i<NtN; i++) {
		aux = c[i] + NIGscale * (c[i] - c_o[i]);
		c_o[i]= c[i];
		c[i]= aux;
	  }
	}
	do_Oth_Prof &= 23u;
		/* 4th bit set to 0 : now c_0 correspond to t */
	do_Oth_Prof |= 2u;
		/* 2nd bit set to 1 : new c_0 available for output */
	ApplyBounds(c);
#endif
	tStepLast = tStep; /* For export into PDM (mainly for use in update_Other() */
	} /* End of Calculation of new ttStep and tt; and for preV4_3, the guess for new c */
	  /* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

	/* Diagnostic output (collected above) every pr_Other time, but usually
	   not sooner than after 10 steps and not later than after 1000 steps: */
	if(++diagCounter == 1000) diagFlag= 1;
	if(monitor<1 && !(diagCounter&7u)) PE".");
	if(diagFlag) {
		Pd"%#15.10g  %10.3e %10.3e  %10.3e %10.3e %4d %4d %6d",
			t-tStep, tStepFloor, tStepCeil, cDiffMmulti, cDiffMceil,
			allIterFloor, allIterCeil, diagCounter);
		if(diagFlag == 2) Pd"  %c", CmntC);
		Pd"\n"); Fd
		diagFlag--;
		diagCounter= 0;
	}

	/* The new time step to be made available for resetting in update_Other(): */
	tStep= tSign*ttStep;
#if PDMtrace
	PE"... update_Other\n");
#endif
	updtOthCode = update_Other();	// <-------  UPDATE_OTHER ----------
	do_Oth_Prof &= 27u;
		/* 3rd bit set to 0: "Other" values now correspond to t */
	do_Oth_Prof |= 1u;
		/* 1st bit set to 1: new "Other" values not yet outputted */
#if preV4_3
	if(updtOthCode) out_Other(k<nOutOTH || ttOld<ttOut);
	if(updtOthCode<0) exitTR(1000-updtOthCode,
		"Termination requested from update_Other at this time");
#endif
#if NodeNumberChangeENABL
	if(nnodes!=oldNnodes) { // can be changed in a PDM in update_Other
		NtN= nnodes*neqns;
		oldNnodes= nnodes;
		changeNoNodes(nnodes);
	}
#endif

#if preV4_3 && RESETtStepAT99
	/* If macro RESETtStepAT99 is defined and nonzero:
	 This segment of code causes to reset the time step
	to the initial value whenever update_Other() returns 99.
	Also resets c to c_o.
	 If the return value is positive and different from 99, ttStep is
	only reset to ttStep/pow(10.,i-1).
	 It was being used to treat moving internal boundaries between two
	media (e.g., water and growing/receding layer of porous precipitate):
	after each boundary move one has, strictly speaking, a new model for
	which the transient solution has to be started from the beginning.
	*/
	if(updtOthCode>1) {
		if(monitor2) aux = ttStep;
		if(updtOthCode==99) ttStep = tStepInit;
		else {
		    ttStep = ttStep/pow(10.,(double)--updtOthCode);
		    if(ttStep<tStepInit) ttStep = tStepInit;
		}
		tt= ttOld + ttStep;
		for(i=0; i<NtN; i++) c[i]= c_o[i];
		ttStepPrev= ttStep;
		diagFlag= 3; // to cause dumping the currently accumulated diag output,
			// followed, after reinitialization, by a marked line,
			// and then another single tStep line.
		if(monitor2) PE"\"RESETtStepAT99\": Scale reduced from %g to %g\n",
				Scale, Scale*ttStep/aux);
	} else {
#endif

	aux= tSign*tStep;
	if(aux<ttStep) {
		if(monitor2) PE"|tStep| decreased in update_Other() from %g to %g\n",
				ttStep, aux);
#if !preV4_3
		Scale *= aux/ttStep; // updated for new initial guess extrapolation
#endif
		ttStep = aux;
		tt= ttOld + ttStep;
#if !preV4_3
		ttStepPrev= ttStep;
		diagFlag= 3; // to cause dumping the currently accumulated diag output,
			// followed - after reinitialization - by a marked line,
			// and then another single tStep line.
#endif
	}
#if !preV4_3
	/* Extrapolate to get initial guess for c[][] for new tt: */
	{ register double NIGscale= NIG*Scale;
	  for(i=0; i<NtN; i++) {
		aux = c[i] + NIGscale * (c[i] - c_o[i]);
		c_o[i]= c[i];
		c[i]= aux;
	  }
	}
	do_Oth_Prof &= 23u;
		/* 4th bit set to 0 : now c_0 correspond to t */
	do_Oth_Prof |= 2u;
		/* 2nd bit set to 1 : new c_0 available for output */
	ApplyBounds(c);
	// In case pr_Other also uses c_o, acting upon updtOthCode was delayed until here:
	if(updtOthCode) out_Other(k<nOutOTH || ttOld<ttOut);
	if(updtOthCode<0) exitTR(1000-updtOthCode,
		"Termination requested from update_Other at this time");
#endif

#if  preV4_3 && RESETtStepAT99
	}
#endif
	//PE"k=%d: ttOut-ttOld=%g; ttOld=%g; ttOut=%g\n", k, ttOut-ttOld, ttOld, ttOut); FE
    } while(ttOld < ttOut); /* End of cycle between two pr_Other events */

    if(diagCounter > 9) diagFlag= 1;
    out_Other(k<nOutOTH);  /* do_Oth_Prof reset there */
    if(!monitor) {
	PE": cDiffM=%-.3le, # tSteps=%ld, # attempts=%ld, # iters=%ld ",
		cDiffM, stepCounter, runbandCounter, iterCounter);
    }
    Fm
  } /* End of for(k=1; k<=nOutOTH; k++) */

  out_profiles(ttOld<ttEnd); /* do_Oth_Prof reset there, too */
  maxIterDone = 0;

  if(tOutMulF>1. && ttEnd-ttOut>nOutOTH*DttOut) DttOut *= tOutMulF;
} /* End of: while(ttOld < ttEnd) */

exitTR(0,"Completed successfully");

} /* End of main() */
