/* File transien.h; TRANSIENT V4.3; 30 Oct 2023
 * (C) 1995-2016,2019,2023 Miroslav Kolar

===================================================================
 Header file for inclusion into a problem description module (PDM)
===================================================================
*/

#include "incl_pri.h"

#include <stdlib.h>

#if defined(vms) || defined(VMS)
/* #include <unistd.h> ?
or: #define _Exit exit*/
#endif
#ifdef __DJGPP
#define _Exit exit
#endif


/* --------------------------------------------------------------------
 * TRANSIENT expects the following functions to be defined
 * in a problem definition module (PDM). Use the files PDMtemplate.c,
 * and demoX.c and demoX.h (X=1,2,3,4) as templates for coding these functions:	*/

int get_neqns_etc(void);
void generate_grid(void);
void Initialize(double *cp, double *c_op);
void read_profiles(char *filename);
void auxiliary_parameters_for_evalf(void);
void init_Other(void);
void init_tStep_for_evalf(void);
int update_Other(void);
void pr_Other(int only_Other);
void pr_profiles(int MaxIter);
void initNL3BANDV(double***LBp,double***UBp,double*F,double*A,double*B,double*D);
void evalf(int j);
/* -------------------------------------------------------------------- */


/* ============================================================================	*/
/* Functions and variables exported currently from TRANSIENT into a PDM module
 * (PDM authors can of course use any additional functions defined in TRANSIENT,
 * listed in Section C of doc/function_lists.txt, by adding respective extern
 * declarations to their PDM; see an example, 'extern void Randomize()' in
 * demo/demo1/demo1.c):	*/

/* 1. All 8 functions from dynarr.h that is included through incl_pri.h above,
	i.e.:
extern int *ivector(int n);
extern double *dvector(int n);
extern double *rdvector(double *v, int n);
extern double **pdvector(int nx);
extern char *makestr(int n);
extern void free_dvector(double **A);
extern void free_ivector(int **A);
extern void free_str(short int by_malloc, char **S);
*/

// 2. ABSV() (macro) from defs.h, also through incl_pri.h above

/* 3. Following 11 functions declared extern below:
 *			exitTR, report_max_NR_error, changeBound,
 * 			out_profiles, out_Other, linInter,
 *			show_dmatrix, show_dvector_f, show_dvector, show_ivector,
 *			pathToThisExecutable (this one only #if PATHtoITSELF set
 *			to 1 also in PDM, e.g. when set in defs.h):
 */

extern void exitTR(int status, char *descr) __attribute__ ((noreturn));
/* When, for whatever reason, terminating the execution from a PDM,
   use exitTR() (instead of exit()). exitTR will update all the output
   files with the results available at the time of termination,
   complete the last line of the 'Output'.diag (diag_out) file, and
   generate the	statistics file, 'Output'.stat (stat_out).

   status (exit code) meaning:
	0	 success;
	-1,-2,-3 reserved for TRANSIENT's initialization failures;
	-4	 convergence not achieved (max # of attempts reached);
	-5	 errors in IN_(STR)PAR macros: unknown format string, illegal default
			value, or overlong parameter name or input value;
	-6	 when attempting to change the number of grid nodes for steady=2;
	-7	 when trying to increase the number of grid nodes above
			initially specified maximum;
	-8	 fatal failure in NL3BANDV (singular matrices encountered
			in all attempts for a time step);
	-9	 cresidT or fresidT is not finite;
	-10	 the time step becomes too small;
	-11	 user defined CHECK_COND() in PDM not satisfied;
	101 to (100+MAXint)     -
		 reserved for termination caused by interrupt catching;
	1 to 100, (100+MAXint+1) to 1000 -
		 reserved for PDM termination codes available to the PDM author
			to be used directly in exitTR();
			[MAXint is the largest interrupt code + 1, can depend on the CPU,
			usually MAXint=65; to find out for your system, see macro
			PRINT_EXIT_CODES_REMINDER in defs.h.]
	> 1000	 reserved for termination requested by the author from update_Other:
			i=update_Other()
			if i<0, execution is terminated with status = 1000-i.
*/

extern void report_max_NR_error(FILE *out, char comment_char);
	/* Defined in transien.c;  to be able, if the user so chooses, to report in
	 * the profiles output files the mean and maximum indicators of the quality
	 * of the convergence of the NR iteration process. */

extern void changeBound(int e, char which, double *BDA);
	/* To change type of "concentration" bound for equation e;
	 * which  is either 'l' (to change the lower bound) or 'u' (the upper bound);
	 * BDA is the address of the new (modified) Bound Definition Array. */

/* The next two functions call the PDM supplied pr_profiles() and pr_Other(),
   respectively. They first check, whether the profiles or the "Other" values
   may have already been printed for the same time. Thus their use avoids duplication:
*/
extern void out_profiles(int NewFile);
	/* Can be used to output additional profiles at arbitrary time instants,
	   in addition to the automatic ones determined by tOutInit and tOutMulF
	   - e.g. by calling out_profiles() at the beginning of evalf()
	   when certain conditions are satisfied.
	   The argument NewFile has an effect only if sepProf >= 0:
		NewFile=0 : the current profiles file will remain open and will
			    be used for the next profiles output, whether
			    automatic, or initiated from the PDM
			    (sepProf=1 temporarily suppressed).
		NewFile=1 : the current profiles file will be closed, and a new
			    one, with the numerical extension increased by one,
			    opened for the next profiles output
			    (standard sepProf=1 behavior).		*/
extern void out_Other(int only_Other);
	/* Can be used to output "Other" values at arbitrary time instants,
	   in addition to the automatic ones ...
	   Argument  only_Other  has the same meaning as in pr_Other().
	   (out_Other is called after completion of update_Other,
	   if update_Other returns a nonzero value.)			*/

extern double linInter(double IN, int NpointsM2, double *inb, double *v, int *lowI);
	/* Defined in InOut.c; to be used for the linear interpolation of
	   tabulated functions, if needed in a PDM.
	v[] are the tabulated function (dependent variable) values, for the
	independent variable values supplied in inb[];
	IN  is the independent variable value for which the function value
	is to be calculated;
	NpointsM2 is the # of points minus 2;
	*lowI is the lower index from the previous call - usually,
	the new value of  IN  will be close to the previous one. */

/* The next four functions, defined in matrixv.c, are for easy printing
   of 2D and 1D double arrays and 1D integer arrays onto the STDERR:.
   Arguments:  name  is an arbitrary string (a "title" of the array printout);
   m  or  v  is the pointer to the array to be printed out;
   order  is the number of columns of a matrix stored linearly;
   (l1,u1), (l2,u2) or (l,u) are the lower to upper ranges of indices of
   elements to be printed;  frm  is a valid C format string for a single
   double value that must start with a space: frm example: " %.12g".
*/
extern void show_dmatrix(char *name, int order, double *m,
			int l1, int u1, int l2, int u2, char *frm);
extern void show_dvector_f(char *name, double *v, int l, int u, char *frm);
extern void show_dvector(char *name, double *v, int l, int u);
extern void show_ivector(char *name, int *v, int l, int u);

#if PATHtoITSELF
extern char *pathToThisExecutable(void);
	/* This function may be useful if one needs to check whether some
	   additional input files needed by a TRANSIENT based program exist
	   in the same directory the executable resides in (the next obvious
	   choice after the current directory). This function returns a
	   pointer to the full or relative path of the program being
	   executed. If not needed anymore, the memory occupied by this path
	   can be released using free_str(&p) (from dynarr.c), where p is
	   the pointer returned by pathToThisExecutable().  */
#endif

/* --- And the following variables:
	Tau, NIG, t, tStep, tStepInv, tStepLast, steady, nnodes, monitor2,
 	par_out, timeDep_out, stat_out, out, title, CmntC, time_units, tStart,
	tEnd, tStepInit, tOutInit, tOutMulF, nOutOTH, init_from_file,
	strict, cDifMxMode, cDifMxInit, cDifMxEnd, cDiffMfExcl: */

extern double Tau;
	/* "implicitness" of the differencing scheme - used
	   only in evalf() (but is a "control" input) 	*/

extern double NIG;
	/* New Initial Guess extrapolation factor for the next time
	 * step. Should be non-negative. Preferably smaller than 1.
	 * Probably not (much) larger than 1. */

extern double t, tStep, tStepInv, tStepLast;
	/* all  are real UNTRANSFORMED times:
	 DURING a time-integration step, tStep = 1/tStepInv is the time
	step currently used to advance the solution to time t;
	t_tau = t - Tau/tStepInv.
	tStep may be reduced by TRANSIENT during the current step,
	but at the beginning of every new attempt (call to runband),
	tStepInv = 1./tStep always (for use in evalf).
	 AFTER a successful completion of a time-integration step,
	t is the current time, tStepLast is the last time step just
	used to reach t, c_o (in transien.c; passed to a PDM through the
	last argument of Initialize) contains current concentration values
	corresponding to t, tStep is a qualified guess for the next
	time step and c contains initial guess (extrapolation)
	for concentrations at t+tStep.
	 The magnitude of tStep may be REDUCED in PDM in update_Other.
	tStepInv is set to 1./tStep only after update_Other is processed.
	 Thus, always use c_0 in pr_profiles, update_Other and pr_Other.
	In Initialize(), put the starting concentrations for t=tStart
	into c_o and the initial guess for concentrations at
	t=tStart+tStepInit into c (but c = c_o is OK to get things going).*/

extern char *title; /* optional title - can be defined in get_neqns_etc	*/

extern int nnodes;  /* total # of nodes - must be set in generate_grid	*/

extern int monitor2;  /* Initially: monitor2 = monitor>1 ? monitor : steady;
			 If steady=2 it is redefined in the transient part of
			 the run as: monitor2 = monitor>1 ? monitor : 0; */

extern FILE *par_out, *timeDep_out, *stat_out, *out; /* all output files */

extern char CmntC;  /* comment character to mark comment
			(unplottable) lines in output files		*/

extern char *time_units; /* name of the time unit used - can be defined
				in get_neqns_etc			*/
extern int steady;
     /* if steady = 1 or 2, steady state is calculated (only/at first);
	then tStepInv = 0, t = tStart, tStepLast undefined. Also c_o is
	unreachable (see Note) in a PDM for steady=2 during the steady
	steady calculation only in a PDM. This doesn't matter because
	tStepInv = 0. When the steady-state calculation is completed,
	the steady state is ALWAYS available via the pointer to c_o passed
	as the last argument of Initialize (for pr_profiles, etc.). Inside
	transien.c, c_o=c during a steady state calculation.
	In the 2nd phase of a steady=2 run (during the transient
	calculation), everything is the same as when steady=0.		*/

extern double tStart, tEnd, tStepInit, tOutInit, tOutMulF;
extern int nOutOTH, strict;
extern int cDifMxMode;
extern double cDifMxInit, cDifMxEnd;
     /* Initial and final times, magnitude of the initial time step,
	initial c output time, output time multiplication factor, and
	relative c tolerance;
	pr_Other frequency factor, strict;
	the mode of interpolation of maximum allowed "relative" change
	in c (cDiffM) for strict>-1 between the initial and final values;
	and these initial, cDifMxInit, and final, cDifMxEnd, values of cDiffM:

	All the above 11 variables are exported into the PDM so that the user
	can define their CUSTOMIZED DEFAULT (most often used) values for
	each problem in the procedure get_neqns_etc(). Only some, or none,
	of these quantities need to be assigned values in get_neqns_etc().

	If tStart != 0 and/or tEnd < tStart, then  tEnd and tOutInit will
	be modified (this happens after get_neqns_etc() is invoked),
	because internally a transformed time is used that always starts
	from t=0 and then grows. tStart always retains its original value
	(to make ther inverse transformation possible).  tStepInit>0 always,
	as it is always the absolute value of the initial step.		*/

extern char *cDiffMfExcl;
     /*	This pointer is NULL by default, which corresponds to all dependent
	variables c contributing to various convergence indicators. It can be
	set in the PDM (e.g., in get_neqns_etc() or Initialize()) to point to
	a non-empty default 'exclusion list', or changed in the input file.
	The exclusion list is the list of indices of equations (or dependent
	variables c) that the changes of which will not be included in the
	calculation the "relative" change in c, cDiffM=|c-c_o|/max(c,c_o),
	in a one time step, and in the NR residuals cresidT/M and fresidT/M.
	The list is comma separated, and preceded by the number of the indices
	in the list and a colon, no spaces allowed. E.g., cDiffMfExcl = "3:2,3,9".
	Double quotes needed only in the C code in a PDM.
	Empty list (NULL pointer assignment) means that that all c are used. */

extern char *init_from_file; /* = pointer to the name of the profiles file
	produced by an earlier run to read initial profiles from. Can also
	be used in init_Other for additional initializations from the
	respective .tdep file, see for example demo1.c.
	It is equal to NULL if no such re-initialization needed (or not yet set). */

/* DO NOT CHANGE  t, tStepInv and tStepLast  at any time.
 * nnodes  can be changed only in init_Other() and update_Other()
 * provided it never exceeds its initial value set in generate_grid().
 * Changing  Tau  during the run may influence results.
 * NIG may be changed at any time. When it's changed during a run,
 * make sure it is nonnegative.  It should not exceed 1 (much).
 * Numerical results will not be affected by changing  CmntC  or
 * the strings pointed to by  title  and  time_units.

 * cDiffMfExcl and init_from_file have two associated variables of type short int,
 * cDiffMfExcl_dyn and init_from_file_dyn. These automatically keep track in TRANSIENT
 * whether cDiffMfExcl and init_from_file were created dynamically or hard-coded as
 * pointers to constant strings in the PDM. They are only important if one ever wanted
 * to use free_str() on cDiffMfExcl or init_from_file in the PDM. Then cDiffMfExcl_dyn
 * or init_from_file_dyn have to be used as the 1st argument of free_str().
 * And one has to include the respective extern declaration(s) in the PDM
 * (or uncomment them here):
 *	extern short int cDiffMfExcl_dyn;
 *	extern short int init_from_file_dyn;
 * ====================================================================	*/


/* Especially below this line, do not change anything unless you are
   100% sure what you are doing */

// Defined in InOut.c:
extern int read_inp_value(char *name, void *var, char *type, char *src_string);
extern int read_inp_str(char *name, char *var, size_t maxLeng, char *src_string);
extern int isTooLong(int lng);

#include <errno.h>

#ifndef _ADDITIONAL_SOURCE_
#define IN_PAR(name,type,initValue,oF_u,expl) type name=initValue;
#define IN_STRPAR(name,leng,initValue,expl) char name[leng+1] = initValue;
#else
/* This branch is used when PDM is split in several files,
   and this header file is included in each of them */
#define IN_PAR(name,type,initValue,oF_u,expl) extern type name;
#define IN_STRPAR(name,leng,initValue,expl) extern char name[leng+1];
#endif

#define CHECK_PAR(cond)

/* Model input parameter declaration: */
#include PDM_H

#undef IN_PAR
#undef IN_STRPAR

#ifndef _ADDITIONAL_SOURCE_

#ifndef ultrix
#define IN_PAR(name,type,iV,outFrmt_unit,expl) else if(!strcmp(str1,#name))\
        { if(!words) name = 0; else if(read_inp_value(#name,&name,#type,str2)) *wrong=1; }
#define IN_STRPAR(name,leng,iV,expl) else if(!strcmp(str1,#name))\
        { if(!words) name[0] = 0; else if(read_inp_str(#name,name,leng,str2)) *wrong=1; }
#else
#define IN_PAR(name,type,iV,outFrmt_unit,expl) else if(!strcmp(str1,"name"))\
        { if(!words) name = 0; else if(read_inp_value("name",&name,"type",str2)) *wrong=1; }
#define IN_STRPAR(name,leng,iV,expl) else if(!strcmp(str1,"name"))\
        { if(!words) name[0] = 0; else if(read_inp_str("name",name,leng,str2)) *wrong=1; }
#endif

int INPT_model_param(int words, char *str1, char *str2, int *wrong)
/*===============================================================*/
{
if(0){}
#include PDM_H
else return 0;
return 1;
}

#undef IN_PAR
#undef IN_STRPAR

#ifndef ultrix
#define IN_PAR(name,type,iV,outFrmt_unit,expl) if(Xplain>0) Pp"\n");\
 Pp"%s"#name" = "#outFrmt_unit"\n",cmntS,name);\
 if(Xplain>0 && strlen(#expl)) Pp"   "#expl"\n");
#define IN_STRPAR(name,leng,iV,expl) if(Xplain>0) Pp"\n");\
 Pp"%s"#name" = %s\n",cmntS,name);\
 if(Xplain>0 && strlen(#expl)) Pp"   "#expl"\n");
#else
#define IN_PAR(name,type,iV,outFrmt_unit,expl) if(Xplain>0) Pp"\n");\
 Pp"%sname = outFrmt_unit\n",cmntS,name);\
 if(Xplain>0 && strlen("expl")) Pp"   expl\n");
#define IN_STRPAR(name,leng,iV,expl) if(Xplain>0) Pp"\n");\
 Pp"%sname = %s\n",cmntS,name);\
 if(Xplain>0 && strlen("expl")) Pp"   expl\n");
#endif

void pr_model(int Xplain, char *cmntS)
/*==================================*/
{
Pp"%s\n",cmntS);
if(Xplain<2 || Xplain==5) Pp"%sPDM parameters:\n%s--------------\n",cmntS,cmntS);
if(Xplain==4) Xplain=0;
#include PDM_H
if(Xplain<0) Pp"%c end of input data\n%c -----------------\n%c\n",
 CmntC, CmntC, CmntC);
(void)fflush(par_out);
}

#undef IN_PAR
#undef IN_STRPAR

#ifndef ultrix
#define IN_PAR(name,type,iV,outFrmt_unit,expl) if(!strncmp(pName,#name,len)){\
	Pp #name" = "#outFrmt_unit"\n",name);\
	if(Xplain && strlen(#expl)) Pp"   "#expl"\n");}
#define IN_STRPAR(name,leng,iV,expl) if(!strncmp(pName,#name,len)){\
	Pp #name" = %s\n",name); if(Xplain && strlen(#expl)) Pp"   "#expl"\n");}
#else
#define IN_PAR(name,type,iV,outFrmt_unit,expl) if(!strncmp(pName,"name",len)){\
	Pp"name = outFrmt_unit\n",name);\
	if(Xplain && strlen("expl")) Pp"   expl\n");}
#define IN_STRPAR(name,leng,iV,expl) if(!strncmp(pName,"name",len)){\
	Pp"name = %s\n",name); if(Xplain && strlen("expl")) Pp"   expl\n");}
#endif

void pr_some_PDMpar(char *pName, int Xplain)
/*========================================*/
{
size_t len= strlen(pName);
if(len==0) return;
#include PDM_H
}

#undef IN_PAR
#undef IN_STRPAR

// ------ checking for illegal default values and long names in control and model header files -------

//To suppress warning: ignoring return value of strtoXX, declared with attribute warn_unused_result:
inline void faking_use(int /*@unused@*/ i) {}

int check_PAR_val_nam(char *name, char *type, char *def_value, char *where)
/*===================--==================================================*/
{
 char *end;
 int Ret = 0;

 errno = 0;
 if(*def_value == '\0') {
	// Compiler will not let this happen; the execution should never get here
	PE"%s: Empty default value given for \"%s\"\n", where, name);
	return 1;
 }
 if(isTooLong(strlen(name))) {
	PE"%s: The name of input parameter \"%s\" is longer than MAX1_STR_LEN-1\
 of InOut.c. Adjust one or the other.\n", where, name);
	return 1;
 }
 if(!(strcmp(type,"int") && strcmp(type,"short int") && strcmp(type,"long int")))
	faking_use((int)strtol(def_value, &end, 10));
 else if(!strcmp(type,"unsigned int")) {
	faking_use((int)strtoul(def_value, &end, 10));
	if(*end=='u' || *end=='U') end++;
 } else if(!strcmp(type,"float")) faking_use((int)strtof(def_value, &end));
 else if(!strcmp(type,"double")) faking_use((int)strtod(def_value, &end));
 else if(!strcmp(type,"long double")) faking_use((int)strtold(def_value, &end));
 else if(!strcmp(type,"char")) {
	if(def_value[0]!='\'') end= def_value;
	else if(def_value[2]!='\'') end= def_value+2;
	else end= def_value+3;
 } else {
	PE"%s: TRANSIENT cannot yet handle type \"%s\" used for input parameter \"%s\"\n",
		where, type, name);
	return 1;
 }
 if(errno != 0) { perror("strtol"); Ret= 1; }
 if(end == def_value || *end != '\0') {
	PE"%s: The default value of \"%s\" (%s), %s, has illegal trailing characters: %s\n",
		where, name, type, def_value, end);
	Ret= 1;
 }
 return Ret;
}

//To expand possible macros in defValue:
#define STRL1(s) #s
#define STRL2(s) STRL1(s)

void check_PAR_format(void)
/*=======================*/
{
 register int wrong=0;
 register char *pdm=strrchr(PDM_H,'/');
 pdm= pdm==NULL ? PDM_H : pdm+1;
#define IN_PAR(name,type,defValue,o,e) if(check_PAR_val_nam(#name,#type,STRL2(defValue),"CTRLpar.h")) wrong=1;
#include "CTRLpar.h"
#undef IN_PAR
#define IN_PAR(name,type,defValue,o,e) if(check_PAR_val_nam(#name,#type,STRL2(defValue),pdm)) wrong=1;
#define IN_STRPAR(name,leng,defV,e) if(strlen(defV)>leng) {\
    PE"%s: Insufficient string length reserved in IN_STRPAR for the default value of \""#name"\"\
 which is %d chars long.\n", pdm, (int)strlen(defV)); wrong=1; }\
else if(isTooLong(strlen(#name))) {\
    PE"%s: The name of input parameter \""#name"\" is longer than MAX1_STR_LEN-1\
 of InOut.c. Adjust one or the other.\n", pdm); wrong=1; }
#include PDM_H
#undef IN_PAR
#undef IN_STRPAR
 if(wrong) {
    PE" ... CORRECT THE ABOVE ERROR(S) AND RECOMPILE ...\n");
    _Exit(-5);
 }
}

// ----------------- end of parameter names and default values checking --------------------

#undef CHECK_PAR

#define IN_PAR(a,b,c,d,e)
#define IN_STRPAR(a,b,c,d)

#ifndef ultrix
#define CHECK_PAR(cond) if(!(cond)){PE"  Required: "#cond"\n");\
	Ps"  Required: "#cond"\n"); wrong=1;}
#else
#define CHECK_PAR(cond) if(!(cond)){PE"  Required: cond\n");\
	Ps"  Required: cond\n"); wrong=1;}
#endif

int WRONG_model_param(void)
/*=======================*/
{
 register int wrong=0;
#include PDM_H
 return wrong;
}

#undef CHECK_PAR

#endif // ! _ADDITIONAL_SOURCE_


// For use anywhere in a PDM *.c file:
#ifndef ultrix
#define CHECK_COND(cond) if(!(cond)){\
	PE"In file "__FILE__" on line "STRL2(__LINE__)" it is required: "#cond);\
	Ps"In file "__FILE__" on line "STRL2(__LINE__)" it is required: "#cond);\
	exitTR(-11,"Check the above error; recompile if needed");}
#else
#define CHECK_COND(cond) if(!(cond)){\
	PE"In file "__FILE__" on line "STRL2(__LINE__)" it is required: cond");\
	Ps"In file "__FILE__" on line "STRL2(__LINE__)" it is required: cond");\
	exitTR(-11,"Check the above error; recompile if needed");}
#endif
